/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2022, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     Rom.cpp
    \brief    global variables & functions
*/

#include "Rom.h"
#include "UnitTools.h"

#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <iomanip>

// ====================================================================================================================
// Initialize / destroy functions
// ====================================================================================================================

#if ENABLE_TRACING
CDTrace *g_trace_ctx = NULL;
#endif
bool g_mctsDecCheckEnabled = false;

//! \ingroup CommonLib
//! \{

MsgLevel g_verbosity = VERBOSE;
#if SIGN_PREDICTION
#if JVET_Y0141_SIGN_PRED_IMPROVE
#if JVET_W0119_LFNST_EXTENSION || EXTENDED_LFNST
int8_t * g_resiBorderTemplateLFNST[6][6][210];
#else
int8_t * g_resiBorderTemplateLFNST[6][6][16];
#endif
int8_t * g_resiBorderTemplate[6][6][NUM_TRANS_TYPE*NUM_TRANS_TYPE];
#else
const int8_t * g_resiBorderTemplate[6][6][NUM_TRANS_TYPE*NUM_TRANS_TYPE];
#endif
// g_initRomSignPred: Format for each [W][H][Idx] in [0-5][0-5][0-8], 0 => empty, 1 => (W+H-1)*16 template coefficients
#if !JVET_W0103_INTRA_MTS
const int8_t g_initRomSignPred[] = {
 1,8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, -8, -8, 8, 5, 5, 5, 5, -10, 10, -4, -10, -4, 5, 10, 10, 10, 10, -13, -6, 6, 13, 6, -6, -13, -10, -4, 5, 10,
 -10, -10, 10, -6, -3, 3, 6, -13, 13, -6, 8, -8, -8, 8, 8, 8, 8, 10, -10, -10, 10, 5, -4, -10, 8, -8, -8, 8, -8, -8, 8, 5, -4, -4, 5, -10, 10, -4, -4, 10, -10, 5, 5, 5,
 5, -6, 13, -13, 6, 3, -3, -6, -4, 10, -10, 5, -4, -4, 5, -3, 6, -6, 3, -6, 6, -3, 0,0,0,1,5, 9, 12, 14, 12, 9, 5, 4, 8, 11, 12, 0, -12, -12, 3, 6, 8, 9, -12, -5, 14, 2,
 3, 4, 5, -12, 14, -9, -12, -12, 0, 12, 11, 8, 4, -11, -11, 0, 11, 0, -11, -11, -8, -8, 0, 8, -11, -4, 12, -4, -4, 0, 4, -11, 12, -8, 14, -5, -12, 9, 8, 6, 3, 12, -4,
 -11, 8, 0, -8, -8, 9, -3, -8, 6, -8, -3, 9, 5, -2, -4, 3, -8, 9, -6, -9, 14, -12, 5, 4, 3, 2, -8, 12, -11, 4, 0, -4, -4, -6, 9, -8, 3, -4, -2, 5, -3, 5, -4, 2, -4, 5,
 -3, 1,14, 12, 9, 5, 4, 3, 2, 12, 11, 8, 4, 0, -4, -4, 9, 8, 6, 3, -4, -2, 5, 5, 4, 3, 2, -4, 5, -3, -12, 0, 12, 12, 11, 8, 4, -11, 0, 11, 11, 0, -11, -11, -8, 0, 8, 8,
 -11, -4, 12, -4, 0, 4, 4, -11, 12, -8, 9, -12, -5, 14, 12, 9, 5, 8, -11, -4, 12, 0, -12, -12, 6, -8, -3, 9, -12, -5, 14, 3, -4, -2, 5, -12, 14, -9, -5, 12, -14, 9, 8, 6,
 3, -4, 11, -12, 8, 0, -8, -8, -3, 8, -9, 6, -8, -3, 9, -2, 4, -5, 3, -8, 9, -6, 0,1,2, 3, 4, 5, 9, 12, 14, 4, 8, 11, 12, 12, 0, -12, 5, 9, 12, 14, -5, -12, 9, 3, 6, 8,
 9, -14, 12, -5, -4, -4, 0, 4, 8, 11, 12, -11, -11, 0, 11, 11, 0, -11, -12, -12, 0, 12, -4, -11, 8, -8, -8, 0, 8, -12, 11, -4, 5, -2, -4, 3, 6, 8, 9, 12, -4, -11, 8, 8,
 0, -8, 14, -5, -12, 9, -3, -8, 6, 9, -3, -8, 6, -9, 8, -3, -3, 5, -4, 2, 3, 4, 5, -8, 12, -11, 4, 4, 0, -4, -9, 14, -12, 5, -2, -4, 3, -6, 9, -8, 3, -5, 4, -2, 1,5, 4,
 3, 2, 3, 4, 5, 12, 11, 8, 4, 4, 0, -4, 14, 12, 9, 5, -2, -4, 3, 9, 8, 6, 3, -5, 4, -2, -4, 0, 4, 4, 8, 11, 12, -11, 0, 11, 11, 11, 0, -11, -12, 0, 12, 12, -4, -11, 8,
 -8, 0, 8, 8, -12, 11, -4, 3, -4, -2, 5, 9, 12, 14, 8, -11, -4, 12, 12, 0, -12, 9, -12, -5, 14, -5, -12, 9, 6, -8, -3, 9, -14, 12, -5, -2, 4, -5, 3, 6, 8, 9, -4, 11, -12,
 8, 8, 0, -8, -5, 12, -14, 9, -3, -8, 6, -3, 8, -9, 6, -9, 8, -3, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, 8, 8, 8, 8,
 -8, -8, 8, 5, 5, 5, 5, 5, 5, 5, 5, -10, 10, -4, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, -14, -12, -8, -3, 3, 8, 12, 14, 6, -6, -14, -11, -9, -6, -2, 2, 6, 9, 11, -11,
 -11, 11, -6, -5, -4, -1, 1, 4, 5, 6, -14, 14, -6, 10, 5, -4, -10, -10, -4, 5, 10, 10, 10, 10, 13, 6, -6, -13, -13, -6, 6, 13, 6, -6, -13, 10, 5, -4, -10, -10, -4, 5, 10,
 -10, -10, 10, 6, 3, -3, -6, -6, -3, 3, 6, -13, 13, -6, -9, 2, 11, 6, -6, -11, -2, 9, 9, 9, 9, -12, 3, 14, 8, -8, -14, -3, 12, 5, -5, -12, -9, 2, 11, 6, -6, -11, -2, 9,
 -9, -9, 9, -5, 1, 6, 4, -4, -6, -1, 5, -12, 12, -5, 0,0,0,1,3, 5, 8, 10, 12, 13, 14, 14, 12, 9, 5, 2, 5, 7, 9, 10, 11, 12, 12, 0, -12, -12, 2, 3, 5, 6, 8, 8, 9, 9, -12,
 -5, 14, 1, 2, 3, 3, 4, 4, 5, 5, -12, 14, -9, -8, -13, -14, -12, -5, 3, 10, 14, 12, 9, 5, -7, -11, -12, -10, -5, 2, 9, 12, 0, -12, -12, -5, -8, -9, -8, -3, 2, 6, 9, -12,
 -5, 14, -3, -4, -5, -4, -2, 1, 3, 5, -12, 14, -9, 12, 14, 5, -8, -14, -10, 3, 13, 11, 8, 4, 10, 12, 5, -7, -12, -9, 2, 11, 0, -11, -11, 8, 9, 3, -5, -9, -6, 2, 8, -11,
 -4, 13, 4, 5, 2, -3, -5, -3, 1, 4, -11, 13, -8, -14, -8, 10, 13, -3, -14, -5, 12, 10, 8, 4, -12, -7, 9, 11, -2, -12, -5, 10, 0, -10, -10, -9, -5, 6, 8, -2, -9, -3, 8,
 -10, -4, 12, -5, -3, 3, 4, -1, -5, -2, 4, -10, 12, -8, 1,14, 14, 13, 12, 10, 8, 5, 3, 2, 2, 1, 12, 12, 11, 10, 9, 7, 5, 2, 0, -2, -2, 9, 9, 8, 8, 6, 5, 3, 2, -2, -1, 3,
 5, 5, 4, 4, 3, 3, 2, 1, -2, 3, -2, -14, -10, -3, 5, 12, 14, 13, 8, 7, 5, 3, -12, -9, -2, 5, 10, 12, 11, 7, 0, -7, -7, -9, -6, -2, 3, 8, 9, 8, 5, -7, -3, 8, -5, -3, -1,
 2, 4, 5, 4, 3, -7, 8, -5, 13, 3, -10, -14, -8, 5, 14, 12, 10, 8, 4, 11, 2, -9, -12, -7, 5, 12, 10, 0, -10, -10, 8, 2, -6, -9, -5, 3, 9, 8, -10, -4, 12, 4, 1, -3, -5, -3,
 2, 5, 4, -10, 12, -8, -12, 5, 14, 3, -13, -10, 8, 14, 12, 9, 5, -10, 5, 12, 2, -11, -9, 7, 12, 0, -12, -12, -8, 3, 9, 2, -8, -6, 5, 9, -12, -5, 14, -4, 2, 5, 1, -4, -3,
 3, 5, -12, 14, -9, 0,1,1, 2, 3, 3, 4, 4, 5, 5, 9, 12, 14, 2, 5, 7, 9, 10, 11, 12, 12, 12, 0, -12, 3, 5, 8, 10, 12, 13, 14, 14, -5, -12, 9, 2, 3, 5, 6, 8, 8, 9, 9, -14,
 12, -5, -3, -4, -5, -4, -2, 1, 3, 5, 9, 12, 14, -7, -11, -12, -10, -5, 2, 9, 12, 12, 0, -12, -8, -13, -14, -12, -5, 3, 10, 14, -5, -12, 9, -5, -8, -9, -8, -3, 2, 6, 9,
 -14, 12, -5, 4, 5, 2, -3, -5, -3, 1, 4, 8, 11, 13, 10, 12, 5, -7, -12, -9, 2, 11, 11, 0, -11, 12, 14, 5, -8, -14, -10, 3, 13, -4, -11, 8, 8, 9, 3, -5, -9, -6, 2, 8, -13,
 11, -4, -5, -3, 3, 4, -1, -5, -2, 4, 8, 10, 12, -12, -7, 9, 11, -2, -12, -5, 10, 10, 0, -10, -14, -8, 10, 13, -3, -14, -5, 12, -4, -10, 8, -9, -5, 6, 8, -2, -9, -3, 8,
 -12, 10, -4, 1,5, 5, 4, 4, 3, 3, 2, 1, 2, 2, 3, 12, 12, 11, 10, 9, 7, 5, 2, 2, 0, -2, 14, 14, 13, 12, 10, 8, 5, 3, -1, -2, 2, 9, 9, 8, 8, 6, 5, 3, 2, -3, 2, -1, -5, -3,
 -1, 2, 4, 5, 4, 3, 5, 7, 8, -12, -9, -2, 5, 10, 12, 11, 7, 7, 0, -7, -14, -10, -3, 5, 12, 14, 13, 8, -3, -7, 5, -9, -6, -2, 3, 8, 9, 8, 5, -8, 7, -3, 4, 1, -3, -5, -3,
 2, 5, 4, 8, 10, 12, 11, 2, -9, -12, -7, 5, 12, 10, 10, 0, -10, 13, 3, -10, -14, -8, 5, 14, 12, -4, -10, 8, 8, 2, -6, -9, -5, 3, 9, 8, -12, 10, -4, -4, 2, 5, 1, -4, -3,
 3, 5, 9, 12, 14, -10, 5, 12, 2, -11, -9, 7, 12, 12, 0, -12, -12, 5, 14, 3, -13, -10, 8, 14, -5, -12, 9, -8, 3, 9, 2, -8, -6, 5, 9, -14, 12, -5, 1,8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, -8, -8, 8,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -10, 10, -4, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, -15, -14, -13, -11, -9, -7, -4,
 -1, 1, 4, 7, 9, 11, 13, 14, 15, 6, -6, -15, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, -11, -11, 11, -6, -6, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6,
 6, 6, -15, 15, -6, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 14, 12, 8, 3, -3, -8, -12, -14, -14, -12, -8, -3, 3, 8, 12, 14, 6, -6, -14,
 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, -11, -11, 11, 6, 5, 4, 1, -1, -4, -5, -6, -6, -5, -4, -1, 1, 4, 5, 6, -14, 14, -6, -11, -7, -1, 5, 10, 11, 9,
 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, -14, -9, -1, 7, 13, 15, 11, 4, -4, -11, -15, -13, -7, 1, 9, 14, 6, -6, -14, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11,
 -10, -5, 1, 7, 11, -11, -11, 11, -6, -4, -1, 3, 6, 6, 5, 2, -2, -5, -6, -6, -3, 1, 4, 6, -14, 14, -6, 0,0,0,1,1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14,
 13, 9, 5, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 0, -13, -13, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, -13, -5, 14, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4,
 4, 5, 5, 5, 5, 5, -13, 14, -9, -4, -8, -11, -13, -14, -14, -13, -11, -8, -4, 0, 4, 8, 11, 13, 14, 13, 9, 5, -4, -7, -10, -12, -13, -13, -12, -10, -7, -4, 0, 4, 7, 10,
 12, 13, 0, -13, -13, -3, -5, -7, -9, -9, -9, -9, -7, -5, -3, 0, 3, 5, 7, 9, 9, -13, -5, 14, -1, -3, -4, -5, -5, -5, -5, -4, -3, -1, 0, 1, 3, 4, 5, 5, -13, 14, -9, 7, 12,
 14, 14, 10, 4, -3, -9, -13, -14, -13, -8, -1, 5, 11, 14, 13, 9, 5, 6, 11, 13, 12, 9, 4, -2, -8, -12, -13, -11, -7, -1, 5, 10, 13, 0, -13, -13, 4, 8, 9, 9, 7, 3, -2, -6,
 -9, -9, -8, -5, -1, 4, 7, 9, -13, -5, 14, 2, 4, 5, 5, 4, 1, -1, -3, -5, -5, -4, -3, 0, 2, 4, 5, -13, 14, -9, -9, -14, -13, -7, 3, 11, 14, 12, 4, -5, -13, -14, -10, -1,
 8, 14, 12, 9, 5, -8, -13, -12, -6, 2, 10, 13, 11, 4, -5, -11, -13, -9, -1, 7, 12, 0, -12, -12, -6, -9, -9, -4, 2, 7, 9, 8, 3, -4, -8, -9, -7, -1, 5, 9, -12, -5, 14, -3,
 -5, -5, -2, 1, 4, 5, 4, 1, -2, -4, -5, -4, 0, 3, 5, -12, 14, -9, 1,14, 14, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 1, 0, 13, 13, 13, 12, 12, 11, 11, 10, 9,
 8, 7, 6, 5, 4, 2, 1, 0, -1, -1, 9, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, -1, 0, 1, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, 1, -1, -14, -13, -11, -8,
 -4, 0, 4, 8, 11, 13, 14, 14, 13, 11, 8, 4, 4, 3, 1, -13, -12, -10, -7, -4, 0, 4, 7, 10, 12, 13, 13, 12, 10, 7, 4, 0, -4, -4, -9, -9, -7, -5, -3, 0, 3, 5, 7, 9, 9, 9, 9,
 7, 5, 3, -4, -1, 4, -5, -5, -4, -3, -1, 0, 1, 3, 4, 5, 5, 5, 5, 4, 3, 1, -4, 4, -3, 14, 11, 5, -1, -8, -13, -14, -13, -9, -3, 4, 10, 14, 14, 12, 7, 6, 4, 2, 13, 10, 5,
 -1, -7, -11, -13, -12, -8, -2, 4, 9, 12, 13, 11, 6, 0, -6, -6, 9, 7, 4, -1, -5, -8, -9, -9, -6, -2, 3, 7, 9, 9, 8, 4, -6, -2, 7, 5, 4, 2, 0, -3, -4, -5, -5, -3, -1, 1,
 4, 5, 5, 4, 2, -6, 7, -4, -14, -8, 1, 10, 14, 13, 5, -4, -12, -14, -11, -3, 7, 13, 14, 9, 8, 6, 3, -12, -7, 1, 9, 13, 11, 5, -4, -11, -13, -10, -2, 6, 12, 13, 8, 0, -8,
 -8, -9, -5, 1, 7, 9, 8, 4, -3, -8, -9, -7, -2, 4, 9, 9, 6, -8, -3, 9, -5, -3, 0, 4, 5, 4, 2, -1, -4, -5, -4, -1, 2, 5, 5, 3, -8, 9, -6, 0,1,0, 1, 1, 2, 2, 3, 3, 4, 4, 4,
 4, 5, 5, 5, 5, 5, 9, 13, 14, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 13, 0, -13, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, -5, -13, 9, 1,
 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, -14, 13, -5, -1, -3, -4, -5, -5, -5, -5, -4, -3, -1, 0, 1, 3, 4, 5, 5, 9, 13, 14, -4, -7, -10, -12, -13, -13, -12, -10, -7,
 -4, 0, 4, 7, 10, 12, 13, 13, 0, -13, -4, -8, -11, -13, -14, -14, -13, -11, -8, -4, 0, 4, 8, 11, 13, 14, -5, -13, 9, -3, -5, -7, -9, -9, -9, -9, -7, -5, -3, 0, 3, 5, 7,
 9, 9, -14, 13, -5, 2, 4, 5, 5, 4, 1, -1, -3, -5, -5, -4, -3, 0, 2, 4, 5, 9, 13, 14, 6, 11, 13, 12, 9, 4, -2, -8, -12, -13, -11, -7, -1, 5, 10, 13, 13, 0, -13, 7, 12, 14,
 14, 10, 4, -3, -9, -13, -14, -13, -8, -1, 5, 11, 14, -5, -13, 9, 4, 8, 9, 9, 7, 3, -2, -6, -9, -9, -8, -5, -1, 4, 7, 9, -14, 13, -5, -3, -5, -5, -2, 1, 4, 5, 4, 1, -2,
 -4, -5, -4, 0, 3, 5, 9, 12, 14, -8, -13, -12, -6, 2, 10, 13, 11, 4, -5, -11, -13, -9, -1, 7, 12, 12, 0, -12, -9, -14, -13, -7, 3, 11, 14, 12, 4, -5, -13, -14, -10, -1,
 8, 14, -5, -12, 9, -6, -9, -9, -4, 2, 7, 9, 8, 3, -4, -8, -9, -7, -1, 5, 9, -14, 12, -5, 1,5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 1, 1, 1, 13, 13, 13, 12, 12,
 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 0, -1, 14, 14, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 0, -1, 1, 9, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, -1, 1,
 0, -5, -5, -4, -3, -1, 0, 1, 3, 4, 5, 5, 5, 5, 4, 3, 1, 3, 4, 4, -13, -12, -10, -7, -4, 0, 4, 7, 10, 12, 13, 13, 12, 10, 7, 4, 4, 0, -4, -14, -13, -11, -8, -4, 0, 4, 8,
 11, 13, 14, 14, 13, 11, 8, 4, -1, -4, 3, -9, -9, -7, -5, -3, 0, 3, 5, 7, 9, 9, 9, 9, 7, 5, 3, -4, 4, -1, 5, 4, 2, 0, -3, -4, -5, -5, -3, -1, 1, 4, 5, 5, 4, 2, 4, 6, 7,
 13, 10, 5, -1, -7, -11, -13, -12, -8, -2, 4, 9, 12, 13, 11, 6, 6, 0, -6, 14, 11, 5, -1, -8, -13, -14, -13, -9, -3, 4, 10, 14, 14, 12, 7, -2, -6, 4, 9, 7, 4, -1, -5, -8,
 -9, -9, -6, -2, 3, 7, 9, 9, 8, 4, -7, 6, -2, -5, -3, 0, 4, 5, 4, 2, -1, -4, -5, -4, -1, 2, 5, 5, 3, 6, 8, 9, -12, -7, 1, 9, 13, 11, 5, -4, -11, -13, -10, -2, 6, 12, 13,
 8, 8, 0, -8, -14, -8, 1, 10, 14, 13, 5, -4, -12, -14, -11, -3, 7, 13, 14, 9, -3, -8, 6, -9, -5, 1, 7, 9, 8, 4, -3, -8, -9, -7, -2, 4, 9, 9, 6, -9, 8, -3, 1,8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, -8, -8,
 8, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -10, 10, -4, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4,
 -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 6, -6, -15, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11,
 11, -11, -11, 11, -6, -6, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, -15, 15, -6, 11, 11, 10, 9, 7, 5, 3, 1,
 -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14,
 -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, 6, -6, -15, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5,
 -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, -11, -11, 11, 6, 6, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -6, -6, -6, -6, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 6, 6, -15, 15,
 -6, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5, -8, -10, -11, -11, -11, -9, -7, -4, 0, 3, 6, 8, 10, 11, 11, 11, 11, -15, -13, -11, -7, -4, 1, 5,
 9, 12, 14, 15, 14, 13, 10, 6, 2, -2, -6, -10, -13, -14, -15, -14, -12, -9, -5, -1, 4, 7, 11, 13, 15, 6, -6, -15, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5,
 2, -2, -5, -8, -10, -11, -11, -11, -9, -7, -4, 0, 3, 6, 8, 10, 11, -11, -11, 11, -6, -6, -5, -3, -2, 0, 2, 4, 5, 6, 6, 6, 5, 4, 3, 1, -1, -3, -4, -5, -6, -6, -6, -5, -4,
 -2, 0, 2, 3, 5, 6, 6, -15, 15, -6, 0,0,0,1,1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 13, 10, 5,
 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 0, -13, -13, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6,
 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, -13, -5, 15, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -13,
 15, -10, -2, -4, -6, -8, -10, -11, -13, -13, -14, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -3, -1, 1, 3, 5, 7, 9, 10, 12, 13, 14, 14, 15, 13, 10, 5, -2, -4, -5,
 -7, -9, -10, -11, -12, -12, -13, -13, -13, -12, -12, -11, -10, -8, -7, -5, -3, -1, 1, 2, 4, 6, 8, 9, 10, 11, 12, 13, 13, 0, -13, -13, -1, -3, -4, -5, -6, -7, -8, -9, -9,
 -10, -10, -9, -9, -9, -8, -7, -6, -5, -4, -2, -1, 0, 2, 3, 5, 6, 7, 8, 8, 9, 9, 10, -13, -5, 15, -1, -1, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3,
 -2, -1, -1, 0, 1, 2, 2, 3, 4, 4, 4, 5, 5, 5, -13, 15, -10, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3,
 7, 10, 12, 14, 15, 13, 10, 5, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3, 0, 3, 6, 9, 11, 12, 13, 0, -13, -13, 2,
 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, -13, -5, 15, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1, 0, -1, -2,
 -3, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, -13, 15, -10, -5, -9, -13, -14, -15, -13, -10, -6, -1, 3, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11, -13,
 -15, -14, -12, -9, -4, 1, 6, 10, 13, 14, 13, 9, 5, -4, -8, -11, -13, -13, -12, -9, -5, -1, 3, 7, 10, 12, 13, 12, 10, 7, 2, -2, -6, -10, -12, -13, -12, -11, -8, -4, 1, 5,
 9, 11, 13, 0, -13, -13, -3, -6, -8, -9, -10, -9, -7, -4, -1, 2, 5, 8, 9, 10, 9, 7, 5, 2, -1, -5, -7, -9, -10, -9, -8, -6, -3, 0, 4, 6, 8, 9, -13, -5, 14, -2, -3, -4, -5,
 -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -4, -3, -1, 0, 2, 3, 4, 5, -13, 14, -9, 1,15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12,
 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, 1, 0, 0, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 2,
 2, 1, 1, 0, -1, -1, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, 0, 1, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4,
 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, -1, 1, 0, -15, -14, -14, -13, -12, -10, -9, -7, -5, -3, -1, 1, 3, 6, 8, 9, 11, 12, 13, 14, 14, 15, 15, 14, 13,
 13, 11, 10, 8, 6, 4, 2, 2, 1, 1, -13, -13, -12, -11, -10, -9, -8, -6, -4, -2, -1, 1, 3, 5, 7, 8, 10, 11, 12, 12, 13, 13, 13, 12, 12, 11, 10, 9, 7, 5, 4, 2, 0, -2, -2,
 -10, -9, -9, -8, -8, -7, -6, -5, -3, -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 4, 3, 1, -2, -1, 2, -5, -5, -5, -4, -4, -4, -3, -2, -2, -1, 0, 1, 1,
 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 1, -2, 2, -1, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14,
 15, 15, 14, 12, 10, 7, 3, 3, 2, 1, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3, 0, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 0, -3, -3,
 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, -3, -1, 3, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5,
 -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1, -3, 3, -2, -14, -13, -10, -6, -1, 4, 9, 12, 14, 15, 13, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -3,
 1, 6, 10, 13, 15, 14, 13, 9, 5, 4, 3, 2, -13, -11, -9, -5, -1, 4, 8, 11, 12, 13, 12, 10, 6, 2, -2, -7, -10, -12, -13, -12, -10, -7, -3, 1, 5, 9, 12, 13, 13, 11, 8, 4, 0,
 -4, -4, -9, -8, -6, -4, 0, 3, 6, 8, 9, 10, 9, 7, 5, 1, -2, -5, -7, -9, -10, -9, -8, -5, -2, 1, 4, 7, 9, 10, 9, 8, 6, 3, -4, -2, 5, -5, -4, -3, -2, 0, 1, 3, 4, 5, 5, 5,
 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2, 4, 5, 5, 5, 4, 3, 2, -4, 5, -3, 1,4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 7, 9, 11, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 0, -9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -4, -9, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, -10, 9, -4, -5, -5, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 10, 13, 15, -13,
 -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 13, 0, -13, -15, -15, -14, -14, -13, -13, -12,
 -11, -10, -9, -8, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, -5, -13, 10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -5, -4, -3, -2, -1,
 0, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, -15, 13, -5, 5, 5, 5, 4, 3, 2, 1, 1, -1, -1, -2, -3, -4, -5, -5, -5, -5, -5, -5, -4, -3, -2, -1, -1, 1, 1, 2, 3, 4,
 5, 5, 5, 10, 13, 15, 13, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -13, -13, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 13, 13, 0, -13, 15, 14,
 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14, -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, -5, -13, 10, 10, 9, 9, 8, 6, 5, 3, 1, -1, -3, -5,
 -6, -8, -9, -9, -10, -10, -9, -9, -8, -6, -5, -3, -1, 1, 3, 5, 6, 8, 9, 9, 10, -15, 13, -5, -5, -5, -4, -3, -1, 0, 2, 3, 4, 5, 5, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -5,
 -5, -4, -3, -2, 0, 1, 3, 4, 5, 5, 10, 13, 15, -13, -12, -10, -7, -3, 1, 4, 8, 11, 12, 13, 13, 11, 9, 5, 2, -2, -5, -9, -11, -13, -13, -12, -11, -8, -4, -1, 3, 7, 10, 12,
 13, 13, 0, -13, -15, -13, -11, -8, -4, 1, 5, 9, 12, 14, 15, 14, 13, 10, 6, 2, -2, -6, -10, -13, -14, -15, -14, -12, -9, -5, -1, 4, 8, 11, 13, 15, -5, -13, 10, -10, -9,
 -7, -5, -2, 0, 3, 6, 8, 9, 10, 9, 8, 7, 4, 1, -1, -4, -7, -8, -9, -10, -9, -8, -6, -3, 0, 2, 5, 7, 9, 10, -15, 13, -5, 1,0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4,
 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 10, 13, 15, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13,
 13, 13, 0, -13, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, -5, -13, 10, 0, 1, 1, 2, 2, 3, 3, 4, 4,
 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, -15, 13, -5, -1, -1, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -1, -1,
 0, 1, 2, 2, 3, 4, 4, 4, 5, 5, 5, 10, 13, 15, -2, -4, -5, -7, -9, -10, -11, -12, -12, -13, -13, -13, -12, -12, -11, -10, -8, -7, -5, -3, -1, 1, 2, 4, 6, 8, 9, 10, 11, 12,
 13, 13, 13, 0, -13, -2, -4, -6, -8, -10, -11, -13, -13, -14, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -3, -1, 1, 3, 5, 7, 9, 10, 12, 13, 14, 14, 15, -5, -13, 10,
 -1, -3, -4, -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -9, -8, -7, -6, -5, -4, -2, -1, 0, 2, 3, 5, 6, 7, 8, 8, 9, 9, 10, -15, 13, -5, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1,
 0, -1, -2, -3, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 10, 13, 15, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12,
 -11, -9, -6, -3, 0, 3, 6, 9, 11, 12, 13, 13, 0, -13, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10,
 12, 14, 15, -5, -13, 10, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, -15, 13, -5, -2, -3, -4, -5,
 -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -4, -3, -1, 0, 2, 3, 4, 5, 9, 13, 14, -4, -8, -11, -13, -13, -12, -9, -5, -1, 3, 7, 10, 12, 13,
 12, 10, 7, 2, -2, -6, -10, -12, -13, -12, -11, -8, -4, 1, 5, 9, 11, 13, 13, 0, -13, -5, -9, -13, -14, -15, -13, -10, -6, -1, 3, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11,
 -13, -15, -14, -12, -9, -4, 1, 6, 10, 13, 14, -5, -13, 9, -3, -6, -8, -9, -10, -9, -7, -4, -1, 2, 5, 8, 9, 10, 9, 7, 5, 2, -1, -5, -7, -9, -10, -9, -8, -6, -3, 0, 4, 6,
 8, 9, -14, 13, -5, 1,5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 1, 1, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11,
 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 0, -1, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6,
 6, 5, 4, 3, 3, 2, 1, 1, 0, -1, 0, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, 1, 0, -5, -5, -5, -4, -4, -4,
 -3, -2, -2, -1, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 1, 1, 2, 2, -13, -13, -12, -11, -10, -9, -8, -6, -4, -2, -1, 1, 3, 5, 7, 8, 10, 11, 12,
 12, 13, 13, 13, 12, 12, 11, 10, 9, 7, 5, 4, 2, 2, 0, -2, -15, -14, -14, -13, -12, -10, -9, -7, -5, -3, -1, 1, 3, 6, 8, 9, 11, 12, 13, 14, 14, 15, 15, 14, 13, 13, 11, 10,
 8, 6, 4, 2, -1, -2, 1, -10, -9, -9, -8, -8, -7, -6, -5, -3, -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 4, 3, 1, -2, 2, -1, 5, 5, 4, 3, 2, 1, 0, -1,
 -2, -3, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1, 2, 3, 3, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3,
 0, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 3, 0, -3, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14,
 12, 10, 7, 3, -1, -3, 2, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, -3, 3, -1, -5, -4, -3, -2, 0,
 1, 3, 4, 5, 5, 5, 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2, 4, 5, 5, 5, 4, 3, 2, 3, 4, 5, -13, -11, -9, -5, -1, 4, 8, 11, 12, 13, 12, 10, 6, 2, -2, -7, -10,
 -12, -13, -12, -10, -7, -3, 1, 5, 9, 12, 13, 13, 11, 8, 4, 4, 0, -4, -14, -13, -10, -6, -1, 4, 9, 12, 14, 15, 13, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -3, 1,
 6, 10, 13, 15, 14, 13, 9, 5, -2, -4, 3, -9, -8, -6, -4, 0, 3, 6, 8, 9, 10, 9, 7, 5, 1, -2, -5, -7, -9, -10, -9, -8, -5, -2, 1, 4, 7, 9, 10, 9, 8, 6, 3, -5, 4, -2, 1,8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, -8, -8, 8, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, -10, 10, -4, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0,
 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13,
 -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 13,
 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 6, -6, -15, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3,
 -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, -11, -11, 11, -6, -6, -6, -6, -6, -6,
 -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6,
 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -15, 15, -6, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11,
 -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 15, 15, 14, 14, 13, 13, 12, 11,
 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -7, -6, -5, -4, -2,
 -1, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 6, -6, -15, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9,
 -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, -11, -11, 11, 6,
 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1,
 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, -15, 15, -6, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11, 11, 11, 11, 10, 9, 8, 7,
 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11, 11, 11, 11, -15, -14, -14,
 -13, -12, -10, -8, -7, -5, -2, 0, 2, 4, 6, 8, 10, 11, 12, 13, 14, 15, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, -15, -14,
 -13, -12, -11, -10, -8, -6, -4, -2, 0, 2, 5, 7, 8, 10, 12, 13, 14, 14, 15, 6, -6, -15, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11,
 11, 11, 11, 10, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11,
 -11, -11, 11, -6, -6, -6, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 6, 6, 6, 6, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -5, -6, -6, -6, -6, -6,
 -6, -6, -5, -5, -4, -3, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 6, 6, 6, -15, 15, -6, 0,0,0,0,1,15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13,
 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5,
 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6,
 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -15, -15, -15, -14, -14, -14, -13, -13,
 -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12,
 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 1, 0, -13, -13, -13, -13, -12, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7,
 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -1, -10, -10, -10, -9, -9, -9, -9, -8,
 -8, -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5,
 4, 3, 3, 2, 1, 1, -1, 0, 1, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, -1, 1, -1, 15, 15, 14, 13, 13, 12, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13,
 -14, -14, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -7, -6, -4, -2, 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2,
 2, 1, 1, 13, 13, 12, 12, 11, 10, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -12, -13, -13, -13, -13, -13, -12, -12, -11, -10, -9, -8, -6, -5, -3, -2, 0,
 1, 3, 4, 6, 7, 9, 10, 11, 11, 12, 13, 13, 13, 13, 13, 12, 12, 11, 10, 9, 8, 6, 5, 3, 2, 0, -2, -2, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 0, -1, -2, -3, -4, -5, -6, -7,
 -8, -8, -9, -9, -10, -10, -10, -9, -9, -9, -8, -7, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, 10, 9, 9, 9, 8, 7, 6, 6, 5, 3, 2, 1, -2, -1, 2, 5,
 5, 5, 5, 4, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5,
 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, -2, 2, -1, -15, -14, -13, -12, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11, 10, 8, 5, 3, 0, -2,
 -5, -7, -9, -11, -12, -14, -14, -15, -15, -14, -13, -12, -10, -9, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15, 14, 14, 13, 11, 9, 7, 5, 2, 2, 2, 1, -13, -13, -12,
 -11, -9, -8, -6, -4, -2, 1, 3, 5, 7, 9, 10, 11, 12, 13, 13, 13, 12, 11, 10, 9, 7, 5, 2, 0, -2, -4, -6, -8, -10, -11, -12, -13, -13, -13, -13, -12, -11, -9, -8, -5, -3,
 -1, 1, 3, 5, 7, 9, 10, 12, 12, 13, 13, 13, 12, 11, 10, 8, 6, 4, 2, 0, -2, -2, -10, -9, -9, -8, -7, -6, -4, -3, -1, 0, 2, 4, 5, 6, 8, 8, 9, 10, 10, 10, 9, 8, 7, 6, 5, 3,
 2, 0, -1, -3, -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -4, -3, -1, 1, 2, 4, 5, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 3, 2, -2, -1, 2, -5, -5, -5, -4, -4,
 -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 4, 4, 5, 5,
 5, 5, 5, 5, 4, 4, 3, 2, 2, 1, -2, 2, -2, 0,0,1,5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11,
 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15,
 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4,
 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5,
 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1,
 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, -13, -13, -13, -13, -12, -12, -12, -11,
 -11, -10, -9, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10,
 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 1, 0, -1, -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, 1, -10, -10, -10, -9, -9, -9, -9, -8,
 -8, -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5,
 4, 3, 3, 2, 1, 1, -1, 1, 0, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -2, -2, -1, -1, 0,
 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 2, 2, 13, 13, 12, 12, 11, 10, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11,
 -12, -13, -13, -13, -13, -13, -12, -12, -11, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 4, 6, 7, 9, 10, 11, 11, 12, 13, 13, 13, 13, 13, 12, 12, 11, 10, 9, 8, 6, 5, 3, 2, 2,
 0, -2, 15, 15, 14, 13, 13, 12, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -7, -6, -4, -2, 0,
 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, -1, -2, 1, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 0, -1, -2, -3, -4, -5, -6, -7,
 -8, -8, -9, -9, -10, -10, -10, -9, -9, -9, -8, -7, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, 10, 9, 9, 9, 8, 7, 6, 6, 5, 3, 2, 1, -2, 2, -1,
 -5, -5, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -1, -1, 0,
 1, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 2, 1, 2, 2, 2, -13, -13, -12, -11, -9, -8, -6, -4, -2, 1, 3, 5, 7, 9, 10, 11, 12, 13, 13, 13, 12, 11, 10, 9, 7, 5, 2, 0, -2,
 -4, -6, -8, -10, -11, -12, -13, -13, -13, -13, -12, -11, -9, -8, -5, -3, -1, 1, 3, 5, 7, 9, 10, 12, 12, 13, 13, 13, 12, 11, 10, 8, 6, 4, 2, 2, 0, -2, -15, -14, -13, -12,
 -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11, 10, 8, 5, 3, 0, -2, -5, -7, -9, -11, -12, -14, -14, -15, -15, -14, -13, -12, -10, -9, -6, -4,
 -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15, 14, 14, 13, 11, 9, 7, 5, 2, -1, -2, 2, -10, -9, -9, -8, -7, -6, -4, -3, -1, 0, 2, 4, 5, 6, 8, 8, 9, 10, 10, 10, 9, 8, 7, 6, 5, 3,
 2, 0, -1, -3, -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -4, -3, -1, 1, 2, 4, 5, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 3, 2, -2, 2, -1, 1,8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, -8, -8, 8, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -10, 10,
 -4, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -7,
 -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5,
 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -12, -11, -11, -11, -11, -10,
 -10, -10, -9, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5,
 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 6, -6, -15, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8,
 -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3,
 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, -11, -11, 11, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4,
 -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3,
 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -15, 15, -6, 11, 11, 11, 11, 11, 11,
 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9,
 -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3,
 -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 15, 15, 15, 14, 14, 14,
 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12,
 -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8,
 -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 6, -6, -15,
 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7,
 -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6,
 -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, -11, -11, 11,
 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5,
 -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1,
 -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -15, 15, -6, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9,
 -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6,
 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8,
 -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, -15, -15, -14, -14, -14, -13, -13, -12, -12,
 -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 6, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11,
 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 0, 0, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13,
 -12, -12, -11, -10, -9, -8, -7, -6, -6, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 6, -6, -15, -11, -11, -11, -11,
 -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10,
 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10,
 -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, -11, -11, 11, -6, -6, -6, -6, -6,
 -6, -6, -5, -5, -5, -5, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4,
 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -3,
 -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, -15, 15, -6, 0,0,0,0,1,15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4,
 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1,
 1, 0, 0, 0, 0, 0, 0, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15, -15, -15, -15,
 -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1,
 -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, 0, 0, -13, -13, -13,
 -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -3, -2, -2, -1,
 -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, -10, -10, -10, -10, -10,
 -10, -9, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2,
 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8,
 8, 8, 8, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -4, -4,
 -4, -4, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11,
 -12, -12, -13, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -13, -12, -11, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, -1,
 0, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5,
 4, 4, 3, 2, 1, 1, 1, 0, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 5, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9,
 -10, -10, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2,
 -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 5, 5,
 4, 3, 2, 2, 1, 0, -1, -1, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8,
 -8, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4,
 5, 5, 6, 6, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 6, 5, 5, 4, 3, 3, 2, 2, 1, 1, -1, 0, 1, 5, 5, 5, 5, 5, 5, 5, 5,
 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5,
 -5, -5, -5, -5, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, -1, 1, -1, -15, -15, -14, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0,
 1, 3, 4, 5, 6, 7, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11,
 -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -10, -9, -8, -6, -5, -4, -3, -2, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13,
 14, 14, 14, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 1, 0, -13, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -3, -2,
 -1, 0, 1, 2, 3, 4, 5, 7, 8, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -9,
 -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11,
 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -1, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2,
 -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -6, -7, -8, -8,
 -8, -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10,
 10, 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, -1, 0, 1, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5,
 -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, -1, 1, -1,
 0,0,1,5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0,
 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8,
 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -2,
 -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -13, -13, -13, -13, -13,
 -13, -13, -13, -12, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0,
 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, -15, -15, -15, -15, -15, -15, -14,
 -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1,
 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14,
 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, 0, 0, -10, -10, -10, -10, -10, -10, -9,
 -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3,
 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8,
 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1,
 -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2,
 -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1,
 1, 0, 1, 1, 1, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 5, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10,
 -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0,
 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 5, 5, 4, 3, 2,
 2, 1, 1, 0, -1, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11,
 -12, -12, -13, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -13, -12, -11, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, -1,
 0, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5,
 4, 4, 3, 2, 1, 0, -1, 1, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8,
 -8, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4,
 5, 5, 6, 6, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 6, 5, 5, 4, 3, 3, 2, 2, 1, 1, -1, 1, 0, -5, -5, -5, -5, -5, -5,
 -4, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1,
 -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4,
 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 1, 1, 1, -13, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0,
 1, 2, 3, 4, 5, 7, 8, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10,
 -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11, 12, 12, 12,
 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 0, -1, -15, -15, -14, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0,
 1, 3, 4, 5, 6, 7, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11,
 -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -10, -9, -8, -6, -5, -4, -3, -2, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13,
 14, 14, 14, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, 1, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1,
 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -6, -7, -8, -8, -8,
 -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10,
 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, -1, 1, 0, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, 10, 10, 10, 10, 5, -4, -10, -10, -4,
 5, 10, 9, 9, 9, 9, -2, -11, -6, 6, 11, 2, -9, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, -14, -6, 6, 14, 12, 8, 3, -3, -8, -12, -14, -13, -6, 6, 13, 6, -6, -13, -13,
 -6, 6, 13, -12, -5, 5, 12, -3, -14, -8, 8, 14, 3, -12, 8, -8, -8, 8, 8, 8, 8, 8, 8, 8, 8, 11, -11, -11, 11, 9, 6, 2, -2, -6, -9, -11, 10, -10, -10, 10, 5, -4, -10, -10,
 -4, 5, 10, 9, -9, -9, 9, -2, -11, -6, 6, 11, 2, -9, -4, 10, -10, 5, 5, 5, 5, 5, 5, 5, 5, -6, 14, -14, 6, 5, 4, 1, -1, -4, -5, -6, -6, 13, -13, 6, 3, -3, -6, -6, -3, 3,
 6, -5, 12, -12, 5, -1, -6, -4, 4, 6, 1, -5, 0,0,0,1,5, 9, 12, 14, 14, 13, 12, 10, 8, 5, 3, 5, 9, 12, 14, 10, 3, -5, -12, -14, -13, -8, 4, 8, 11, 13, 3, -10, -14, -8, 5,
 14, 12, 4, 8, 10, 12, -5, -14, -3, 13, 10, -8, -14, -12, -12, 0, 12, 12, 11, 10, 9, 7, 5, 2, -12, -12, 0, 12, 9, 2, -5, -10, -12, -11, -7, -11, -11, 0, 11, 2, -9, -12,
 -7, 5, 12, 10, -10, -10, 0, 10, -5, -12, -2, 11, 9, -7, -12, 14, -5, -12, 9, 9, 8, 8, 6, 5, 3, 2, 14, -5, -12, 9, 6, 2, -3, -8, -9, -8, -5, 13, -4, -11, 8, 2, -6, -9,
 -5, 3, 9, 8, 12, -4, -10, 8, -3, -9, -2, 8, 6, -5, -9, -9, 14, -12, 5, 5, 4, 4, 3, 3, 2, 1, -9, 14, -12, 5, 3, 1, -2, -4, -5, -4, -3, -8, 13, -11, 4, 1, -3, -5, -3, 2,
 5, 4, -8, 12, -10, 4, -2, -5, -1, 4, 3, -3, -5, 1,14, 12, 9, 5, 5, 4, 4, 3, 3, 2, 1, 14, 12, 9, 5, 3, 1, -2, -4, -5, -4, -3, 13, 11, 8, 4, 1, -3, -5, -3, 2, 5, 4, 12,
 10, 8, 4, -2, -5, -1, 4, 3, -3, -5, -12, 0, 12, 12, 12, 11, 10, 9, 7, 5, 2, -12, 0, 12, 12, 9, 2, -5, -10, -12, -11, -7, -11, 0, 11, 11, 2, -9, -12, -7, 5, 12, 10, -10,
 0, 10, 10, -5, -12, -2, 11, 9, -7, -12, 9, -12, -5, 14, 14, 13, 12, 10, 8, 5, 3, 9, -12, -5, 14, 10, 3, -5, -12, -14, -13, -8, 8, -11, -4, 13, 3, -10, -14, -8, 5, 14,
 12, 8, -10, -4, 12, -5, -14, -3, 13, 10, -8, -14, -5, 12, -14, 9, 9, 8, 8, 6, 5, 3, 2, -5, 12, -14, 9, 6, 2, -3, -8, -9, -8, -5, -4, 11, -13, 8, 2, -6, -9, -5, 3, 9, 8,
 -4, 10, -12, 8, -3, -9, -2, 8, 6, -5, -9, 0,1,1, 2, 2, 3, 5, 8, 10, 12, 13, 14, 14, 3, 5, 7, 8, 13, 14, 12, 5, -3, -10, -14, 4, 8, 10, 12, 14, 5, -8, -14, -10, 3, 13, 5,
 9, 12, 14, 8, -10, -13, 3, 14, 5, -12, -2, -2, 0, 2, 5, 7, 9, 10, 11, 12, 12, -7, -7, 0, 7, 11, 12, 10, 5, -2, -9, -12, -10, -10, 0, 10, 12, 5, -7, -12, -9, 2, 11, -12,
 -12, 0, 12, 7, -9, -11, 2, 12, 5, -10, 3, -1, -2, 2, 3, 5, 6, 8, 8, 9, 9, 8, -3, -7, 5, 8, 9, 8, 3, -2, -6, -9, 12, -4, -10, 8, 9, 3, -5, -9, -6, 2, 8, 14, -5, -12, 9,
 5, -6, -8, 2, 9, 3, -8, -2, 3, -2, 1, 2, 3, 3, 4, 4, 5, 5, -5, 8, -7, 3, 4, 5, 4, 2, -1, -3, -5, -8, 12, -10, 4, 5, 2, -3, -5, -3, 1, 4, -9, 14, -12, 5, 3, -3, -4, 1, 5,
 2, -4, 1,3, 2, 2, 1, 2, 3, 3, 4, 4, 5, 5, 8, 7, 5, 3, 4, 5, 4, 2, -1, -3, -5, 12, 10, 8, 4, 5, 2, -3, -5, -3, 1, 4, 14, 12, 9, 5, 3, -3, -4, 1, 5, 2, -4, -2, 0, 2, 2, 5,
 7, 9, 10, 11, 12, 12, -7, 0, 7, 7, 11, 12, 10, 5, -2, -9, -12, -10, 0, 10, 10, 12, 5, -7, -12, -9, 2, 11, -12, 0, 12, 12, 7, -9, -11, 2, 12, 5, -10, 2, -2, -1, 3, 5, 8,
 10, 12, 13, 14, 14, 5, -7, -3, 8, 13, 14, 12, 5, -3, -10, -14, 8, -10, -4, 12, 14, 5, -8, -14, -10, 3, 13, 9, -12, -5, 14, 8, -10, -13, 3, 14, 5, -12, -1, 2, -3, 2, 3,
 5, 6, 8, 8, 9, 9, -3, 7, -8, 5, 8, 9, 8, 3, -2, -6, -9, -4, 10, -12, 8, 9, 3, -5, -9, -6, 2, 8, -5, 12, -14, 9, 5, -6, -8, 2, 9, 3, -8, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, -10, -4, 5, 10, 9, 9, 9, 9, 9, 9, 9, 9, -2, -11, -6,
 6, 11, 2, -9, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, -15, -13, -9, -3, 3, 9, 13, 15, 13, 9, 3, -3, -9, -13, -15, -14, -12, -8, -3, 3, 8, 12, 14, 6,
 -6, -14, -14, -6, 6, 14, -13, -11, -7, -3, 3, 7, 11, 13, -3, -15, -9, 9, 15, 3, -13, 10, 5, -4, -10, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 14, 6, -6, -14, -14, -6,
 6, 14, 12, 8, 3, -3, -8, -12, -14, 13, 6, -6, -13, -13, -6, 6, 13, 6, -6, -13, -13, -6, 6, 13, 12, 5, -5, -12, -12, -5, 5, 12, -3, -14, -8, 8, 14, 3, -12, -9, 2, 11, 6,
 -6, -11, -2, 9, 9, 9, 9, 9, 9, 9, 9, -13, 3, 15, 9, -9, -15, -3, 13, 11, 7, 3, -3, -7, -11, -13, -12, 3, 14, 8, -8, -14, -3, 12, 5, -5, -12, -12, -5, 5, 12, -11, 3, 13,
 7, -7, -13, -3, 11, -3, -13, -7, 7, 13, 3, -11, 0,0,0,1,3, 5, 8, 10, 12, 13, 14, 14, 14, 13, 12, 10, 8, 5, 3, 3, 5, 8, 10, 12, 13, 14, 14, 10, 3, -5, -12, -14, -13, -8,
 3, 5, 7, 9, 11, 12, 13, 13, 3, -10, -14, -8, 5, 14, 12, 2, 4, 6, 8, 10, 11, 12, 12, -5, -14, -3, 13, 10, -8, -14, -8, -13, -14, -12, -5, 3, 10, 14, 14, 13, 12, 10, 8, 5,
 3, -8, -13, -14, -12, -5, 3, 10, 14, 10, 3, -5, -12, -14, -13, -8, -7, -12, -13, -11, -5, 3, 9, 13, 3, -10, -14, -8, 5, 14, 12, -6, -11, -12, -10, -4, 2, 8, 12, -5, -14,
 -3, 13, 10, -8, -14, 12, 14, 5, -8, -14, -10, 3, 13, 13, 12, 11, 9, 7, 5, 3, 12, 14, 5, -8, -14, -10, 3, 13, 9, 3, -5, -11, -13, -12, -7, 11, 13, 5, -7, -13, -9, 3, 12,
 3, -9, -13, -7, 5, 13, 11, 10, 12, 4, -6, -12, -8, 2, 11, -5, -13, -3, 12, 9, -7, -13, -14, -8, 10, 13, -3, -14, -5, 12, 12, 11, 10, 8, 6, 4, 2, -14, -8, 10, 13, -3,
 -14, -5, 12, 8, 2, -4, -10, -12, -11, -6, -13, -7, 9, 12, -3, -13, -5, 11, 2, -8, -12, -6, 4, 12, 10, -12, -6, 8, 11, -2, -12, -4, 10, -4, -12, -2, 11, 8, -6, -12, 1,14,
 14, 13, 12, 10, 8, 5, 3, 3, 3, 2, 2, 2, 1, 1, 14, 14, 13, 12, 10, 8, 5, 3, 2, 1, -1, -2, -3, -3, -2, 13, 13, 12, 11, 9, 7, 5, 3, 1, -2, -3, -2, 1, 3, 2, 12, 12, 11, 10,
 8, 6, 4, 2, -1, -3, -1, 3, 2, -2, -3, -14, -10, -3, 5, 12, 14, 13, 8, 8, 7, 6, 5, 4, 3, 2, -14, -10, -3, 5, 12, 14, 13, 8, 5, 2, -3, -6, -8, -7, -4, -13, -9, -3, 5, 11,
 13, 12, 7, 2, -5, -8, -4, 3, 8, 6, -12, -8, -2, 4, 10, 12, 11, 6, -3, -8, -2, 7, 5, -4, -8, 13, 3, -10, -14, -8, 5, 14, 12, 12, 11, 10, 8, 6, 4, 2, 13, 3, -10, -14, -8,
 5, 14, 12, 8, 2, -4, -10, -12, -11, -6, 12, 3, -9, -13, -7, 5, 13, 11, 2, -8, -12, -6, 4, 12, 10, 11, 2, -8, -12, -6, 4, 12, 10, -4, -12, -2, 11, 8, -6, -12, -12, 5, 14,
 3, -13, -10, 8, 14, 14, 13, 12, 10, 8, 5, 3, -12, 5, 14, 3, -13, -10, 8, 14, 10, 3, -5, -12, -14, -13, -8, -11, 5, 13, 3, -12, -9, 7, 13, 3, -10, -14, -8, 5, 14, 12,
 -10, 4, 12, 2, -11, -8, 6, 12, -5, -14, -3, 13, 10, -8, -14, 0,1,1, 1, 2, 2, 2, 3, 3, 3, 5, 8, 10, 12, 13, 14, 14, 2, 3, 4, 5, 6, 7, 8, 8, 13, 14, 12, 5, -3, -10, -14,
 2, 4, 6, 8, 10, 11, 12, 12, 14, 5, -8, -14, -10, 3, 13, 3, 5, 8, 10, 12, 13, 14, 14, 8, -10, -13, 3, 14, 5, -12, -2, -3, -3, -2, -1, 1, 2, 3, 5, 8, 10, 12, 13, 14, 14,
 -4, -7, -8, -6, -3, 2, 5, 8, 13, 14, 12, 5, -3, -10, -14, -6, -11, -12, -10, -4, 2, 8, 12, 14, 5, -8, -14, -10, 3, 13, -8, -13, -14, -12, -5, 3, 10, 14, 8, -10, -13, 3,
 14, 5, -12, 2, 3, 1, -2, -3, -2, 1, 3, 5, 7, 9, 11, 12, 13, 13, 6, 8, 3, -4, -8, -5, 2, 7, 12, 13, 11, 5, -3, -9, -13, 10, 12, 4, -6, -12, -8, 2, 11, 13, 5, -7, -13, -9,
 3, 12, 12, 14, 5, -8, -14, -10, 3, 13, 7, -9, -12, 3, 13, 5, -11, -3, -2, 2, 3, -1, -3, -1, 2, 4, 6, 8, 10, 11, 12, 12, -8, -4, 5, 7, -2, -8, -3, 6, 11, 12, 10, 4, -2,
 -8, -12, -12, -6, 8, 11, -2, -12, -4, 10, 12, 4, -6, -12, -8, 2, 11, -14, -8, 10, 13, -3, -14, -5, 12, 6, -8, -11, 2, 12, 4, -10, 1,3, 3, 3, 2, 2, 2, 1, 1, 1, 2, 2, 2,
 3, 3, 3, 8, 8, 7, 6, 5, 4, 3, 2, 3, 3, 2, 1, -1, -2, -3, 12, 12, 11, 10, 8, 6, 4, 2, 3, 1, -2, -3, -2, 1, 3, 14, 14, 13, 12, 10, 8, 5, 3, 2, -2, -3, 1, 3, 1, -2, -3, -2,
 -1, 1, 2, 3, 3, 2, 3, 4, 5, 6, 7, 8, 8, -8, -5, -2, 3, 6, 8, 7, 4, 7, 8, 6, 3, -2, -5, -8, -12, -8, -2, 4, 10, 12, 11, 6, 8, 3, -4, -8, -5, 2, 7, -14, -10, -3, 5, 12,
 14, 13, 8, 4, -5, -7, 2, 8, 3, -6, 3, 1, -2, -3, -2, 1, 3, 2, 4, 6, 8, 10, 11, 12, 12, 7, 2, -5, -8, -4, 3, 8, 6, 11, 12, 10, 4, -2, -8, -12, 11, 2, -8, -12, -6, 4, 12,
 10, 12, 4, -6, -12, -8, 2, 11, 13, 3, -10, -14, -8, 5, 14, 12, 6, -8, -11, 2, 12, 4, -10, -2, 1, 3, 1, -3, -2, 2, 3, 5, 8, 10, 12, 13, 14, 14, -6, 3, 8, 2, -7, -5, 4, 8,
 13, 14, 12, 5, -3, -10, -14, -10, 4, 12, 2, -11, -8, 6, 12, 14, 5, -8, -14, -10, 3, 13, -12, 5, 14, 3, -13, -10, 8, 14, 8, -10, -13, 3, 14, 5, -12, 1,8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, -10, -4, 5, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, -2, -11, -6, 6, 11, 2, -9, -11, -11, -10, -9, -7, -5, -3,
 -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, 13, 9, 3, -3, -9, -13, -16, -15, -14, -13,
 -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, 6, -6, -15, -15, -6, 6, 15, -13, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 13, -3, -16, -9, 9, 16, 3, -13,
 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 15, 13, 9, 3, -3, -9, -13, -15, -15, -13, -9, -3, 3, 9, 13, 15, 13, 9, 3, -3, -9,
 -13, -15, 14, 12, 8, 3, -3, -8, -12, -14, -14, -12, -8, -3, 3, 8, 12, 14, 6, -6, -14, -14, -6, 6, 14, 13, 11, 7, 3, -3, -7, -11, -13, -13, -11, -7, -3, 3, 7, 11, 13, -3,
 -15, -9, 9, 15, 3, -13, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, 11, 11, 11, 11, -15, -10, -2, 7, 14, 16, 12, 4, -4, -12, -16, -14, -7,
 2, 10, 15, 13, 8, 3, -3, -8, -13, -15, -14, -9, -1, 7, 13, 15, 11, 4, -4, -11, -15, -13, -7, 1, 9, 14, 6, -6, -14, -14, -6, 6, 14, -13, -8, -1, 6, 12, 13, 10, 4, -4,
 -10, -13, -12, -6, 1, 8, 13, -3, -15, -8, 8, 15, 3, -13, 0,0,0,1,1, 3, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 13, 12, 10, 8, 6, 3, 1, 3, 4, 5, 7, 8, 9,
 10, 11, 12, 13, 13, 14, 14, 15, 15, 10, 3, -5, -12, -15, -13, -8, 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 13, 13, 3, -10, -15, -8, 6, 15, 12, 1, 2, 3, 5, 6, 7,
 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, -5, -15, -3, 13, 10, -8, -15, -4, -8, -11, -14, -15, -15, -14, -11, -8, -4, 0, 4, 8, 11, 14, 15, 15, 13, 12, 10, 8, 6, 3, -4, -8,
 -11, -13, -15, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13, 15, 10, 3, -5, -12, -15, -13, -8, -4, -7, -10, -12, -13, -13, -12, -10, -7, -4, 0, 4, 7, 10, 12, 13, 3, -10, -15,
 -8, 6, 15, 12, -3, -7, -9, -11, -12, -12, -11, -9, -7, -3, 0, 3, 7, 9, 11, 12, -5, -15, -3, 13, 10, -8, -15, 7, 12, 15, 14, 10, 4, -3, -9, -14, -15, -13, -8, -1, 6, 11,
 15, 14, 13, 12, 10, 8, 5, 3, 7, 12, 15, 14, 10, 4, -3, -9, -13, -15, -13, -8, -1, 5, 11, 14, 10, 3, -5, -12, -15, -13, -8, 6, 11, 13, 13, 9, 4, -3, -8, -12, -13, -12,
 -7, -1, 5, 10, 13, 3, -10, -15, -8, 5, 14, 12, 6, 10, 12, 12, 9, 3, -2, -8, -11, -12, -11, -7, -1, 5, 9, 12, -5, -15, -3, 13, 10, -8, -14, -9, -15, -14, -7, 3, 11, 15,
 12, 4, -6, -13, -15, -10, -1, 8, 14, 14, 13, 12, 10, 8, 5, 3, -9, -14, -13, -7, 3, 11, 15, 12, 4, -5, -13, -15, -10, -1, 8, 14, 10, 3, -5, -12, -14, -13, -8, -8, -13,
 -12, -6, 3, 10, 13, 11, 4, -5, -12, -13, -9, -1, 7, 13, 3, -10, -14, -8, 5, 14, 12, -8, -12, -11, -6, 2, 9, 12, 10, 3, -5, -11, -12, -9, -1, 7, 12, -5, -14, -3, 13, 10,
 -8, -14, 1,15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 0, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 0, 0, -1, -1, -1,
 -1, 13, 13, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 1, 0, -1, -1, -1, 1, 1, 1, 12, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 3, 2, 1, 0, -1, 0, 1, 1, -1, -1, -15,
 -14, -11, -8, -4, 0, 4, 8, 11, 14, 15, 15, 14, 11, 8, 4, 4, 4, 3, 3, 2, 2, 1, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13, 15, 15, 13, 11, 8, 4, 3, 1, -2, -3, -4, -4, -2,
 -13, -12, -10, -7, -4, 0, 4, 7, 10, 12, 13, 13, 12, 10, 7, 4, 1, -3, -4, -2, 2, 4, 3, -12, -11, -9, -7, -3, 0, 3, 7, 9, 11, 12, 12, 11, 9, 7, 3, -2, -4, -1, 4, 3, -2,
 -4, 15, 11, 6, -1, -8, -13, -15, -14, -9, -3, 4, 10, 14, 15, 12, 7, 7, 6, 6, 5, 4, 3, 1, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4, 10, 14, 15, 12, 7, 5, 1, -2, -6,
 -7, -6, -4, 13, 10, 5, -1, -7, -12, -13, -12, -8, -3, 4, 9, 13, 13, 11, 6, 1, -5, -7, -4, 3, 7, 6, 12, 9, 5, -1, -7, -11, -12, -11, -8, -2, 3, 9, 12, 12, 10, 6, -2, -7,
 -1, 6, 5, -4, -7, -14, -8, 1, 10, 15, 13, 6, -4, -12, -15, -11, -3, 7, 14, 15, 9, 9, 8, 8, 6, 5, 3, 2, -14, -8, 1, 10, 15, 13, 5, -4, -12, -15, -11, -3, 7, 13, 14, 9, 6,
 2, -3, -8, -9, -8, -5, -13, -7, 1, 9, 13, 12, 5, -4, -11, -13, -10, -3, 6, 12, 13, 8, 2, -6, -9, -5, 3, 9, 8, -12, -7, 1, 9, 12, 11, 5, -3, -10, -12, -9, -2, 6, 11, 12,
 8, -3, -9, -2, 8, 6, -5, -9, 0,1,0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 6, 8, 10, 12, 13, 15, 15, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 13, 15, 12, 6,
 -3, -10, -15, 1, 2, 3, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, 15, 6, -8, -15, -10, 3, 13, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 8, -10, -13, 3,
 15, 6, -12, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 6, 8, 10, 12, 13, 15, 15, -2, -4, -6, -7, -8, -8, -7, -6, -4, -2, 0, 2, 4, 6, 7, 8, 13, 15, 12, 6,
 -3, -10, -15, -3, -7, -9, -11, -12, -12, -11, -9, -7, -3, 0, 3, 7, 9, 11, 12, 15, 6, -8, -15, -10, 3, 13, -4, -8, -11, -13, -15, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13,
 15, 8, -10, -13, 3, 15, 6, -12, 1, 2, 3, 3, 2, 1, -1, -2, -3, -3, -3, -2, 0, 1, 2, 3, 5, 8, 10, 12, 13, 14, 15, 4, 7, 8, 8, 6, 2, -2, -5, -7, -8, -7, -4, -1, 3, 6, 8,
 13, 15, 12, 5, -3, -10, -14, 6, 10, 12, 12, 9, 3, -2, -8, -11, -12, -11, -7, -1, 5, 9, 12, 14, 5, -8, -15, -10, 3, 13, 7, 12, 15, 14, 10, 4, -3, -9, -13, -15, -13, -8,
 -1, 5, 11, 14, 8, -10, -13, 3, 15, 5, -12, -2, -3, -3, -1, 1, 2, 3, 2, 1, -1, -3, -3, -2, 0, 2, 3, 5, 8, 10, 12, 13, 14, 14, -5, -8, -7, -4, 2, 6, 8, 7, 2, -3, -7, -8,
 -6, -1, 4, 8, 13, 14, 12, 5, -3, -10, -14, -8, -12, -11, -6, 2, 9, 12, 10, 3, -5, -11, -12, -9, -1, 7, 12, 14, 5, -8, -14, -10, 3, 13, -9, -14, -13, -7, 3, 11, 15, 12,
 4, -5, -13, -15, -10, -1, 8, 14, 8, -10, -13, 3, 14, 5, -12, 1,3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 4, 4,
 3, 2, 2, 1, 1, 1, 1, 1, 0, -1, -1, 12, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 3, 2, 1, 1, 1, -1, -1, -1, 0, 1, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3,
 1, 1, -1, -1, 0, 1, 1, -1, -3, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 3, 3, 2, 2, 1, 2, 2, 3, 3, 4, 4, 4, -8, -7, -6, -4, -2, 0, 2, 4, 6, 7, 8, 8, 7, 6, 4, 2, 4, 4, 3, 2, -1,
 -3, -4, -12, -11, -9, -7, -3, 0, 3, 7, 9, 11, 12, 12, 11, 9, 7, 3, 4, 2, -2, -4, -3, 1, 4, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13, 15, 15, 13, 11, 8, 4, 2, -3, -4, 1, 4,
 2, -3, 3, 2, 1, 0, -2, -3, -3, -3, -2, -1, 1, 2, 3, 3, 2, 1, 3, 4, 5, 6, 6, 7, 7, 8, 6, 3, -1, -4, -7, -8, -7, -5, -2, 2, 6, 8, 8, 7, 4, 6, 7, 6, 3, -1, -5, -7, 12, 9,
 5, -1, -7, -11, -12, -11, -8, -2, 3, 9, 12, 12, 10, 6, 7, 3, -4, -7, -5, 1, 6, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4, 10, 14, 15, 12, 7, 4, -5, -6, 1, 7, 3, -6,
 -3, -2, 0, 2, 3, 3, 1, -1, -2, -3, -2, -1, 1, 3, 3, 2, 3, 5, 6, 8, 8, 9, 9, -8, -4, 1, 6, 8, 7, 3, -2, -7, -8, -6, -2, 4, 7, 8, 5, 8, 9, 8, 3, -2, -6, -9, -12, -7, 1, 9,
 12, 11, 5, -3, -10, -12, -9, -2, 6, 11, 12, 8, 9, 3, -5, -9, -6, 2, 8, -14, -8, 1, 10, 15, 13, 5, -4, -12, -15, -11, -3, 7, 13, 14, 9, 5, -6, -8, 2, 9, 3, -8, 1,8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, -10, -4, 5, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, -2, -11, -6, 6, 11, 2, -9, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 13, 9, 3, -3, -9, -13,
 -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 6, -6, -15, -15, -6, 6, 15, -13,
 -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, -3, -16, -9, 9, 16, 3, -13, 11, 11, 10, 9, 7,
 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 15, 14, 12, 10, 7, 4, 2, -2,
 -4, -7, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, 13, 9, 3, -3, -9, -13, -16, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7,
 -9, -11, -13, -14, -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, 6, -6, -15, -15, -6, 6, 15, 13, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -12,
 -13, -13, -13, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 13, -3, -16, -9, 9, 16, 3, -13, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5,
 -8, -10, -11, -11, -11, -9, -7, -4, 0, 3, 6, 8, 10, 11, 11, 11, 11, 11, 11, 11, 11, -16, -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15,
 -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 13, 9, 3, -3, -9, -13, -16, -15, -13, -11, -7, -4, 1, 5, 9, 12, 14, 15, 14, 13, 10, 6, 2, -2, -6, -10, -13, -14, -15, -14,
 -12, -9, -5, -1, 4, 7, 11, 13, 15, 6, -6, -15, -15, -6, 6, 15, -13, -12, -10, -7, -3, 1, 5, 8, 11, 12, 13, 13, 11, 9, 6, 2, -2, -6, -9, -11, -13, -13, -12, -11, -8, -5,
 -1, 3, 7, 10, 12, 13, -3, -16, -9, 9, 16, 3, -13, 0,0,0,1,1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15,
 15, 15, 14, 12, 11, 8, 6, 3, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 11, 3, -6, -12, -15, -14,
 -8, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 3, -11, -15, -8, 6, 15, 12, 1, 1, 2, 2, 3, 4, 4, 5,
 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, -6, -15, -3, 14, 11, -8, -15, -2, -4, -6, -8, -10, -11, -13, -14, -14, -15, -15,
 -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 15, 14, 12, 11, 8, 6, 3, -2, -4, -6, -8, -10, -11, -13, -14, -14, -15, -15, -15, -14,
 -13, -12, -11, -9, -8, -6, -3, -1, 1, 3, 5, 7, 9, 10, 12, 13, 14, 14, 15, 11, 3, -6, -12, -15, -14, -8, -2, -4, -6, -8, -9, -10, -12, -12, -13, -14, -14, -13, -13, -12,
 -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 13, 13, 14, 3, -11, -15, -8, 6, 15, 12, -2, -4, -5, -7, -8, -9, -11, -11, -12, -12, -12, -12, -12, -11, -10, -9,
 -8, -6, -5, -3, -1, 1, 2, 4, 6, 7, 9, 10, 11, 12, 12, 12, -6, -15, -3, 14, 11, -8, -15, 4, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 4, 0, -4, -7, -10, -12, -14, -15, -15,
 -14, -12, -10, -7, -4, 0, 4, 7, 10, 12, 14, 15, 15, 14, 12, 10, 8, 6, 3, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10,
 -7, -3, 0, 3, 7, 10, 12, 14, 15, 10, 3, -6, -12, -15, -14, -8, 3, 6, 9, 11, 13, 14, 14, 13, 11, 9, 6, 3, 0, -3, -6, -9, -11, -13, -14, -14, -13, -11, -9, -6, -3, 0, 3,
 6, 9, 11, 13, 14, 3, -10, -15, -8, 6, 15, 12, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12,
 -6, -15, -3, 14, 10, -8, -15, -5, -9, -13, -15, -15, -13, -11, -6, -2, 4, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11, -14, -15, -14, -12, -9, -4, 1, 6, 10, 13, 15, 15, 13,
 12, 10, 8, 6, 3, -5, -9, -13, -14, -15, -13, -10, -6, -1, 3, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11, -14, -15, -14, -12, -9, -4, 1, 6, 10, 13, 15, 10, 3, -5, -12, -15,
 -13, -8, -5, -9, -12, -13, -14, -12, -10, -6, -1, 3, 8, 11, 13, 14, 13, 10, 7, 3, -2, -6, -10, -12, -14, -13, -11, -8, -4, 1, 5, 9, 12, 13, 3, -10, -15, -8, 6, 15, 12,
 -4, -8, -11, -12, -12, -11, -9, -5, -1, 3, 7, 10, 12, 12, 12, 9, 6, 2, -2, -6, -9, -11, -12, -12, -10, -7, -4, 1, 5, 8, 11, 12, -5, -15, -3, 13, 10, -8, -15, 1,15, 15,
 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13,
 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, 0, 0, 0, -1, -1, -1, 0, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 10, 10, 10, 9,
 9, 8, 8, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, 0, -1, 0, 0, 1, 1, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2,
 1, 1, 0, -1, 0, 1, 0, 0, -1, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, 14, 14, 13, 11, 10, 8, 6, 4, 2, 2, 2, 2, 2, 1,
 1, 0, -15, -14, -14, -13, -12, -10, -9, -7, -5, -3, -1, 1, 3, 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, 14, 14, 13, 11, 10, 8, 6, 4, 2, 2, 0, -1, -2, -2, -2, -1, -14, -13,
 -13, -12, -11, -10, -8, -6, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 12, 13, 13, 14, 14, 13, 12, 12, 10, 9, 8, 6, 4, 2, 0, -2, -2, -1, 1, 2, 2, -12, -12, -12, -11, -10, -9,
 -7, -6, -4, -2, -1, 1, 3, 5, 6, 8, 9, 10, 11, 12, 12, 12, 12, 12, 11, 11, 9, 8, 7, 5, 4, 2, -1, -2, 0, 2, 2, -1, -2, 15, 14, 12, 10, 7, 4, 0, -4, -7, -10, -12, -14, -15,
 -15, -14, -12, -10, -7, -4, 0, 4, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 4, 3, 3, 3, 2, 2, 1, 1, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10,
 -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 2, 1, -1, -3, -4, -3, -2, 14, 13, 11, 9, 6, 3, 0, -3, -6, -9, -11, -13, -14, -14, -13, -11, -9, -6, -3, 0, 3, 6,
 9, 11, 13, 14, 14, 13, 11, 9, 6, 3, 1, -2, -4, -2, 1, 3, 3, 12, 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 12, 12,
 10, 8, 6, 3, -1, -4, -1, 3, 2, -2, -3, -15, -13, -10, -6, -1, 4, 9, 12, 14, 15, 14, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -4, 2, 6, 11, 13, 15, 15, 13, 9, 5, 5,
 5, 4, 4, 3, 2, 1, -15, -13, -10, -6, -1, 4, 9, 12, 14, 15, 14, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -3, 1, 6, 10, 13, 15, 14, 13, 9, 5, 4, 1, -2, -4, -5, -5,
 -3, -13, -12, -9, -5, -1, 4, 8, 11, 13, 14, 12, 10, 6, 2, -3, -7, -10, -13, -14, -13, -11, -8, -3, 1, 6, 10, 12, 14, 13, 12, 9, 5, 1, -4, -5, -3, 2, 5, 4, -12, -11, -8,
 -5, -1, 4, 7, 10, 12, 12, 11, 9, 6, 2, -2, -6, -9, -12, -12, -12, -10, -7, -3, 1, 5, 9, 11, 12, 12, 11, 8, 4, -2, -5, -1, 5, 4, -3, -5, 1,2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 8, 9, 10, 11, 11, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
 6, 6, 6, 6, 6, 6, 10, 11, 9, 4, -2, -7, -11, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 11, 4, -6, -11, -7, 2, 10,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 6, -7, -10, 2, 11, 4, -9, -3, -3, -3, -3,
 -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 6, 8, 11, 12, 14, 15, 15, -8, -8, -8, -8, -7, -7, -7, -6, -5, -5, -4, -3,
 -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 7, 7, 7, 8, 8, 8, 8, 14, 15, 12, 6, -3, -11, -15, -12, -12, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2,
 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 12, 12, 15, 6, -8, -15, -11, 3, 14, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8,
 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 8, -11, -14, 3, 15, 6, -12, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2,
 3, 3, 3, 6, 8, 11, 12, 14, 15, 15, 8, 8, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -8, -8, -8, -8, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7, 8, 8, 14, 15, 12, 6, -3,
 -11, -15, 12, 12, 11, 10, 8, 6, 3, 1, -1, -3, -6, -8, -10, -11, -12, -12, -12, -12, -11, -10, -8, -6, -3, -1, 1, 3, 6, 8, 10, 11, 12, 12, 15, 6, -8, -15, -11, 3, 14, 15,
 14, 13, 12, 9, 7, 4, 1, -1, -4, -7, -9, -12, -13, -14, -15, -15, -14, -13, -12, -9, -7, -4, -1, 1, 4, 7, 9, 12, 13, 14, 15, 8, -11, -14, 3, 15, 6, -12, -3, -3, -2, -2,
 -1, 0, 1, 2, 2, 3, 3, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -3, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 6, 8, 11, 12, 14, 15, 15, -8, -7, -6, -4, -2, 0, 3, 5, 7, 8, 8, 8, 7, 5, 3,
 1, -1, -3, -5, -7, -8, -8, -8, -7, -5, -3, 0, 2, 4, 6, 7, 8, 14, 15, 12, 6, -3, -11, -15, -12, -11, -9, -6, -3, 1, 4, 7, 10, 12, 12, 12, 11, 8, 5, 2, -2, -5, -8, -11,
 -12, -12, -12, -10, -7, -4, -1, 3, 6, 9, 11, 12, 15, 6, -8, -15, -11, 3, 14, -15, -14, -11, -8, -4, 1, 5, 9, 12, 14, 15, 15, 13, 10, 6, 2, -2, -6, -10, -13, -15, -15,
 -14, -12, -9, -5, -1, 4, 8, 11, 14, 15, 8, -11, -14, 3, 15, 6, -12, 1,0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 6,
 8, 11, 12, 14, 15, 15, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 14, 15, 12, 6, -3, -11, -15, 1, 1, 2, 2, 3, 4, 4,
 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 15, 6, -8, -15, -11, 3, 14, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10,
 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 8, -11, -14, 3, 15, 6, -12, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -1,
 -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 6, 8, 11, 12, 14, 15, 15, -1, -2, -3, -4, -5, -6, -7, -7, -8, -8, -8, -8, -8, -7, -7, -6, -5, -4, -3, -2, -1, 0, 2, 3, 4, 5, 6,
 6, 7, 8, 8, 8, 14, 15, 12, 6, -3, -11, -15, -2, -4, -5, -7, -8, -9, -11, -11, -12, -12, -12, -12, -12, -11, -10, -9, -8, -6, -5, -3, -1, 1, 2, 4, 6, 7, 9, 10, 11, 12,
 12, 12, 15, 6, -8, -15, -11, 3, 14, -2, -4, -6, -8, -10, -11, -13, -14, -14, -15, -15, -15, -14, -13, -12, -11, -9, -8, -6, -3, -1, 1, 3, 5, 7, 9, 10, 12, 13, 14, 14,
 15, 8, -11, -14, 3, 15, 6, -12, 1, 1, 2, 2, 3, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 6, 8, 10, 12, 14, 15, 15, 2,
 4, 5, 7, 8, 8, 8, 8, 7, 5, 4, 2, 0, -2, -4, -5, -7, -8, -8, -8, -8, -7, -5, -4, -2, 0, 2, 4, 5, 7, 8, 8, 14, 15, 12, 6, -3, -10, -15, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8,
 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 15, 6, -8, -15, -10, 3, 14, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7,
 -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 8, -10, -14, 3, 15, 6, -12, -1, -2, -3, -3, -3, -3, -2, -1, 0, 1, 2, 2, 3, 3, 3, 2, 2, 1, 0, -1,
 -2, -3, -3, -3, -2, -2, -1, 0, 1, 2, 3, 3, 6, 8, 10, 12, 13, 15, 15, -3, -5, -7, -8, -8, -7, -6, -3, -1, 2, 4, 6, 8, 8, 8, 6, 4, 2, -1, -4, -6, -7, -8, -8, -7, -5, -2,
 0, 3, 5, 7, 8, 13, 15, 12, 6, -3, -10, -15, -4, -8, -11, -12, -12, -11, -9, -5, -1, 3, 7, 10, 12, 12, 12, 9, 6, 2, -2, -6, -9, -11, -12, -12, -10, -7, -4, 1, 5, 8, 11,
 12, 15, 6, -8, -15, -10, 3, 13, -5, -9, -13, -14, -15, -13, -10, -6, -1, 3, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11, -14, -15, -14, -12, -9, -4, 1, 6, 10, 13, 15, 8,
 -10, -13, 3, 15, 6, -12, 1,3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 8, 8, 8, 8, 8, 8, 8, 8,
 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 1, 1, 1, 0, 0, 0, -1, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7,
 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 0, 0, -1, 0, 0, 1, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1,
 1, 0, 0, -1, 0, 1, 0, -1, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 1, 1, 2, 2, 2, 2, 2, -8, -8, -8, -7,
 -6, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 8, 8, 8, 8, 7, 7, 6, 5, 4, 3, 2, 1, 2, 2, 2, 1, 0, -2, -2, -12, -12, -12, -11, -10, -9, -7, -6, -4, -2, -1, 1, 3,
 5, 6, 8, 9, 10, 11, 12, 12, 12, 12, 12, 11, 11, 9, 8, 7, 5, 4, 2, 2, 1, -1, -2, -2, 0, 2, -15, -14, -14, -13, -12, -10, -9, -7, -5, -3, -1, 1, 3, 6, 8, 9, 11, 12, 13,
 14, 15, 15, 15, 14, 14, 13, 11, 10, 8, 6, 4, 2, 1, -2, -2, 0, 2, 1, -2, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 3, 3,
 2, 2, 1, 1, 1, 2, 2, 3, 3, 3, 4, 8, 8, 7, 5, 4, 2, 0, -2, -4, -5, -7, -8, -8, -8, -8, -7, -5, -4, -2, 0, 2, 4, 5, 7, 8, 8, 8, 8, 7, 5, 4, 2, 3, 4, 3, 1, -1, -2, -3, 12,
 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6, 3, 3, 1, -2, -4, -2, 1, 3, 15, 14, 12, 10, 7, 3, 0,
 -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 2, -2, -3, 1, 4, 1, -3, -3, -3, -2, -1, 0, 1, 2, 2, 3, 3, 3, 2, 1,
 0, -1, -2, -2, -3, -3, -3, -2, -2, -1, 0, 1, 2, 3, 3, 3, 3, 2, 1, 2, 3, 4, 4, 5, 5, 5, -8, -7, -5, -3, 0, 2, 5, 7, 8, 8, 7, 6, 4, 1, -2, -4, -6, -8, -8, -8, -6, -4, -2,
 1, 3, 6, 7, 8, 8, 7, 5, 3, 5, 5, 4, 2, -1, -4, -5, -12, -11, -8, -5, -1, 4, 7, 10, 12, 12, 11, 9, 6, 2, -2, -6, -9, -12, -12, -12, -10, -7, -3, 1, 5, 9, 11, 12, 12, 11,
 8, 4, 5, 2, -3, -5, -4, 1, 5, -15, -13, -10, -6, -1, 4, 9, 12, 14, 15, 14, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -3, 1, 6, 10, 13, 15, 14, 13, 9, 5, 3, -4, -5,
 1, 5, 2, -4, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9,
 -11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, -10, -4, 5, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, -2, -11, -6, 6, 11, 2,
 -9, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5,
 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13,
 -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14,
 15, 15, 15, 15, 16, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7,
 -6, -5, -5, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 6, -6, -15, -15, -6,
 6, 15, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4,
 4, 5, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, -3, -16, -9, 9, 16, 3, -13, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3,
 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8,
 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 13, 9, 3, -3, -9,
 -13, -16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -14, -14, -13, -13,
 -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 6, -6, -15, -15, -6, 6, 15, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7,
 6, 5, 3, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 3,
 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, -3, -16, -9, 9, 16, 3, -13, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11, 11, 11, 11,
 10, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 6, 8, 10, 12, 13, 14, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 3, 1, -1, -3, -6, -8, -10, -11, -13,
 -14, -15, -16, -16, -16, -15, -14, -13, -12, -10, -8, -6, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 13, 9, 3, -3, -9, -13, -16, -15, -14, -14, -13, -12, -10, -8,
 -7, -5, -2, 0, 2, 4, 6, 8, 10, 11, 12, 13, 14, 15, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, -15, -14, -13, -12, -11, -10,
 -8, -6, -4, -2, 0, 2, 5, 7, 8, 10, 12, 13, 14, 14, 15, 6, -6, -15, -15, -6, 6, 15, -13, -13, -12, -12, -10, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7, 9, 10, 11, 12, 13, 13, 13,
 13, 13, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -6, -8, -10, -11, -12, -13, -13, -13, -13, -13, -12, -11, -10, -9, -7, -5, -4, -2, 0, 2, 4, 6, 8, 9, 10, 12, 12, 13, 13,
 -3, -16, -9, 9, 16, 3, -13, 0,0,0,0,1,15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11,
 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14,
 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9,
 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0,
 0, 0, 0, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 14, 14,
 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 0, 0, -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11,
 -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9,
 8, 7, 6, 5, 4, 3, 2, 1, 1, 0, 0, -1, -1, -1, -1, -14, -14, -14, -13, -13, -13, -12, -12, -11, -11, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7,
 8, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -1, -1, 0, 1, 1, -12, -12, -12,
 -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 0, -1, 0, 1, 1, -1, -1, 15, 15, 14, 14, 13, 12, 11, 9, 8, 6, 4, 3, 1, -1, -3, -5, -6, -8, -10, -11, -12,
 -13, -14, -15, -15, -15, -15, -15, -14, -14, -13, -12, -10, -9, -7, -6, -4, -2, 0, 2, 3, 5, 7, 8, 10, 11, 12, 13, 14, 15, 15, 15, 15, 15, 14, 13, 13, 11, 10, 9, 7, 5, 4,
 2, 2, 2, 2, 1, 1, 1, 0, 15, 15, 14, 14, 13, 12, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -15, -15, -14, -13, -13, -11, -10, -9,
 -7, -6, -4, -2, 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, 1, 0, -1, -2, -2, -2, -1, 14, 14, 13, 12, 12, 11, 10, 8, 7,
 5, 4, 2, 1, -1, -3, -4, -6, -7, -9, -10, -11, -12, -13, -13, -14, -14, -14, -13, -13, -12, -12, -11, -9, -8, -7, -5, -4, -2, 0, 1, 3, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14,
 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 3, 2, 0, -1, -2, -1, 1, 2, 2, 12, 12, 12, 11, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -12, -12, -12,
 -12, -12, -12, -12, -11, -11, -10, -9, -7, -6, -5, -3, -2, 0, 1, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 12, 12, 12, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, -1, -2, 0, 2, 1, -1,
 -2, -15, -15, -14, -13, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 5, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14,
 -12, -11, -9, -6, -4, -2, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11, 10, 7, 5, 3, 2, 2, 2, 2, 1, 1, 0, -15, -14, -14, -12, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10,
 12, 13, 14, 15, 15, 15, 14, 13, 11, 10, 8, 5, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -14, -13, -12, -10, -9, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15,
 15, 14, 13, 11, 9, 7, 5, 2, 2, 0, -1, -2, -3, -2, -1, -14, -13, -12, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 14, 14, 13, 12, 11, 9, 7, 5, 3, 0, -2, -4,
 -6, -8, -10, -12, -13, -13, -14, -14, -13, -12, -11, -10, -8, -6, -4, -1, 1, 3, 5, 8, 9, 11, 12, 13, 14, 14, 13, 13, 12, 10, 9, 7, 5, 2, 0, -2, -3, -1, 1, 2, 2, -12,
 -12, -11, -10, -9, -7, -6, -4, -2, 1, 3, 5, 7, 8, 10, 11, 12, 12, 12, 12, 12, 11, 10, 8, 6, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -12, -12, -12, -12, -11, -10, -9, -7,
 -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 12, 12, 12, 12, 12, 11, 9, 8, 6, 4, 2, -1, -3, 0, 2, 2, -1, -2, 0,0,1,3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8,
 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14,
 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0,
 0, 0, 0, 0, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 1, 1,
 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 1, 1, 0, 0, -1, -1, -12, -12, -12, -12, -12,
 -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 1, 0, -1, -1, -1, 0, 1, -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1,
 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, -1, -1, 0, 1,
 0, -1, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2,
 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 8, 8, 8, 7, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -7, -8, -8, -8, -8,
 -8, -8, -7, -7, -6, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 8, 8, 8, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 2, 2, 2, 1, 0, -1, -2, 12, 12, 12, 11, 11, 10,
 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -12, -12, -12, -12, -12, -12, -12, -11, -11, -10, -9, -7, -6, -5, -3, -2, 0, 1, 3, 4, 6, 7, 8, 9, 10, 11, 12,
 12, 12, 12, 12, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 2, 1, -1, -2, -1, 0, 2, 15, 15, 14, 14, 13, 12, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13, -14,
 -14, -15, -15, -15, -15, -14, -13, -13, -11, -10, -9, -7, -6, -4, -2, 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, 1, -1,
 -2, 0, 2, 1, -2, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2,
 -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 2, 2, 1, 1, 0, 1, 1, 2, 2, 2, 2, 3, -8, -8, -7, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8, 8, 8, 8, 7, 6, 5, 4,
 3, 2, 0, -1, -3, -4, -5, -6, -7, -7, -8, -8, -8, -8, -7, -7, -6, -5, -3, -2, -1, 1, 2, 3, 4, 6, 6, 7, 8, 8, 8, 8, 8, 7, 6, 5, 4, 3, 1, 2, 3, 2, 1, 0, -2, -2, -12, -12,
 -11, -10, -9, -7, -6, -4, -2, 1, 3, 5, 7, 8, 10, 11, 12, 12, 12, 12, 12, 11, 10, 8, 6, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -12, -12, -12, -12, -11, -10, -9, -7, -5,
 -3, -1, 1, 3, 5, 7, 9, 10, 11, 12, 12, 12, 12, 12, 11, 9, 8, 6, 4, 2, 2, 1, -1, -3, -2, 0, 2, -15, -14, -14, -12, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15,
 15, 15, 14, 13, 11, 10, 8, 5, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -14, -13, -12, -10, -9, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11,
 9, 7, 5, 2, 1, -2, -2, 0, 3, 1, -2, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 5, -4, -10, -10, -4, 5,
 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, -2, -11, -6, 6, 11, 2, -9, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10,
 -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1,
 -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -15, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5,
 -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12,
 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -15, -15, -15, -15, -15, -15,
 -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -9, -8, -8,
 -8, -7, -7, -7, -6, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9,
 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 6, -6, -15, -15, -6,
 6, 15, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -12, -12, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10,
 -9, -9, -9, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4,
 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, -3, -16, -9, 9, 16, 3, -13, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2,
 -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10,
 -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1,
 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15,
 -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10,
 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10,
 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14,
 -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0,
 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 6, -6, -15, -15, -6, 6, 15, 13, 13, 13, 13, 13, 13, 13,
 12, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -4, -5, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11,
 -11, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6,
 -5, -5, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, -3, -16, -9, 9, 16, 3,
 -13, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, -16, -16, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14,
 14, 15, 15, 15, 15, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14,
 -14, -15, -15, -15, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 10, 11, 12,
 13, 13, 14, 14, 15, 15, 15, 16, 16, 13, 9, 3, -3, -9, -13, -16, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2,
 3, 4, 6, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 0, 0, -2, -3, -4, -5, -6, -7,
 -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -8, -7, -6, -6, -4, -3, -2, -1, 0, 1,
 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 6, -6, -15, -15, -6, 6, 15, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -7,
 -6, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3,
 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -7,
 -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, -3, -16, -9, 9, 16, 3, -13, 0,0,0,0,1,15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6,
 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4,
 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10,
 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2,
 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5,
 -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1,
 0, 0, 0, 0, 0, 0, 0, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6,
 -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3,
 2, 2, 1, 0, 0, 0, 0, 0, -1, 0, 0, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -7,
 -7, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13,
 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 4,
 3, 2, 2, 2, 1, 0, 0, 0, -1, 0, 0, 0, 0, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7,
 -7, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 4,
 3, 3, 2, 2, 1, 1, 0, 0, -1, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7,
 -8, -9, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -9, -8, -8,
 -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12,
 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 0, 0, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -2,
 -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11,
 -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14,
 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 4, 4, 3, 2, 1, 1, 0, 0, -1, -1, -1, -1, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4,
 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -14, -14, -13, -13, -13, -12, -12,
 -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14,
 14, 14, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -1, -1, 0, 1, 1, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 9, 8, 8, 7,
 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12,
 -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 2, 1, 0, -1, 0, 1, 1, -1, -1, -15, -15, -15, -14, -14, -14, -13, -12, -11,
 -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1,
 -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0,
 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 0, -15, -15, -15, -14,
 -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 3, 4, 5, 6, 7, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10,
 9, 8, 7, 6, 4, 3, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -9, -8,
 -6, -5, -4, -3, -2, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 0, 0, -1, -1,
 -1, -1, -14, -14, -13, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -5, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 14, 13,
 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -14, -14, -13, -13, -12, -12,
 -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 5,
 4, 2, 1, 0, -1, -1, -1, 1, 1, 1, -12, -12, -12, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -12, -12, -12, -12, -12, -12,
 -12, -12, -11, -11, -10, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 9, 9,
 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, 0, 1, 1, -1, -1, 0,0,1,3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 6,
 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6,
 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3,
 -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -4,
 -4, -4, -4, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, -12,
 -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2,
 -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, -1, 0, 0, 0, -15,
 -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3,
 -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, 0, 0, 0,
 1, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3,
 -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 5, 5,
 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7, -7,
 -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7,
 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, -1,
 -1, -2, -3, -4, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -11, -10, -10,
 -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 2, 1, 1, 0, -1, -1, -1, 0, 1, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3,
 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13,
 -13, -12, -12, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15,
 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 4, 4, 3, 2, 1, 1, -1, -1, 0, 1, 0, -1, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1,
 -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2,
 -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, -8, -8, -8, -8, -8, -7, -7, -6, -6, -6, -5, -4, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7,
 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7,
 -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 1, 1, 1,
 1, 0, -1, -1, -12, -12, -12, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11,
 -10, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2,
 1, 1, 1, -1, -1, -1, 0, 1, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 3, 4, 5, 6, 7, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15,
 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, -15, -15,
 -14, -14, -14, -13, -12, -12, -11, -10, -9, -8, -6, -5, -4, -3, -2, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12,
 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, -1, -1, 0, 1, 1, -1, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7,
 -9, -10, -11, -11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11, -10,
 -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, -15, -6, 6, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14, -15, -14, -6, 6, 14, 12, 8,
 3, -3, -8, -12, -14, -14, -12, -8, -3, 3, 8, 12, 14, -14, -6, 6, 14, 9, 1, -7, -13, -15, -11, -4, 4, 11, 15, 13, 7, -1, -9, -14, 8, -8, -8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 11, -11, -11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11, -11, -11, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11,
 11, -11, -11, 11, 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11, -4, 10, -10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -6, 15, -15, 6, 6, 6, 5, 4, 3,
 2, 1, -1, -2, -3, -4, -5, -6, -6, -6, -6, 14, -14, 6, 5, 4, 1, -1, -4, -5, -6, -6, -5, -4, -1, 1, 4, 5, 6, -6, 14, -14, 6, 4, 1, -3, -6, -6, -5, -2, 2, 5, 6, 6, 3, -1,
 -4, -6, 0,0,0,1,5, 9, 13, 14, 14, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 5, 9, 13, 14, 13, 11, 8, 4, 0, -4, -8, -11, -13, -14, -14, -13, -11, -8, -4, 5, 9, 13,
 14, 11, 5, -1, -8, -13, -14, -13, -9, -3, 4, 10, 14, 14, 12, 7, 5, 9, 12, 14, 8, -1, -10, -14, -13, -5, 4, 12, 14, 11, 3, -7, -13, -14, -9, -13, -13, 0, 13, 13, 13, 12,
 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, -13, -13, 0, 13, 12, 10, 7, 4, 0, -4, -7, -10, -12, -13, -13, -12, -10, -7, -4, -13, -13, 0, 13, 10, 5, -1, -7, -11, -13, -12,
 -8, -2, 4, 9, 12, 13, 11, 6, -12, -12, 0, 12, 7, -1, -9, -13, -11, -5, 4, 11, 13, 10, 2, -6, -12, -13, -8, 14, -5, -13, 9, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1,
 14, -5, -13, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, -9, -9, -7, -5, -3, 14, -5, -13, 9, 7, 4, -1, -5, -8, -9, -9, -6, -2, 3, 7, 9, 9, 8, 4, 14, -5, -12, 9, 5, -1, -7, -9,
 -8, -4, 3, 8, 9, 7, 2, -4, -9, -9, -6, -9, 14, -13, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, -9, 14, -13, 5, 5, 4, 3, 1, 0, -1, -3, -4, -5, -5, -5, -5, -4, -3,
 -1, -9, 14, -13, 5, 4, 2, 0, -3, -4, -5, -5, -3, -1, 1, 4, 5, 5, 4, 2, -9, 14, -12, 5, 3, 0, -4, -5, -4, -2, 1, 4, 5, 4, 1, -2, -5, -5, -3, 1,14, 13, 9, 5, 5, 5, 5, 5,
 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 14, 13, 9, 5, 5, 4, 3, 1, 0, -1, -3, -4, -5, -5, -5, -5, -4, -3, -1, 14, 13, 9, 5, 4, 2, 0, -3, -4, -5, -5, -3, -1, 1, 4, 5, 5, 4, 2,
 14, 12, 9, 5, 3, 0, -4, -5, -4, -2, 1, 4, 5, 4, 1, -2, -5, -5, -3, -13, 0, 13, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, -13, 0, 13, 13, 12, 10, 7, 4, 0,
 -4, -7, -10, -12, -13, -13, -12, -10, -7, -4, -13, 0, 13, 13, 10, 5, -1, -7, -11, -13, -12, -8, -2, 4, 9, 12, 13, 11, 6, -12, 0, 12, 12, 7, -1, -9, -13, -11, -5, 4, 11,
 13, 10, 2, -6, -12, -13, -8, 9, -13, -5, 14, 14, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 9, -13, -5, 14, 13, 11, 8, 4, 0, -4, -8, -11, -13, -14, -14, -13, -11,
 -8, -4, 9, -13, -5, 14, 11, 5, -1, -8, -13, -14, -13, -9, -3, 4, 10, 14, 14, 12, 7, 9, -12, -5, 14, 8, -1, -10, -14, -13, -5, 4, 12, 14, 11, 3, -7, -13, -14, -9, -5, 13,
 -14, 9, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, -5, 13, -14, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, -9, -9, -7, -5, -3, -5, 13, -14, 9, 7, 4, -1, -5, -8, -9, -9, -6,
 -2, 3, 7, 9, 9, 8, 4, -5, 12, -14, 9, 5, -1, -7, -9, -8, -4, 3, 8, 9, 7, 2, -4, -9, -9, -6, 0,1,0, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 1, 3,
 4, 4, 8, 11, 13, 14, 14, 13, 11, 8, 4, 0, -4, -8, -11, -13, -14, 2, 4, 6, 7, 12, 14, 14, 10, 4, -3, -9, -13, -14, -13, -8, -1, 5, 11, 14, 3, 6, 8, 9, 14, 13, 7, -3, -11,
 -14, -12, -4, 5, 13, 14, 10, 1, -8, -14, -1, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, -4, -4, 0, 4, 7, 10, 12, 13, 13, 12, 10, 7, 4, 0, -4, -7,
 -10, -12, -13, -6, -6, 0, 6, 11, 13, 12, 9, 4, -2, -8, -12, -13, -11, -7, -1, 5, 10, 13, -8, -8, 0, 8, 13, 12, 6, -2, -10, -13, -11, -4, 5, 11, 13, 9, 1, -7, -12, 1, 0,
 -1, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, 4, -1, -4, 3, 5, 7, 9, 9, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, 7, -2, -6, 4, 8, 9, 9, 7, 3, -2, -6, -9, -9, -8, -5,
 -1, 4, 7, 9, 9, -3, -8, 6, 9, 9, 4, -2, -7, -9, -8, -3, 4, 8, 9, 7, 1, -5, -9, -1, 1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, -3, 4, -4, 1, 3, 4, 5, 5, 5, 5,
 4, 3, 1, 0, -1, -3, -4, -5, -5, -4, 7, -6, 2, 4, 5, 5, 4, 1, -1, -3, -5, -5, -4, -3, 0, 2, 4, 5, -6, 9, -8, 3, 5, 5, 2, -1, -4, -5, -4, -1, 2, 4, 5, 4, 0, -3, -5, 1,1,
 1, 1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 4, 4, 3, 1, 3, 4, 5, 5, 5, 5, 4, 3, 1, 0, -1, -3, -4, -5, -5, 7, 6, 4, 2, 4, 5, 5, 4, 1, -1, -3, -5, -5, -4, -3, 0,
 2, 4, 5, 9, 8, 6, 3, 5, 5, 2, -1, -4, -5, -4, -1, 2, 4, 5, 4, 0, -3, -5, -1, 0, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, -4, 0, 4, 4, 7, 10, 12, 13,
 13, 12, 10, 7, 4, 0, -4, -7, -10, -12, -13, -6, 0, 6, 6, 11, 13, 12, 9, 4, -2, -8, -12, -13, -11, -7, -1, 5, 10, 13, -8, 0, 8, 8, 13, 12, 6, -2, -10, -13, -11, -4, 5,
 11, 13, 9, 1, -7, -12, 1, -1, 0, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 14, 3, -4, -1, 4, 8, 11, 13, 14, 14, 13, 11, 8, 4, 0, -4, -8, -11, -13, -14, 4, -6,
 -2, 7, 12, 14, 14, 10, 4, -3, -9, -13, -14, -13, -8, -1, 5, 11, 14, 6, -8, -3, 9, 14, 13, 7, -3, -11, -14, -12, -4, 5, 13, 14, 10, 1, -8, -14, 0, 1, -1, 1, 2, 3, 4, 4,
 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, -1, 4, -4, 3, 5, 7, 9, 9, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, -2, 6, -7, 4, 8, 9, 9, 7, 3, -2, -6, -9, -9, -8, -5, -1, 4, 7, 9, -3, 8,
 -9, 6, 9, 9, 4, -2, -7, -9, -8, -3, 4, 8, 9, 7, 1, -5, -9, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10,
 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -13, -9, -3, 3,
 9, 13, 16, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -16, -15, -13, -9, -3, 3, 9, 13, 15, 13, 9, 3, -3, -9, -13, -15, -15, -13, -9, -3, 3, 9, 13, 15, -15,
 -13, -8, -3, 3, 8, 13, 15, 10, 2, -7, -14, -16, -12, -4, 4, 12, 16, 14, 7, -2, -10, -15, 10, 5, -4, -10, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 15, 6, -6, -15, -15, -6, 6, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14, -15, 14, 6, -6, -14, -14, -6, 6, 14, 12, 8, 3, -3, -8, -12, -14, -14,
 -12, -8, -3, 3, 8, 12, 14, 14, 6, -6, -14, -14, -6, 6, 14, 9, 1, -7, -13, -15, -11, -4, 4, 11, 15, 13, 7, -1, -9, -14, -9, 2, 11, 6, -6, -11, -2, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, -13, 3, 16, 9, -9, -16, -3, 13, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -13, -13, 3, 15, 9, -9, -15, -3, 13, 11, 7, 3, -3, -7,
 -11, -13, -13, -11, -7, -3, 3, 7, 11, 13, -13, 3, 15, 8, -8, -15, -3, 13, 8, 1, -6, -12, -13, -10, -4, 4, 10, 13, 12, 6, -1, -8, -13, 0,0,0,1,3, 6, 8, 10, 12, 13, 15,
 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 4, 3, 1, 3, 6, 8, 10, 12, 13, 15, 15, 14, 11, 8, 4, 0, -4, -8, -11, -14, -15, -15, -14, -11, -8, -4, 3, 5, 8, 10, 12, 13,
 14, 15, 11, 6, -1, -8, -13, -15, -14, -9, -3, 4, 10, 14, 15, 12, 7, 3, 5, 8, 10, 12, 13, 14, 14, 8, -1, -10, -15, -13, -6, 4, 12, 15, 11, 3, -7, -14, -15, -9, -8, -13,
 -15, -12, -5, 3, 10, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, -8, -13, -15, -12, -5, 3, 10, 15, 13, 11, 8, 4, 0, -4, -8, -11, -13, -15, -15, -13, -11,
 -8, -4, -8, -13, -15, -12, -5, 3, 10, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4, 10, 14, 15, 12, 7, -8, -13, -14, -12, -5, 3, 10, 14, 8, -1, -10, -15, -13, -5, 4, 12,
 15, 11, 3, -7, -13, -14, -9, 12, 15, 6, -8, -15, -10, 3, 13, 13, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 1, 12, 15, 6, -8, -15, -10, 3, 13, 12, 10, 7, 4, 0, -4, -7,
 -10, -12, -13, -13, -12, -10, -7, -4, 12, 14, 5, -8, -15, -10, 3, 13, 10, 5, -1, -7, -12, -13, -12, -8, -3, 4, 9, 13, 13, 11, 6, 12, 14, 5, -8, -14, -10, 3, 13, 7, -1,
 -9, -13, -12, -5, 4, 11, 13, 10, 3, -6, -12, -13, -8, -15, -8, 10, 13, -3, -15, -5, 12, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 3, 2, 1, -15, -8, 10, 13, -3, -15, -5,
 12, 11, 9, 7, 3, 0, -3, -7, -9, -11, -12, -12, -11, -9, -7, -3, -14, -8, 10, 13, -3, -15, -5, 12, 9, 5, -1, -7, -11, -12, -11, -8, -2, 3, 9, 12, 12, 10, 6, -14, -8, 10,
 13, -3, -14, -5, 12, 7, -1, -9, -12, -11, -5, 3, 10, 12, 9, 2, -6, -11, -12, -8, 1,15, 15, 13, 12, 10, 8, 6, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 15, 15, 13,
 12, 10, 8, 6, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, 15, 14, 13, 12, 10, 8, 5, 3, 2, 1, 0, -2, -3, -3, -3, -2, -1, 1, 2, 3, 3, 2, 1, 14, 14, 13, 12,
 10, 8, 5, 3, 2, 0, -2, -3, -3, -1, 1, 2, 3, 2, 1, -1, -3, -3, -2, -15, -10, -3, 6, 12, 15, 13, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 4, 4, 3, 2, 2, 1, -15, -10, -3, 6, 12, 15,
 13, 8, 7, 6, 4, 2, 0, -2, -4, -6, -7, -8, -8, -7, -6, -4, -2, -14, -10, -3, 5, 12, 15, 13, 8, 6, 3, -1, -4, -7, -8, -7, -5, -2, 2, 6, 8, 8, 7, 4, -14, -10, -3, 5, 12,
 14, 13, 8, 4, -1, -6, -8, -7, -3, 2, 7, 8, 6, 2, -4, -7, -8, -5, 13, 3, -10, -15, -8, 6, 15, 12, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 3, 2, 1, 13, 3, -10, -15, -8,
 6, 15, 12, 11, 9, 7, 3, 0, -3, -7, -9, -11, -12, -12, -11, -9, -7, -3, 13, 3, -10, -15, -8, 5, 14, 12, 9, 5, -1, -7, -11, -12, -11, -8, -2, 3, 9, 12, 12, 10, 6, 13, 3,
 -10, -14, -8, 5, 14, 12, 7, -1, -9, -12, -11, -5, 3, 10, 12, 9, 2, -6, -11, -12, -8, -12, 6, 15, 3, -13, -10, 8, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1,
 -12, 6, 15, 3, -13, -10, 8, 15, 13, 11, 8, 4, 0, -4, -8, -11, -13, -15, -15, -13, -11, -8, -4, -12, 5, 15, 3, -13, -10, 8, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4,
 10, 14, 15, 12, 7, -12, 5, 14, 3, -13, -10, 8, 14, 8, -1, -10, -15, -13, -5, 4, 12, 15, 11, 3, -7, -13, -14, -9, 0,1,0, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 9, 10, 11,
 12, 13, 14, 14, 15, 15, 15, 1, 2, 2, 3, 3, 4, 4, 4, 8, 11, 14, 15, 15, 14, 11, 8, 4, 0, -4, -8, -11, -14, -15, 1, 3, 4, 5, 6, 6, 7, 7, 12, 15, 14, 10, 4, -3, -9, -14,
 -15, -13, -8, -1, 6, 11, 15, 2, 3, 5, 6, 8, 8, 9, 9, 15, 14, 7, -3, -11, -15, -12, -4, 6, 13, 15, 10, 1, -8, -14, -1, -1, -1, -1, 0, 0, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11,
 12, 13, 13, 14, 14, 15, 15, -2, -4, -4, -3, -2, 1, 3, 4, 8, 11, 13, 15, 15, 13, 11, 8, 4, 0, -4, -8, -11, -13, -15, -4, -6, -7, -6, -2, 1, 5, 7, 12, 15, 14, 10, 4, -3,
 -9, -13, -15, -13, -8, -1, 5, 11, 14, -5, -8, -9, -8, -3, 2, 6, 9, 14, 13, 7, -3, -11, -15, -12, -4, 5, 13, 15, 10, 1, -8, -14, 1, 1, 1, -1, -1, -1, 0, 1, 3, 4, 5, 6, 7,
 8, 9, 10, 11, 12, 12, 13, 13, 13, 13, 3, 4, 2, -2, -4, -3, 1, 4, 7, 10, 12, 13, 13, 12, 10, 7, 4, 0, -4, -7, -10, -12, -13, 6, 7, 3, -4, -7, -5, 1, 6, 11, 13, 13, 9, 4,
 -3, -8, -12, -13, -12, -7, -1, 5, 10, 13, 8, 9, 3, -5, -9, -6, 2, 8, 13, 12, 6, -3, -10, -13, -11, -4, 5, 12, 13, 9, 1, -7, -13, -1, -1, 1, 1, 0, -1, 0, 1, 2, 3, 5, 6,
 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, -4, -2, 3, 4, -1, -4, -2, 3, 7, 9, 11, 12, 12, 11, 9, 7, 3, 0, -3, -7, -9, -11, -12, -7, -4, 5, 6, -1, -7, -2, 6, 10, 12, 12, 9,
 3, -2, -8, -11, -12, -11, -7, -1, 5, 9, 12, -9, -5, 6, 8, -2, -9, -3, 8, 12, 11, 6, -2, -9, -12, -10, -3, 5, 11, 12, 9, 1, -7, -12, 1,1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 3, 3, 2, 2, 1, 2, 2, 3, 3, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, 7, 7, 6, 6, 5, 4, 3, 1, 2, 3, 3, 2, 1, -1, -2, -3, -3, -3, -2,
 0, 1, 2, 3, 9, 9, 8, 8, 6, 5, 3, 2, 3, 3, 1, -1, -2, -3, -2, -1, 1, 3, 3, 2, 0, -2, -3, -1, -1, 0, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, -4, -3,
 -1, 2, 3, 4, 4, 2, 4, 6, 7, 8, 8, 7, 6, 4, 2, 0, -2, -4, -6, -7, -8, -7, -5, -1, 3, 6, 7, 6, 4, 7, 8, 8, 6, 2, -2, -5, -7, -8, -7, -4, -1, 3, 6, 8, -9, -6, -2, 3, 8, 9,
 8, 5, 8, 7, 4, -2, -6, -8, -7, -2, 3, 7, 8, 6, 1, -4, -8, 1, 0, -1, -1, -1, 1, 1, 1, 2, 3, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, 4, 1, -3, -4, -2, 2, 4, 3, 7, 9,
 11, 12, 12, 11, 9, 7, 3, 0, -3, -7, -9, -11, -12, 6, 1, -5, -7, -4, 3, 7, 6, 10, 12, 12, 9, 3, -2, -8, -11, -12, -11, -7, -1, 5, 9, 12, 8, 2, -6, -9, -5, 3, 9, 8, 12,
 11, 6, -2, -9, -12, -10, -3, 5, 11, 12, 9, 1, -7, -12, -1, 1, 1, 0, -1, -1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, -3, 2, 4, 1, -4, -3, 2, 4, 8, 11,
 13, 15, 15, 13, 11, 8, 4, 0, -4, -8, -11, -13, -15, -6, 3, 7, 1, -6, -5, 4, 7, 12, 15, 14, 10, 4, -3, -9, -13, -15, -13, -8, -1, 5, 11, 14, -8, 3, 9, 2, -8, -6, 5, 9,
 14, 13, 7, -3, -11, -15, -12, -4, 5, 13, 15, 10, 1, -8, -14, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5,
 -1, -7, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -14, -12, -10, -8, -4,
 -2, 2, 4, 8, 10, 12, 14, 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, 13, 9,
 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8,
 -2, -10, -15, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 13, 9, 3, -3, -9, -13, -16,
 -16, -13, -9, -3, 3, 9, 13, 16, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -16, 15, 13, 9, 3, -3, -9, -13, -15, -15, -13, -9, -3, 3, 9, 13, 15, 13, 9, 3,
 -3, -9, -13, -15, -15, -13, -9, -3, 3, 9, 13, 15, 15, 13, 8, 3, -3, -8, -13, -15, -15, -13, -8, -3, 3, 8, 13, 15, 10, 2, -7, -14, -16, -12, -4, 4, 12, 16, 14, 7, -2,
 -10, -15, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -15, -10, -2, 8, 14, 16, 12, 4, -4,
 -12, -16, -14, -8, 2, 10, 15, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -15, -15, -10, -2, 7, 14, 16, 12, 4, -4, -12, -16, -14, -7, 2, 10, 15, 13, 8, 3,
 -3, -8, -13, -15, -15, -13, -8, -3, 3, 8, 13, 15, -15, -10, -2, 7, 14, 15, 12, 4, -4, -12, -15, -14, -7, 2, 10, 15, 10, 2, -7, -14, -15, -12, -4, 4, 12, 15, 14, 7, -2,
 -10, -15, 0,0,0,1,1, 3, 4, 6, 7, 8, 9, 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 9, 8, 7, 6, 4, 3, 1, 1, 3, 4, 6, 7, 8, 9, 11, 12, 13, 13, 14,
 15, 15, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 1, 3, 4, 6, 7, 8, 9, 11, 12, 12, 13, 14, 14, 15, 15, 15, 12, 6, -1, -8, -13, -15, -14, -9,
 -3, 4, 11, 15, 15, 13, 7, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 8, -1, -11, -15, -13, -6, 4, 13, 15, 12, 3, -7, -14, -15, -9, -4, -8, -12, -14, -15,
 -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 15, 15, 14, 13, 13, 12, 11, 9, 8, 7, 6, 4, 3, 1, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 14,
 12, 8, 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 12, 6, -1, -8, -13, -15, -14, -9, -3, 4,
 11, 15, 15, 13, 7, -4, -8, -11, -13, -15, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13, 15, 8, -1, -11, -15, -13, -6, 4, 13, 15, 12, 3, -7, -14, -15, -9, 7, 13, 15, 15, 11, 4,
 -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 15, 15, 14, 14, 13, 12, 12, 11, 9, 8, 7, 6, 4, 3, 1, 7, 13, 15, 15, 11, 4, -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 14, 12, 8,
 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 7, 12, 15, 14, 11, 4, -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 12, 6, -1, -8, -13, -15, -14, -9, -3, 4, 11, 14, 15,
 12, 7, 7, 12, 15, 14, 10, 4, -3, -9, -13, -15, -13, -8, -1, 5, 11, 14, 8, -1, -11, -15, -13, -6, 4, 12, 15, 12, 3, -7, -14, -15, -9, -9, -15, -14, -7, 3, 12, 15, 13, 4,
 -6, -13, -15, -11, -1, 8, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, -9, -15, -14, -7, 3, 12, 15, 13, 4, -6, -13, -15, -11, -1, 8, 15, 13, 11, 8, 4, 0, -4,
 -8, -11, -13, -15, -15, -13, -11, -8, -4, -9, -15, -14, -7, 3, 12, 15, 12, 4, -6, -13, -15, -11, -1, 8, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4, 10, 14, 15, 12, 7,
 -9, -14, -13, -7, 3, 11, 15, 12, 4, -5, -13, -15, -10, -1, 8, 14, 8, -1, -10, -15, -13, -5, 4, 12, 15, 11, 3, -7, -13, -14, -9, 1,15, 15, 15, 15, 14, 13, 13, 12, 11, 9,
 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 15, 15, 15, 15, 14, 13, 13, 12, 11, 9, 8, 7, 6, 4, 3, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1,
 0, 15, 15, 15, 14, 14, 13, 12, 12, 11, 9, 8, 7, 6, 4, 3, 1, 1, 1, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1,
 0, -1, -1, -1, -1, 0, 1, 1, 1, 0, -1, -1, -1, -1, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, -15, -14,
 -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -3, -2, -1, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4,
 3, 2, 0, -2, -4, -4, -4, -3, -1, 1, 3, 4, 4, 4, 2, -15, -13, -11, -8, -4, 0, 4, 8, 11, 13, 15, 15, 13, 11, 8, 4, 2, 0, -3, -4, -4, -2, 1, 4, 4, 3, 1, -2, -4, -4, -3, 15,
 12, 6, -1, -8, -13, -15, -14, -9, -3, 4, 11, 15, 15, 13, 7, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 15, 12, 6, -1, -8, -13, -15, -14, -9, -3, 4, 11, 15, 15, 13, 7,
 6, 5, 4, 2, 0, -2, -4, -5, -6, -7, -7, -6, -5, -4, -2, 15, 12, 6, -1, -8, -13, -15, -14, -9, -3, 4, 11, 14, 15, 12, 7, 5, 3, -1, -4, -6, -7, -6, -4, -1, 2, 5, 7, 7, 6,
 3, 14, 11, 5, -1, -8, -13, -15, -13, -9, -3, 4, 10, 14, 15, 12, 7, 4, -1, -5, -7, -6, -3, 2, 6, 7, 5, 1, -3, -6, -7, -4, -15, -8, 1, 11, 15, 13, 6, -4, -13, -15, -12,
 -3, 7, 14, 15, 9, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, -15, -8, 1, 11, 15, 13, 6, -4, -13, -15, -12, -3, 7, 14, 15, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, -9, -9,
 -7, -5, -3, -14, -8, 1, 11, 15, 13, 6, -4, -12, -15, -12, -3, 7, 14, 15, 9, 7, 4, -1, -5, -8, -9, -9, -6, -2, 3, 7, 9, 9, 8, 4, -14, -8, 1, 10, 15, 13, 5, -4, -12, -15,
 -11, -3, 7, 13, 14, 9, 5, -1, -7, -9, -8, -4, 3, 8, 9, 7, 2, -4, -9, -9, -6, 0,1,0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 9, 11, 12, 13, 13, 14,
 15, 15, 15, 15, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7,
 7, 13, 15, 15, 11, 4, -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, 15, 14, 7, -3, -12, -15, -13, -4, 6, 13, 15, 11, 1, -8,
 -15, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 3, 4, 6, 7, 8, 9, 11, 12, 13, 13, 14, 15, 15, 15, 15, -1, -2, -3, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,
 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -2, -4, -5, -6, -7, -7, -6, -5, -4, -2, 0, 2, 4, 5, 6, 7, 13, 15, 15, 11, 4, -3, -9, -14, -15, -13, -8, -1,
 6, 12, 15, -3, -5, -7, -9, -9, -9, -9, -7, -5, -3, 0, 3, 5, 7, 9, 9, 15, 14, 7, -3, -12, -15, -13, -4, 6, 13, 15, 11, 1, -8, -15, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1,
 -1, 0, 1, 1, 1, 3, 4, 6, 7, 8, 9, 11, 12, 12, 13, 14, 14, 15, 15, 15, 2, 4, 4, 4, 3, 1, -1, -3, -4, -4, -4, -2, 0, 2, 3, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8,
 -12, -14, -15, 3, 6, 7, 7, 5, 2, -1, -4, -6, -7, -6, -4, -1, 3, 5, 7, 12, 15, 14, 11, 4, -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 4, 8, 9, 9, 7, 3, -2, -6, -9, -9, -8,
 -5, -1, 4, 7, 9, 15, 14, 7, -3, -12, -15, -12, -4, 6, 13, 15, 11, 1, -8, -14, -1, -1, -1, -1, 0, 1, 1, 1, 0, -1, -1, -1, -1, 0, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13,
 13, 14, 14, 15, 15, -3, -4, -4, -2, 1, 3, 4, 4, 1, -2, -4, -4, -3, 0, 2, 4, 8, 11, 13, 15, 15, 13, 11, 8, 4, 0, -4, -8, -11, -13, -15, -4, -7, -6, -3, 1, 5, 7, 6, 2, -3,
 -6, -7, -5, -1, 4, 7, 12, 15, 14, 10, 4, -3, -9, -13, -15, -13, -8, -1, 5, 11, 14, -6, -9, -9, -4, 2, 7, 9, 8, 3, -4, -8, -9, -7, -1, 5, 9, 14, 13, 7, -3, -11, -15, -12,
 -4, 5, 13, 15, 10, 1, -8, -14, 1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1,
 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, 7, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 1, 1, 1, 9, 9, 9, 9, 9,
 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 1, 1, 1, 0, -1, -1, -1, 0, 1, 1, 1, 1, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4,
 4, 4, 4, 4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 4, 4, 3, 2, 1, 2, 3, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -7, -6, -5, -4, -2, 0, 2, 4, 5, 6, 7, 7, 6, 5, 4, 2,
 4, 4, 4, 3, 1, -1, -3, -4, -4, -4, -2, 0, 2, 3, 4, -9, -9, -7, -5, -3, 0, 3, 5, 7, 9, 9, 9, 9, 7, 5, 3, 4, 4, 2, -1, -3, -4, -4, -1, 2, 4, 4, 3, 0, -2, -4, 1, 1, 1, 0,
 -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 7, 4, 3, 2, 0, -2, -4, -4, -4, -3, -1, 1, 3, 4, 4, 4, 2, 4, 5, 6, 7, 7, 6, 5, 4, 2, 0,
 -2, -4, -5, -6, -7, 7, 5, 3, -1, -4, -6, -7, -6, -4, -1, 2, 5, 7, 7, 6, 3, 6, 7, 7, 5, 2, -1, -4, -6, -7, -6, -4, -1, 3, 5, 7, 9, 7, 4, -1, -5, -8, -9, -9, -6, -2, 3, 7,
 9, 9, 8, 4, 7, 6, 3, -1, -5, -7, -6, -2, 3, 6, 7, 5, 1, -4, -7, -1, -1, 0, 1, 1, 1, 1, 0, -1, -1, -1, 0, 1, 1, 1, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 9, -4, -2,
 0, 3, 4, 4, 2, -1, -4, -4, -3, -1, 2, 4, 4, 3, 5, 7, 9, 9, 9, 9, 7, 5, 3, 0, -3, -5, -7, -9, -9, -7, -4, 1, 5, 7, 6, 3, -2, -6, -7, -5, -1, 3, 6, 7, 4, 8, 9, 9, 7, 3,
 -2, -6, -9, -9, -8, -5, -1, 4, 7, 9, -9, -5, 1, 7, 9, 8, 4, -3, -8, -9, -7, -2, 4, 9, 9, 6, 9, 9, 4, -2, -7, -9, -8, -3, 4, 8, 9, 7, 1, -5, -9, 1,8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11,
 -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 15, 14, 12, 10, 8, 4,
 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16,
 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 12, 13,
 14, 14, 15, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7,
 -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16,
 -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, 16, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10,
 -12, -14, -15, -16, -16, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 15, 14, 12, 10, 7,
 4, 2, -2, -4, -7, -10, -12, -14, -15, -15, -15, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10,
 -15, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5, -8, -10, -11, -11, -11, -9, -7, -4, 0, 3, 6, 8, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, -16, -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 15, 14, 12, 10, 8,
 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14,
 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15, -14, -11, -8, -4, 1, 5, 9, 12, 14, 15, 15, 13, 10, 6, 2, -2, -6, -10, -13, -15, -15, -14, -12, -9,
 -5, -1, 4, 8, 11, 14, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 0,0,0,1,1, 2, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13,
 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 2, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13,
 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12,
 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11,
 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 8, -1, -11, -15, -14, -6, 4, 13, 15, 12, 3, -7, -14, -15, -10, -2, -4, -7, -9, -10, -12, -13, -14, -15, -15,
 -15, -15, -15, -14, -13, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, -2, -4, -7, -9, -10, -12,
 -13, -14, -15, -15, -15, -15, -15, -14, -13, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8,
 -4, -2, -4, -6, -8, -10, -12, -13, -14, -15, -15, -15, -15, -14, -14, -13, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 12, 6, -1, -8, -14, -15, -14,
 -10, -3, 4, 11, 15, 15, 13, 7, -2, -4, -6, -8, -10, -11, -13, -14, -14, -15, -15, -15, -14, -13, -12, -11, -9, -8, -6, -3, -1, 1, 3, 5, 7, 9, 10, 12, 13, 14, 14, 15, 8,
 -1, -11, -15, -14, -6, 4, 13, 15, 12, 3, -7, -14, -15, -10, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4,
 7, 10, 13, 14, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10,
 -7, -4, 0, 4, 7, 10, 13, 14, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14,
 -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 12, 6, -1, -8, -13, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3,
 -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 8, -1, -11, -15, -13, -6, 4, 13, 15, 12, 3, -7, -14, -15, -10, -5, -10, -13, -15, -15, -14,
 -11, -7, -2, 4, 9, 12, 15, 15, 14, 12, 8, 3, -2, -7, -11, -14, -15, -15, -13, -9, -4, 1, 6, 10, 13, 15, 15, 15, 15, 14, 13, 13, 12, 11, 9, 8, 7, 6, 4, 3, 1, -5, -10,
 -13, -15, -15, -14, -11, -7, -2, 4, 9, 12, 15, 15, 14, 12, 8, 3, -2, -7, -11, -14, -15, -15, -13, -9, -4, 1, 6, 10, 13, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -15,
 -14, -12, -8, -4, -5, -10, -13, -15, -15, -14, -11, -6, -2, 4, 8, 12, 14, 15, 14, 12, 8, 3, -2, -7, -11, -14, -15, -15, -13, -9, -4, 1, 6, 10, 13, 15, 12, 6, -1, -8,
 -13, -15, -14, -9, -3, 4, 11, 15, 15, 13, 7, -5, -9, -13, -14, -15, -13, -10, -6, -1, 3, 8, 12, 14, 15, 14, 11, 8, 3, -2, -7, -11, -14, -15, -14, -12, -9, -4, 1, 6, 10,
 13, 15, 8, -1, -11, -15, -13, -6, 4, 13, 15, 12, 3, -7, -14, -15, -9, 1,15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5,
 4, 4, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3,
 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2,
 2, 1, 1, 0, 0, 0, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 0, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1,
 0, 0, 0, -1, -1, 0, 0, 1, 1, 1, 0, 0, -1, -1, 0, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 11, 13, 14, 15, 15, 15, 15, 15, 14, 13, 12, 10, 9, 7,
 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 11, 13, 14, 15, 15, 15, 15, 15, 14, 13, 12, 10, 9,
 7, 4, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -2, -2, -2, -1, -1, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 11, 13, 14, 14, 15, 15, 15, 15, 14, 13,
 12, 10, 8, 6, 4, 2, 2, 1, 0, -1, -2, -2, -2, -1, 0, 1, 2, 2, 2, 2, 1, -15, -14, -14, -13, -12, -10, -9, -7, -5, -3, -1, 1, 3, 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, 14,
 14, 13, 11, 10, 8, 6, 4, 2, 1, 0, -2, -2, -2, -1, 1, 2, 2, 2, 0, -1, -2, -2, -1, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4,
 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0,
 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 3, 3, 2, 1, 0, -1, -2, -3, -3, -4, -4, -3, -3, -2, -1, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13,
 -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 3, 1, 0, -2, -3, -4, -3, -2, -1, 1, 3, 3, 4, 3, 2, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15,
 -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 2, 0, -3, -4, -3, -1, 1, 3, 4, 3, 1, -2, -3, -4, -2, -15, -13, -10, -6, -1, 4, 9, 13, 15, 15, 14,
 11, 7, 2, -3, -8, -12, -14, -15, -15, -12, -9, -4, 2, 7, 11, 14, 15, 15, 13, 10, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, -15, -13, -10, -6, -1, 4, 9, 13, 15, 15,
 14, 11, 7, 2, -3, -8, -12, -14, -15, -15, -12, -9, -4, 2, 7, 11, 14, 15, 15, 13, 10, 5, 5, 4, 3, 1, 0, -1, -3, -4, -5, -5, -5, -5, -4, -3, -1, -15, -13, -10, -6, -1, 4,
 9, 13, 15, 15, 14, 11, 7, 2, -3, -8, -12, -14, -15, -14, -12, -8, -4, 2, 6, 11, 14, 15, 15, 13, 10, 5, 4, 2, 0, -3, -5, -5, -5, -3, -1, 1, 4, 5, 5, 4, 2, -15, -13, -10,
 -6, -1, 4, 9, 12, 14, 15, 14, 11, 7, 2, -3, -8, -11, -14, -15, -14, -12, -8, -3, 1, 6, 10, 13, 15, 14, 13, 9, 5, 3, 0, -4, -5, -5, -2, 1, 4, 5, 4, 1, -2, -5, -5, -3, 1,
1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 10, 11, 11, 11, 11, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 6, 9, 10, 11, 11, 10, 9, 6, 3, 0, -3, -6, -8, -10, -11, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 9, 11, 11, 8, 3, -2, -7, -10, -11, -10, -6, -1, 4, 9, 11, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 11, 10, 5, -2, -8, -11, -9, -3, 4, 10, 11, 8, 1, -6, -11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2,
 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -7, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3,
 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -5, -4, -3, -2, -1, 0, 0, 1, 2,
 3, 4, 5, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 4, 4, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4, -3,
 -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, 7, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -7, -7, -7, -6, -5,
 -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 10, 9, 9, 8, 6, 5, 3, 1, -1, -3, -5, -6, -8, -9, -9, -10, -10, -9,
 -9, -8, -6, -5, -3, -1, 1, 3, 5, 6, 8, 9, 9, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1,
 -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, -4, -4, -3, -2, -1, 0, 2, 3, 4, 4, 4, 4, 4, 3, 2, 1, -1, -2, -3,
 -4, -4, -4, -4, -4, -3, -2, 0, 1, 2, 3, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -7, -6, -5, -4, -2, 0, 2, 4, 6, 7, 7, 7, 6, 5, 3, 1, -1, -3, -5,
 -6, -7, -7, -7, -6, -4, -2, 0, 2, 4, 5, 6, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, -10, -9, -7, -5, -2, 0, 3, 6, 8, 9, 10, 9, 8, 7, 4, 1, -1,
 -4, -7, -8, -9, -10, -9, -8, -6, -3, 0, 2, 5, 7, 9, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 1,0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7,
 7, 7, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10,
 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, -1, -1, -2, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4,
 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -1, -2, -3, -4, -5, -5, -6, -6, -7, -7, -7, -7, -7, -6, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,
 5, 6, 6, 7, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, -1, -3, -4, -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -9, -8, -7, -6, -5, -4, -2, -1, 0,
 2, 3, 5, 6, 7, 8, 8, 9, 9, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, 0, 0, 0, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 13, 14, 15, 15, 15, 15, 1, 2, 3, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -3, -2, -1,
 0, 1, 2, 3, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, 2, 3, 5, 6, 7, 7, 7, 7, 6, 5, 3, 2, 0, -2, -3, -5, -6, -7, -7, -7, -7, -6, -5, -3, -2, 0,
 2, 3, 5, 6, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -13, -8, -1, 6, 12, 15, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5,
 -2, 0, 2, 5, 6, 8, 9, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 13, 15, 11, 1, -8, -15, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, 0, 0, 1, 1, 1, 1, 3, 4, 6, 7, 8, 9, 11, 12, 13, 13, 14, 15, 15, 15, 15, -1, -3, -4, -4, -4, -4, -3, -2, 0, 1, 2, 4, 4, 4, 4, 3, 2, 1, -1, -2, -3, -4, -4, -4,
 -4, -3, -1, 0, 2, 3, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -2, -4, -6, -7, -7, -6, -5, -3, -1, 2, 4, 6, 7, 7, 7, 5, 4, 1, -1, -3, -5, -6, -7,
 -7, -6, -4, -2, 0, 3, 5, 6, 7, 13, 15, 15, 11, 4, -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, -3, -6, -8, -9, -10, -9, -7, -4, -1, 2, 5, 8, 9, 10, 9, 7, 5, 2, -1, -5, -7,
 -9, -10, -9, -8, -6, -3, 0, 4, 6, 8, 9, 15, 14, 7, -3, -12, -15, -13, -4, 6, 13, 15, 11, 1, -8, -15, 1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0,
 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 0, 0, 0, 0,
 -1, -1, -1, 0, 0, 0, 1, 1, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 1, 1, 0, 0, -1, -1, -1, 0, 0, 1, 1, 0, 0,
 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, -4, -4, -4,
 -4, -4, -3, -3, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -7, -7, -7, -6, -6, -5,
 -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 6, 7, 7, 7, 7, 7, 6, 6, 5, 5, 4, 3, 2, 1, 2, 2, 2, 2, 1, 0, -1, -2, -2, -2, -1, 0, 1, 2, 2, -10, -9, -9, -8, -8, -7, -6, -5, -3,
 -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 4, 3, 1, 2, 2, 1, 0, -2, -2, -2, -1, 1, 2, 2, 2, 0, -1, -2, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -3,
 -2, -1, 0, 1, 2, 3, 4, 4, 4, 4, 4, 4, 3, 2, 1, 2, 3, 3, 4, 4, 3, 3, 2, 1, 0, -1, -2, -3, -3, -4, 7, 7, 6, 5, 3, 2, 0, -2, -3, -5, -6, -7, -7, -7, -7, -6, -5, -3, -2, 0,
 2, 3, 5, 6, 7, 7, 7, 7, 6, 5, 3, 2, 3, 4, 3, 3, 1, -1, -2, -3, -4, -3, -2, 0, 1, 3, 4, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5,
 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 4, 3, 2, -1, -3, -4, -3, -1, 1, 3, 4, 3, 0, -2, -3, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1,
 1, 1, 1, 1, 1, 1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, -4, -4, -3, -2, 0, 1, 3, 4, 4, 4, 4, 3, 2, 1, -1, -2, -3, -4, -4, -4, -4, -2, -1, 0, 2, 3, 4, 4, 4, 4,
 3, 1, 3, 4, 5, 5, 5, 5, 4, 3, 1, 0, -1, -3, -4, -5, -5, -7, -6, -5, -3, 0, 2, 4, 6, 7, 7, 6, 5, 3, 1, -1, -4, -5, -7, -7, -7, -6, -4, -2, 1, 3, 5, 6, 7, 7, 6, 4, 2, 4,
 5, 5, 4, 1, -1, -3, -5, -5, -5, -3, 0, 2, 4, 5, -9, -8, -6, -4, 0, 3, 6, 8, 9, 10, 9, 7, 5, 1, -2, -5, -7, -9, -10, -9, -8, -5, -2, 1, 4, 7, 9, 10, 9, 8, 6, 3, 5, 5, 2,
 -1, -4, -5, -4, -1, 2, 5, 5, 4, 0, -3, -5, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 7, 1, -5, -10,
 -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3, -3, -2,
 -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1,
 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15,
 -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4,
 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15, -15,
 -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -7, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 7,
 8, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 11, 11, 11, 11, 10, 10,
 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0,
 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2,
 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9,
 11, 12, 13, 14, 14, 15, 15, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4,
 -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13,
 14, 14, 15, 15, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -8, -9,
 -10, -11, -12, -13, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 12, 13, 14, 14, 15,
 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11, 11, 11, 11, 10, 9,
 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1,
 -4, -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8,
 -10, -12, -14, -15, -16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 6, 8, 10, 12, 13, 14, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 3, 1, -1, -3, -6, -8,
 -10, -11, -13, -14, -15, -16, -16, -16, -15, -14, -13, -12, -10, -8, -6, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3,
 9, 13, 16, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, 0, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 15, 15, 15, 14, 12, 11, 10, 7, 6, 3, 1, -1, -3, -6, -7, -10, -11, -12,
 -14, -15, -15, -15, -15, -15, -14, -13, -12, -10, -8, -6, -4, -2, 0, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 0,
0,0,0,1,15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 9, 8,
 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12,
 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5,
 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -9, -8, -7, -5,
 -4, -3, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -9, -8, -7, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1,
 -1, -1, 0, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14,
 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, -15, -15, -15, -14, -14,
 -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14,
 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 0, -1, -1, -1, 0, 0, 1, 1, 1, 0, -1, -1, -1, -1, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 4, 3, 1, -1, -3, -5, -7,
 -8, -10, -11, -12, -13, -14, -15, -15, -15, -15, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 3, 5, 7, 9, 10, 11, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 13,
 12, 10, 9, 7, 6, 4, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 4, 3, 1, -1, -3, -5, -7, -8, -10, -11, -12, -13, -14, -15, -15,
 -15, -15, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 3, 5, 7, 9, 10, 11, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 13, 12, 10, 9, 7, 6, 4, 2, 2, 1, 1, 1, 0,
 -1, -1, -1, -2, -2, -2, -2, -1, -1, -1, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 4, 3, 1, -1, -3, -5, -6, -8, -10, -11, -12, -13, -14, -15, -15, -15, -15, -15, -14, -14,
 -13, -12, -11, -9, -7, -6, -4, -2, 0, 2, 3, 5, 7, 8, 10, 11, 12, 13, 14, 15, 15, 15, 15, 15, 14, 14, 13, 12, 10, 9, 7, 5, 4, 2, 1, 1, 0, -1, -2, -2, -2, -1, 0, 1, 1, 2,
 2, 2, 1, 15, 15, 14, 14, 13, 12, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -15, -15, -14, -13, -13, -11, -10, -9, -7, -6, -4, -2,
 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, 1, 0, -1, -2, -2, -1, 1, 2, 2, 1, 0, -1, -2, -2, -1, -15, -15, -14, -13,
 -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 14, 15, 15, 15, 15, 15, 13, 12, 10, 8, 6, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -4,
 -2, 1, 4, 6, 9, 11, 12, 14, 15, 15, 15, 15, 14, 13, 12, 10, 8, 5, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, -15, -15, -14, -13, -11, -9, -7, -4, -2, 1, 3, 6, 8,
 10, 12, 14, 15, 15, 15, 15, 15, 13, 12, 10, 8, 6, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 12, 14, 15, 15,
 15, 15, 14, 13, 12, 10, 8, 5, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -2, -2, -1, -1, -15, -15, -14, -13, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15, 15, 15,
 14, 13, 12, 10, 8, 5, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14, -12, -11, -9, -6, -4, -2, 1, 4, 6, 8, 11, 12, 14, 15, 15, 15, 15, 14, 13, 12, 10, 7,
 5, 3, 2, 1, 0, -1, -2, -3, -2, -2, 0, 1, 2, 2, 3, 2, 1, -15, -14, -14, -12, -11, -9, -7, -4, -2, 1, 3, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11, 10, 8, 5, 3, 0, -2,
 -5, -7, -9, -11, -13, -14, -15, -15, -15, -14, -13, -12, -10, -9, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 11, 9, 7, 5, 2, 1, 0, -2, -3, -2, -1, 1, 2,
 3, 2, 0, -1, -2, -3, -2, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4,
 -8, -12, -14, -15, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5,
 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 15, 14, 7, -3, -12, -15, -13, -4,
 6, 14, 15, 11, 1, -8, -15, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 0, -1, -1, -1, -2, -2, -2, -2,
 -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3,
 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -1, -1, -2, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7,
 -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 13, 15, 15,
 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, -1, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9,
 -8, -8, -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15,
 11, 1, -8, -15, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 8, 12,
 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 5, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7,
 -7, -7, -7, -7, -7, -6, -6, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 4, 4, 5, 6, 6, 6, 7, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 1, 2, 3,
 5, 6, 6, 7, 8, 9, 9, 9, 10, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -7, -8, -9, -9, -9, -10, -10, -10, -9, -9, -8, -8, -7, -6, -5, -4, -3,
 -2, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 3, 4,
 6, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, -1, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2,
 1, 1, 0, -1, -2, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, 0, 1, 1, 2, 3, 3, 4, 4, 4, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 0, -4, -8, -12, -14, -15, -1,
 -2, -3, -4, -5, -6, -7, -7, -7, -7, -7, -6, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 6, 7, 7, 7, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -4, -5, -5, -6, -7, -7, -7, -7, -7,
 -6, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 6, 7, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, -2, -3, -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -8,
 -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -7, -8, -9, -10, -10, -10, -9, -8, -8, -6, -5, -4, -2, 0, 1, 3, 4, 6, 7,
 8, 9, 9, 10, 15, 14, 7, -3, -12, -15, -13, -4, 6, 14, 15, 11, 1, -8, -15, 1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4,
 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2,
 -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 0, 0, 0, -1, -1, -1, -1, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7,
 -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2,
 1, 1, 1, 1, 1, 0, -1, -1, -1, 0, 0, 1, 1, 1, 0, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 4, 4, 4, 4, 4, 3, 3, 3,
 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 3, 3, 3, 2, 2, 1, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 0, -1, -1, -1, -2, -2, 7, 7, 7, 6, 6, 6, 5, 4, 4, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -6, -6, -6, -7, -7, -7, -7,
 -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 2, 2, 2, 1, 1, 0, -1, -2, -2, -2, -1, 0, 1, 1,
 2, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 0, -1, -2, -3, -4, -5, -6, -7, -8, -8, -9, -9, -10, -10, -10, -9, -9, -9, -8, -7, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6,
 7, 8, 8, 9, 9, 10, 10, 10, 9, 9, 9, 8, 7, 6, 6, 5, 3, 2, 1, 2, 2, 1, 0, -1, -2, -2, -1, 1, 2, 2, 1, 0, -1, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, -4, -4, -4, -4, -3, -3, -2, -1, -1, 0, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2, 2, 1, 0, -1, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4,
 -4, -4, -3, -3, -2, -1, 0, 0, 1, 2, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2, 1, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -7, -7, -6, -6, -5, -4, -3, -2, -1, 0,
 2, 3, 4, 5, 6, 6, 7, 7, 7, 7, 7, 6, 5, 5, 4, 3, 1, 0, -1, -2, -3, -4, -5, -6, -6, -7, -7, -7, -7, -6, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 6, 7, 7, 7, 7, 7, 6, 5,
 4, 3, 2, 1, 2, 3, 2, 2, 1, 0, -2, -2, -3, -2, -1, 0, 1, 2, 3, -10, -9, -9, -8, -7, -6, -4, -3, -1, 0, 2, 4, 5, 6, 8, 8, 9, 10, 10, 10, 9, 8, 7, 6, 5, 3, 2, 0, -1, -3,
 -5, -6, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -4, -3, -1, 1, 2, 4, 5, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 6, 5, 3, 2, 3, 2, 1, 0, -2, -3, -2, -1, 1, 2, 3, 2, 0, -1,
 -2, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 7, 1, -5, -10, -11, -9, -3, 3, 9, 11, 10, 5, -1, -7, -11, -11, -11, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5,
 -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8,
 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11,
 -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4,
 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4,
 -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 13,
 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15,
 -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -13, -12, -12, -12, -12, -12, -11, -11, -11, -11, -10,
 -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6,
 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2,
 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8,
 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13,
 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15,
 -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4,
 -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4,
 -8, -10, -12, -14, -15, -16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4,
 -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14,
 -14, -14, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13,
 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12,
 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -13, -14, -14, -14,
 -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3,
 -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16,
 14, 8, -2, -10, -15, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1,
 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4,
 -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4,
 -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -16, -15, -15,
 -15, -14, -14, -13, -13, -12, -11, -10, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 15,
 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16,
 -15, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 13,
 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, -15, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6,
 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8,
 -9, -10, -11, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2,
 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 15, 15, 10, 2, -8, -14, -16, -12, -4, 4, 12, 16, 14, 8, -2, -10, -15, 0,0,0,0,1,15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6,
 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6,
 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6,
 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5,
 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -15, -15, -15, -15, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2,
 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5,
 -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0,
 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -10,
 -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13,
 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7,
 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 1, 0, 0, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12,
 -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9,
 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12,
 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12,
 12, 11, 10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15,
 -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13,
 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13,
 -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1,
 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4,
 3, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3,
 -4, -4, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -13, -12, -11,
 -11, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 0, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 11,
 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13,
 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 4, 4, 3, 2, 1, 1, 0, -1, -1, -1, 0, 0, 1, 1, 1, 0, 0, -1, -1,
 -1, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -13,
 -13, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 12, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6,
 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10,
 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14,
 -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -13, -13, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 12, 12, 13, 14, 14, 15, 15, 15, 15,
 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 4, 3, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, -15, -15, -15, -15, -14, -14, -13, -12, -11, -11,
 -10, -8, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1,
 -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2,
 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 7, 6, 5, 4, 3, 1, 1, 1, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1,
 1, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 3, 4, 5, 6, 7, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14,
 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12,
 -12, -11, -10, -9, -8, -6, -5, -4, -3, -2, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4,
 2, 1, 1, 0, -1, -1, -1, -1, 0, 1, 1, 1, 0, -1, -1, -1, -1, 0,0,1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3,
 -3, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2,
 -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1,
 -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1,
 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9,
 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2,
 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3,
 -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3,
 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 0, 0,
 -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -5, -6, -6, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4,
 -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3,
 2, 2, 1, 1, 0, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3,
 -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -8, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3,
 -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 6, 5, 5, 4, 3, 3, 2,
 2, 1, 1, 1, 1, 0, 0, -1, -1, -1, 0, 0, 1, 1, 1, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2,
 -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1,
 -1, -7, -7, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 4, 4,
 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -5, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1,
 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 1, 1, 1, -10, -10, -9, -9,
 -9, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 1,
 0, 0, -1, -2, -3, -4, -4, -5, -6, -6, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 1, 2, 3, 4,
 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 1, 0, -1, -1, -1, 0, 1, 1, 1, 1, 0, -1, -1, 1,8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4,
 -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3,
 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11, -10, -8, -5, -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -10, -4, 5,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, -15, -6, 6, 15, 15, 14, 14, 13, 13, 12,
 11, 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -6, 6, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13,
 -14, -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, -15, -6, 6, 15, 13, 11, 7, 4, -1, -5, -9, -12, -14, -15, -14, -13, -10, -6, -2, 2, 6, 10, 13,
 14, 15, 14, 12, 9, 5, 1, -4, -7, -11, -13, -15, 8, -8, -8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, -11, -11,
 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, -11, -11, 11, 11, 10, 9, 7, 5, 3, 1, -1,
 -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, -11, -11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11, -10, -8, -5,
 -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -4, 10, -10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, -6, 15, -15, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -6, -6, -6, 15, -15, 6, 6, 6, 5, 4, 3, 2, 1, -1,
 -2, -3, -4, -5, -6, -6, -6, -6, -6, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 6, 6, -6, 15, -15, 6, 6, 5, 3, 2, 0, -2, -4, -5, -6, -6, -6, -5, -4, -3, -1, 1, 3, 4, 5, 6,
 6, 6, 5, 4, 2, 0, -2, -3, -5, -6, -6, 0,1,11, 9, 7, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 15, 13, 10, 5, 5, 5,
 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, 15, 13, 10, 5, 5, 5, 4, 3, 2, 1, 1, -1, -1, -2, -3, -4, -5, -5, -5,
 -5, -5, -5, -4, -3, -2, -1, -1, 1, 1, 2, 3, 4, 5, 5, 5, 15, 13, 10, 5, 5, 4, 3, 1, 0, -2, -3, -4, -5, -5, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 5, 5, 4, 3, 2, 0, -1, -3,
 -4, -5, -5, -9, 0, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, -13, 0, 13, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8,
 7, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, 0, 13, 13, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -13,
 -13, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 13, -13, 0, 13, 13, 12, 10, 7, 3, -1, -4, -8, -11, -12, -13, -13, -11, -9, -5, -2, 2, 5, 9, 11, 13, 13, 12,
 11, 8, 4, 1, -3, -7, -10, -12, -13, 7, -9, -4, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 10, -13, -5, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, 10, -13, -5, 15,
 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14, -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13, 14, 15, 10, -13, -5, 15, 13, 11, 8, 4, -1, -5, -9,
 -12, -14, -15, -14, -13, -10, -6, -2, 2, 6, 10, 13, 14, 15, 14, 12, 9, 5, 1, -4, -8, -11, -13, -15, -4, 9, -10, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, -5, 13, -15, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10,
 -5, 13, -15, 10, 9, 9, 8, 6, 5, 3, 1, -1, -3, -5, -6, -8, -9, -9, -10, -10, -9, -9, -8, -6, -5, -3, -1, 1, 3, 5, 6, 8, 9, 9, 10, -5, 13, -15, 10, 9, 7, 5, 2, 0, -3, -6,
 -8, -9, -10, -9, -8, -7, -4, -1, 1, 4, 7, 8, 9, 10, 9, 8, 6, 3, 0, -2, -5, -7, -9, -10, 0,1,5, 10, 13, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 11, 11,
 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, 5, 10, 13, 15, 14, 14, 13, 12, 10, 9, 7, 5, 3, 1, -1, -3, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -14, -13, -13,
 -11, -10, -8, -6, -4, -2, 5, 10, 13, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 5,
 9, 13, 14, 13, 10, 6, 1, -4, -9, -12, -14, -15, -13, -11, -7, -2, 3, 8, 11, 14, 15, 14, 12, 8, 3, -1, -6, -10, -13, -15, -14, -13, -9, -5, -13, -13, 0, 13, 13, 13, 13,
 13, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, -13, -13, 0, 13, 13, 12, 11, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -7, -8,
 -10, -11, -12, -12, -13, -13, -13, -12, -12, -11, -10, -9, -7, -5, -4, -2, -13, -13, 0, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3, 0,
 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, -13, -13, 0, 13, 11, 9, 5, 1, -4, -8, -11, -12, -13, -12, -10, -6, -2, 2, 7, 10, 12, 13, 12, 10, 7, 3, -1, -5, -9, -12, -13,
 -13, -11, -8, -4, 15, -5, -13, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 15, -5, -13, 10, 9, 9, 8, 8, 7, 6, 5,
 3, 2, 0, -1, -2, -4, -5, -6, -7, -8, -9, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -4, -3, -1, 15, -5, -13, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8,
 -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 14, -5, -13, 9, 8, 6, 4, 0, -3, -6, -8, -9, -10, -9, -7, -5, -1, 2, 5, 7, 9, 10, 9, 8, 5, 2, -1, -4, -7, -9, -10,
 -9, -8, -6, -3, -10, 15, -13, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, -10, 15, -13, 5, 5, 5, 4, 4, 4, 3, 2, 2, 1,
 0, -1, -1, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -1, -1, -10, 15, -13, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -5, -5, -4, -3, -2, -1,
 0, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1, -9, 14, -13, 5, 4, 3, 2, 0, -1, -3, -4, -5, -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -4, -3, -2, 1,
15, 13, 10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 15, 13, 10, 5, 5, 5, 4, 4, 4, 3, 2, 2, 1, 0, -1, -1, -2, -3,
 -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -1, -1, 15, 13, 10, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5,
 5, 5, 4, 3, 2, 1, 14, 13, 9, 5, 4, 3, 2, 0, -1, -3, -4, -5, -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -4, -3, -2, -13, 0, 13, 13, 13, 13,
 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, -13, 0, 13, 13, 13, 12, 11, 10, 9, 8, 6, 4, 2, 1, -1, -3, -5, -7, -8,
 -10, -11, -12, -12, -13, -13, -13, -12, -12, -11, -10, -9, -7, -5, -4, -2, -13, 0, 13, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3, 0,
 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, -13, 0, 13, 13, 11, 9, 5, 1, -4, -8, -11, -12, -13, -12, -10, -6, -2, 2, 7, 10, 12, 13, 12, 10, 7, 3, -1, -5, -9, -12, -13,
 -13, -11, -8, -4, 10, -13, -5, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, 10, -13, -5, 15, 14, 14,
 13, 12, 10, 9, 7, 5, 3, 1, -1, -3, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -14, -13, -13, -11, -10, -8, -6, -4, -2, 10, -13, -5, 15, 14, 12, 10, 7, 3, 0, -3, -7,
 -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 9, -13, -5, 14, 13, 10, 6, 1, -4, -9, -12, -14, -15, -13, -11, -7, -2, 3,
 8, 11, 14, 15, 14, 12, 8, 3, -1, -6, -10, -13, -15, -14, -13, -9, -5, -5, 13, -15, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2,
 2, 1, 1, 0, -5, 13, -15, 10, 9, 9, 8, 8, 7, 6, 5, 3, 2, 0, -1, -2, -4, -5, -6, -7, -8, -9, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -4, -3, -1, -5, 13, -15, 10, 9, 8,
 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, -5, 13, -14, 9, 8, 6, 4, 0, -3, -6, -8, -9, -10, -9, -7, -5, -1,
 2, 5, 7, 9, 10, 9, 8, 5, 2, -1, -4, -7, -9, -10, -9, -8, -6, -3, 0,1,0, 0, 1, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14,
 14, 14, 14, 15, 15, 15, 1, 1, 2, 2, 4, 6, 8, 10, 11, 13, 13, 14, 15, 15, 14, 14, 13, 12, 11, 9, 8, 6, 3, 1, -1, -3, -5, -7, -9, -10, -12, -13, -14, -14, -15, 1, 2, 3, 3,
 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 2, 3, 4, 5, 9, 13, 14, 15, 13, 10, 6, 1, -3,
 -8, -12, -14, -15, -14, -11, -8, -3, 2, 7, 11, 13, 15, 14, 12, 9, 4, -1, -6, -10, -13, -14, -1, -1, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11,
 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, -2, -2, 0, 2, 4, 5, 7, 9, 10, 11, 12, 12, 13, 13, 13, 12, 12, 11, 10, 8, 7, 5, 3, 1, -1, -2, -4, -6, -8, -9, -10, -11,
 -12, -13, -13, -3, -3, 0, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12, -11, -9, -6, -3, 0, 3, 6, 9, 11, 12, 13, -4, -4, 0, 4, 8, 11,
 13, 13, 12, 9, 5, 1, -3, -7, -10, -12, -13, -12, -10, -7, -2, 2, 6, 10, 12, 13, 12, 11, 8, 4, -1, -5, -9, -11, -13, 1, 0, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6,
 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 2, -1, -2, 1, 3, 4, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -5, -6, -7, -8, -8, -9,
 -9, -10, 3, -1, -3, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 5, -2, -4, 3, 6, 8, 9, 10, 9, 7, 4,
 1, -2, -5, -8, -9, -10, -9, -7, -5, -2, 1, 5, 7, 9, 10, 9, 8, 6, 3, 0, -4, -6, -8, -9, 0, 1, -1, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, -1, 2, -2, 1, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, -1, -2, -2, -3, -4, -4, -4, -5, -5, -5, -2, 3, -3, 1, 2, 3, 4, 5, 5,
 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, -3, 5, -4, 2, 3, 4, 5, 5, 5, 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2,
 4, 5, 5, 5, 4, 3, 1, 0, -2, -3, -4, -5, 1,1, 1, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 2, 2, 1, 1, 1, 2, 3,
 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, -1, -2, -2, -3, -4, -4, -4, -5, -5, -5, 3, 3, 2, 1, 2, 3, 4, 5, 5, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -5,
 -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 5, 5, 4, 3, 2, 3, 4, 5, 5, 5, 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2, 4, 5, 5, 5, 4, 3, 1, 0, -2, -3, -4, -5, -1, 0, 1,
 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, -2, 0, 2, 2, 4, 5, 7, 9, 10, 11, 12, 12, 13, 13, 13, 12,
 12, 11, 10, 8, 7, 5, 3, 1, -1, -2, -4, -6, -8, -9, -10, -11, -12, -13, -13, -3, 0, 3, 3, 6, 9, 11, 12, 13, 13, 12, 11, 9, 6, 3, 0, -3, -6, -9, -11, -12, -13, -13, -12,
 -11, -9, -6, -3, 0, 3, 6, 9, 11, 12, 13, -4, 0, 4, 4, 8, 11, 13, 13, 12, 9, 5, 1, -3, -7, -10, -12, -13, -12, -10, -7, -2, 2, 6, 10, 12, 13, 12, 11, 8, 4, -1, -5, -9,
 -11, -13, 0, -1, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 1, -2, -1, 2, 4, 6, 8, 10, 11, 13,
 13, 14, 15, 15, 14, 14, 13, 12, 11, 9, 8, 6, 3, 1, -1, -3, -5, -7, -9, -10, -12, -13, -14, -14, -15, 2, -3, -1, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7,
 -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 3, -4, -2, 5, 9, 13, 14, 15, 13, 10, 6, 1, -3, -8, -12, -14, -15, -14, -11, -8, -3, 2, 7, 11,
 13, 15, 14, 12, 9, 4, -1, -6, -10, -13, -14, 0, 1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, -1, 2, -2, 1,
 3, 4, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -5, -6, -7, -8, -8, -9, -9, -10, -1, 3, -3, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6,
 -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, -2, 4, -5, 3, 6, 8, 9, 10, 9, 7, 4, 1, -2, -5, -8, -9, -10, -9, -7, -5, -2, 1, 5, 7, 9, 10, 9, 8, 6, 3, 0,
 -4, -6, -8, -9, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5,
 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11,
 -11, -11, -10, -8, -5, -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -13, -9, -3, 3, 9, 13, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4,
 -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -16, -16, -15, -14,
 -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, -16, -13, -9, -3, 3, 9, 13, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15, 16,
 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 10, 5, -4, -10, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 15, 6, -6, -15, -15, -6, 6, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -13, -13,
 -14, -14, -15, -15, 15, 6, -6, -15, -15, -6, 6, 15, 14, 13, 11, 9, 7, 4, 1, -1, -4, -7, -9, -11, -13, -14, -15, -15, -14, -13, -11, -9, -7, -4, -1, 1, 4, 7, 9, 11, 13,
 14, 15, 15, 6, -6, -15, -15, -6, 6, 15, 13, 11, 7, 4, -1, -5, -9, -12, -14, -15, -14, -13, -10, -6, -2, 2, 6, 10, 13, 14, 15, 14, 12, 9, 5, 1, -4, -7, -11, -13, -15, -9,
 2, 11, 6, -6, -11, -2, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 12, 12,
 11, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, 3, 16, 9, -9, -16, -3, 13, 13, 12, 10, 8, 6, 4, 1, -1,
 -4, -6, -8, -10, -12, -13, -13, -13, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 13, -13, 3, 16, 9, -9, -16, -3, 13, 12, 10, 7, 3, -1, -5, -8, -11, -12, -13,
 -13, -11, -9, -6, -2, 2, 6, 9, 11, 13, 13, 12, 11, 8, 5, 1, -3, -7, -10, -12, -13, 0,1,11, 11, 10, 9, 8, 6, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3,
 -3, -3, 15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 15, 15, 14, 12, 11, 8, 6,
 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -11, -7, -2, 4, 9, 11, 10, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -15, -11, -3, 6, 12, 15, 14, 8, 8, 8, 8, 7, 7, 7, 6, 5, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5,
 -5, -6, -7, -7, -7, -8, -8, -8, -8, -15, -11, -3, 6, 12, 15, 14, 8, 8, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -8, -8, -8, -8, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6,
 7, 8, 8, -15, -11, -3, 6, 12, 15, 14, 8, 7, 6, 4, 2, 0, -3, -5, -7, -8, -8, -8, -7, -5, -3, -1, 1, 3, 5, 7, 8, 8, 8, 7, 5, 3, 0, -2, -4, -6, -7, -8, 10, 2, -7, -11, -6,
 4, 11, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 14, 3, -11, -15, -8, 6, 15, 12, 12, 12, 12, 11, 11, 10, 9, 8, 7,
 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -12, -12, 14, 3, -11, -15, -8, 6, 15, 12, 12, 11, 10, 8, 6, 3, 1, -1, -3, -6, -8, -10,
 -11, -12, -12, -12, -12, -11, -10, -8, -6, -3, -1, 1, 3, 6, 8, 10, 11, 12, 12, 14, 3, -11, -15, -8, 6, 15, 12, 11, 9, 6, 3, -1, -4, -7, -10, -12, -12, -12, -11, -8, -5,
 -2, 2, 5, 8, 11, 12, 12, 12, 10, 7, 4, 1, -3, -6, -9, -11, -12, -9, 4, 11, 2, -10, -7, 6, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -12, 6, 15, 3, -14, -11, 8, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -8, -9, -10,
 -11, -12, -13, -14, -14, -15, -15, -15, -12, 6, 15, 3, -14, -11, 8, 15, 14, 13, 12, 9, 7, 4, 1, -1, -4, -7, -9, -12, -13, -14, -15, -15, -14, -13, -12, -9, -7, -4, -1,
 1, 4, 7, 9, 12, 13, 14, 15, -12, 6, 15, 3, -14, -11, 8, 15, 14, 11, 8, 4, -1, -5, -9, -12, -14, -15, -15, -13, -10, -6, -2, 2, 6, 10, 13, 15, 15, 14, 12, 9, 5, 1, -4,
 -8, -11, -14, -15, 0,1,3, 6, 8, 11, 12, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 2, 1, 3, 6,
 8, 11, 12, 14, 15, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -14, -14, -13, -11, -10, -8, -6, -4, -2, 3, 6, 8, 10,
 12, 14, 15, 15, 14, 12, 10, 7, 4, 0, -4, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -4, 0, 4, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 4, 3, 6, 8, 10, 12, 13, 15, 15,
 13, 10, 6, 1, -4, -9, -12, -14, -15, -14, -11, -7, -2, 3, 8, 11, 14, 15, 14, 12, 8, 4, -2, -6, -11, -13, -15, -15, -13, -9, -5, -8, -14, -15, -12, -6, 3, 11, 15, 15, 15,
 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, -8, -14, -15, -12, -6, 3, 11, 15, 14, 14, 13, 12, 10, 9, 7, 5, 3,
 1, -1, -3, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -14, -14, -13, -11, -10, -8, -6, -4, -2, -8, -14, -15, -12, -6, 3, 10, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10,
 -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, -8, -13, -15, -12, -5, 3, 10, 15, 13, 10, 6, 1, -4, -9, -12, -14, -15, -14,
 -11, -7, -2, 3, 8, 11, 14, 15, 14, 12, 8, 3, -1, -6, -10, -13, -15, -14, -13, -9, -5, 12, 15, 6, -8, -15, -11, 3, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11,
 10, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 12, 15, 6, -8, -15, -11, 3, 14, 13, 13, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -12, -13, -13,
 -14, -14, -13, -12, -12, -10, -9, -8, -6, -4, -2, 12, 15, 6, -8, -15, -10, 3, 14, 13, 11, 9, 6, 3, 0, -3, -6, -9, -11, -13, -14, -14, -13, -11, -9, -6, -3, 0, 3, 6, 9,
 11, 13, 14, 14, 13, 11, 9, 6, 3, 12, 15, 6, -8, -15, -10, 3, 13, 12, 9, 5, 1, -4, -8, -11, -13, -14, -12, -10, -6, -2, 3, 7, 10, 13, 14, 13, 11, 8, 3, -1, -6, -10, -12,
 -14, -13, -12, -9, -5, -15, -8, 11, 14, -3, -15, -6, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, -15,
 -8, 11, 14, -3, -15, -6, 12, 12, 12, 11, 10, 9, 7, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -10, -11, -12, -12, -12, -12, -12, -11, -11, -9, -8, -7, -5, -4, -2, -15, -8, 10,
 14, -3, -15, -6, 12, 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6, 3, -15, -8, 10, 13, -3, -15, -5,
 12, 11, 8, 5, 1, -4, -7, -10, -12, -12, -11, -9, -6, -2, 2, 6, 9, 12, 12, 12, 10, 7, 3, -1, -5, -9, -11, -12, -12, -11, -8, -4, 1,15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -2,
 -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -1, -1, 0, 15, 15, 14, 12, 10, 8, 6, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3,
 3, 3, 3, 2, 2, 1, 1, 15, 15, 13, 12, 10, 8, 6, 3, 3, 2, 1, 0, -1, -2, -2, -3, -3, -3, -2, -1, 0, 1, 2, 2, 3, 3, 3, 2, 2, 1, 0, -1, -2, -3, -3, -3, -3, -2, -1, -15, -11,
 -3, 6, 12, 15, 14, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, -15, -11, -3, 6, 12, 15, 14, 8, 8, 8, 7, 6, 6, 5, 4,
 3, 2, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -8, -8, -8, -8, -7, -7, -6, -5, -4, -3, -2, -1, -15, -10, -3, 6, 12, 15, 14, 8, 8, 7, 5, 4, 2, 0, -2, -4, -5, -7, -8, -8,
 -8, -8, -7, -5, -4, -2, 0, 2, 4, 5, 7, 8, 8, 8, 8, 7, 5, 4, 2, -15, -10, -3, 6, 12, 15, 13, 8, 7, 5, 3, 0, -2, -5, -7, -8, -8, -7, -6, -4, -1, 2, 4, 6, 8, 8, 8, 6, 4, 2,
 -1, -3, -6, -7, -8, -8, -7, -5, -3, 14, 3, -11, -15, -8, 6, 15, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2,
 1, 1, 14, 3, -11, -15, -8, 6, 15, 12, 12, 12, 11, 10, 9, 7, 6, 4, 2, 1, -1, -3, -5, -6, -8, -9, -10, -11, -12, -12, -12, -12, -12, -11, -11, -9, -8, -7, -5, -4, -2, 14,
 3, -10, -15, -8, 6, 15, 12, 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6, 3, 13, 3, -10, -15, -8,
 6, 15, 12, 11, 8, 5, 1, -4, -7, -10, -12, -12, -11, -9, -6, -2, 2, 6, 9, 12, 12, 12, 10, 7, 3, -1, -5, -9, -11, -12, -12, -11, -8, -4, -12, 6, 15, 3, -14, -11, 8, 15,
 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 1, -12, 6, 15, 3, -14, -11, 8, 15, 14, 14, 13, 12, 10, 9, 7,
 5, 3, 1, -1, -3, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -14, -14, -13, -11, -10, -8, -6, -4, -2, -12, 6, 15, 3, -14, -10, 8, 15, 14, 12, 10, 7, 3, 0, -3, -7,
 -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, -12, 6, 15, 3, -13, -10, 8, 15, 13, 10, 6, 1, -4, -9, -12, -14, -15, -14,
 -11, -7, -2, 3, 8, 11, 14, 15, 14, 12, 8, 3, -1, -6, -10, -13, -15, -14, -13, -9, -5, 0,1,0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 11, 11, 11,
 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 0, 1, 1, 2, 2, 2, 2, 2, 4, 6, 8, 10, 11, 13, 14, 14, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 4, 2, -1, -3, -5, -7,
 -9, -11, -12, -13, -14, -15, -15, 1, 1, 2, 2, 3, 3, 3, 4, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 4, 0, -4, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -4, 0, 4, 7,
 10, 12, 14, 15, 1, 2, 3, 4, 4, 5, 5, 5, 9, 13, 15, 15, 13, 11, 6, 2, -4, -8, -12, -14, -15, -14, -11, -8, -3, 2, 7, 11, 14, 15, 14, 12, 9, 4, -1, -6, -10, -13, -15, 0,
 -1, -1, -1, 0, 0, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, -1, -2, -2, -2, -1, 0, 2, 2, 4, 6,
 8, 10, 11, 13, 14, 14, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 3, 1, -1, -3, -5, -7, -9, -10, -12, -13, -14, -14, -15, -2, -3, -4, -3, -1, 1, 2, 3, 7, 10, 12, 14, 15, 15,
 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, -3, -5, -5, -4, -2, 1, 4, 5, 9, 13, 14, 15, 13, 10, 6, 1, -3, -8,
 -12, -14, -15, -14, -11, -8, -3, 2, 7, 11, 14, 15, 14, 12, 9, 4, -1, -6, -10, -13, -15, 1, 1, 0, 0, -1, 0, 0, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 10,
 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 2, 2, 1, -1, -2, -2, 0, 2, 4, 6, 8, 9, 10, 12, 12, 13, 14, 14, 13, 13, 12, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -6,
 -8, -10, -11, -12, -13, -13, -14, 3, 3, 1, -2, -4, -2, 1, 3, 6, 9, 11, 13, 14, 14, 13, 11, 9, 6, 3, 0, -3, -6, -9, -11, -13, -14, -14, -13, -11, -9, -6, -3, 0, 3, 6, 9,
 11, 13, 14, 4, 5, 2, -3, -5, -4, 1, 5, 9, 12, 13, 14, 12, 10, 6, 1, -3, -8, -11, -13, -14, -13, -10, -7, -3, 2, 6, 10, 12, 14, 13, 11, 8, 4, -1, -5, -9, -12, -13, -1, 0,
 0, 1, 0, -1, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, -2, -1, 2, 2, 0, -2, -1, 2, 4, 5, 7, 8, 9,
 11, 11, 12, 12, 12, 12, 12, 11, 10, 9, 8, 6, 5, 3, 1, -1, -2, -4, -6, -7, -9, -10, -11, -12, -12, -12, -3, -2, 2, 3, -1, -4, -1, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6,
 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, -5, -3, 4, 5, -1, -5, -2, 4, 8, 11, 12, 12, 11, 9, 5, 1, -3, -7, -10, -12, -12, -12,
 -9, -6, -2, 2, 6, 9, 11, 12, 12, 10, 7, 4, -1, -5, -8, -11, -12, 1,1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, 4, 3, 3, 3, 2, 2, 1, 1, 1,
 2, 2, 3, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 5, 5, 5, 4, 4, 3, 2, 1, 2, 3, 3, 3, 3, 2, 1, 0, -1, -2, -2, -3, -3,
 -3, -2, -2, -1, 0, 1, 2, 3, 3, 3, 2, 2, 1, 0, -1, -2, -3, -3, -1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8,
 8, 8, 8, 8, -2, -2, 0, 1, 2, 2, 2, 1, 2, 3, 4, 5, 6, 7, 7, 8, 8, 8, 8, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -6, -7, -8, -8, -8, -3, -2, -1, 1, 3, 4, 3, 2,
 4, 5, 7, 8, 8, 8, 8, 7, 5, 4, 2, 0, -2, -4, -5, -7, -8, -8, -8, -8, -7, -5, -4, -2, 0, 2, 4, 5, 7, 8, 8, -5, -4, -1, 2, 4, 5, 5, 3, 5, 7, 8, 8, 7, 6, 3, 1, -2, -4, -6,
 -8, -8, -8, -6, -4, -2, 1, 4, 6, 7, 8, 8, 7, 5, 2, 0, -3, -5, -7, -8, 1, 0, 0, -1, 0, 0, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 11,
 12, 12, 12, 12, 12, 12, 12, 12, 2, 0, -2, -2, -1, 1, 2, 2, 4, 5, 7, 8, 9, 11, 11, 12, 12, 12, 12, 12, 11, 10, 9, 8, 6, 5, 3, 1, -1, -2, -4, -6, -7, -9, -10, -11, -12,
 -12, -12, 3, 1, -2, -4, -2, 1, 3, 3, 6, 8, 10, 12, 12, 12, 12, 10, 8, 6, 3, 0, -3, -6, -8, -10, -12, -12, -12, -12, -10, -8, -6, -3, 0, 3, 6, 8, 10, 12, 12, 5, 1, -4,
 -5, -3, 2, 5, 4, 8, 11, 12, 12, 11, 9, 5, 1, -3, -7, -10, -12, -12, -12, -9, -6, -2, 2, 6, 9, 11, 12, 12, 10, 7, 4, -1, -5, -8, -11, -12, -1, 0, 1, 0, -1, 0, 0, 1, 1, 2,
 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, -2, 1, 2, 0, -2, -2, 1, 2, 4, 6, 8, 10, 11, 13, 14, 14, 15, 15,
 15, 14, 13, 12, 11, 9, 8, 6, 3, 1, -1, -3, -5, -7, -9, -10, -12, -13, -14, -14, -15, -3, 1, 4, 1, -3, -2, 2, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10,
 -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, -4, 2, 5, 1, -5, -4, 3, 5, 9, 13, 14, 15, 13, 10, 6, 1, -3, -8, -12, -14, -15, -14, -11, -8, -3, 2,
 7, 11, 14, 15, 14, 12, 9, 4, -1, -6, -10, -13, -15, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7,
 -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11,
 -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11, -10,
 -8, -5, -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 15, 15, 14,
 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14,
 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -15, -14, -12, -10, -8, -4,
 -2, 2, 4, 8, 10, 12, 14, 15, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 11, 9, 6,
 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14,
 -14, -15, -15, -16, -16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -16, -16, -15, -14,
 -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14,
 -11, -7, -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -15, -10, -2, 8, 14, 16, 12, 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 14,
 14, 13, 12, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -15, -10, -2, 8, 14, 16, 12, 4, -4, -12, -16, -14, -8,
 2, 10, 15, 15, 14, 12, 10, 7, 4, 2, -2, -4, -7, -10, -12, -14, -15, -15, -15, -15, -14, -12, -10, -7, -4, -2, 2, 4, 7, 10, 12, 14, 15, 15, -15, -10, -2, 8, 14, 16, 12,
 4, -4, -12, -16, -14, -8, 2, 10, 15, 14, 11, 8, 4, -1, -5, -9, -12, -14, -15, -15, -13, -10, -6, -2, 2, 6, 10, 13, 15, 15, 14, 12, 9, 5, 1, -4, -8, -11, -14, -15, 0,1,
11, 11, 11, 11, 10, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 15, 15, 15, 15, 14, 14,
 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 15, 15, 15, 15, 14, 14, 13,
 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 15, 15, 15, 15, 14, 14, 13, 12,
 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -11, -10, -8, -6, -3, 0, 3, 6, 9, 10,
 11, 11, 10, 9, 6, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12,
 8, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12,
 8, 4, 4, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 4, 4, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12,
 8, 4, 4, 3, 2, 1, 0, -2, -3, -4, -4, -4, -4, -4, -3, -2, -1, 1, 2, 3, 4, 4, 4, 4, 4, 3, 2, 0, -1, -2, -3, -4, -4, 11, 9, 4, -1, -6, -10, -11, -10, -7, -2, 3, 8, 11, 11,
 9, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 7,
 7, 6, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -7, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7,
 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -7, -7, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 7, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13,
 7, 6, 5, 4, 2, 0, -2, -4, -6, -7, -7, -7, -6, -5, -3, -1, 1, 3, 5, 6, 7, 7, 7, 6, 4, 2, 0, -2, -4, -5, -6, -7, -11, -6, 1, 8, 11, 10, 4, -3, -9, -11, -8, -2, 5, 10, 11,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 10, 9,
 9, 9, 8, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10,
 9, 9, 8, 6, 5, 3, 1, -1, -3, -5, -6, -8, -9, -9, -10, -10, -9, -9, -8, -6, -5, -3, -1, 1, 3, 5, 6, 8, 9, 9, 10, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14,
 15, 10, 9, 7, 5, 2, 0, -3, -6, -8, -9, -10, -9, -8, -7, -4, -1, 1, 4, 7, 8, 9, 10, 9, 8, 6, 3, 0, -2, -5, -7, -9, -10, 0,1,1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 2, 1, 1, 3, 4, 6, 7, 8, 10, 11, 12, 13, 14, 14, 15,
 15, 15, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11, -13, -14, -15, -15, -15, -15, -15, -14, -13, -12, -10, -9, -7, -4, -2, 1, 3, 4, 6, 7, 8, 10, 11,
 12, 13, 13, 14, 15, 15, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 1, 3, 4, 6,
 7, 8, 9, 11, 12, 13, 13, 14, 15, 15, 15, 15, 13, 10, 6, 1, -4, -9, -13, -15, -15, -14, -11, -7, -2, 3, 8, 12, 14, 15, 15, 12, 9, 4, -2, -7, -11, -14, -15, -15, -13, -10,
 -5, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4,
 3, 2, 2, 1, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11, -13, -14, -15, -15, -15, -15,
 -15, -14, -13, -12, -10, -9, -7, -4, -2, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14,
 -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, -4, -8, -12, -14, -15, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 13, 10, 6, 1, -4, -9, -13, -15, -15,
 -14, -11, -7, -2, 3, 8, 12, 14, 15, 15, 12, 9, 4, -2, -7, -11, -14, -15, -15, -13, -10, -5, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 4, 3, 2, 2, 1, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15, -14, -8, -1, 6, 12, 15, 15,
 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11, -13, -14, -14, -15, -15, -15, -15, -14, -13, -12, -10, -8, -6, -4, -2, 7, 13, 15, 15, 11, 4, -3, -10, -14, -15,
 -13, -8, -1, 6, 12, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 7, 13, 15, 15, 11, 4,
 -3, -9, -14, -15, -13, -8, -1, 6, 12, 15, 13, 10, 6, 1, -4, -9, -13, -15, -15, -14, -11, -7, -2, 3, 8, 12, 14, 15, 14, 12, 8, 4, -2, -6, -11, -14, -15, -15, -13, -10,
 -5, -10, -15, -14, -7, 3, 12, 15, 13, 4, -6, -14, -15, -11, -1, 8, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3,
 3, 2, 1, 1, -10, -15, -14, -7, 3, 12, 15, 13, 4, -6, -14, -15, -11, -1, 8, 15, 14, 14, 13, 12, 10, 9, 7, 5, 3, 1, -1, -3, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15,
 -14, -14, -13, -11, -10, -8, -6, -4, -2, -10, -15, -14, -7, 3, 12, 15, 13, 4, -6, -13, -15, -11, -1, 8, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14,
 -12, -10, -7, -3, 0, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, -9, -15, -14, -7, 3, 12, 15, 13, 4, -6, -13, -15, -11, -1, 8, 15, 13, 10, 6, 1, -4, -9, -12, -14, -15,
 -14, -11, -7, -2, 3, 8, 11, 14, 15, 14, 12, 8, 3, -1, -6, -10, -13, -15, -14, -13, -9, -5, 1,15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 15, 15, 15, 15, 14, 13, 13, 12, 11, 9, 8, 7, 6, 4, 3, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0,
 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3,
 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4,
 -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, -1, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4,
 -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 4, 4, 4, 4, 3, 2, 1, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 3, 2, 0, -1, -3, -4, -4, -4, -4, -3, -2, -1, 1,
 2, 3, 4, 4, 4, 4, 2, 1, 0, -2, -3, -4, -4, -4, -4, -3, -1, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5,
 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5,
 -6, -6, -7, -7, -7, -7, -7, -6, -6, -5, -5, -4, -3, -2, -1, 15, 12, 6, -1, -8, -13, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 6, 5, 3, 2, 0, -2, -3, -5, -6, -7, -7,
 -7, -7, -6, -5, -3, -2, 0, 2, 3, 5, 6, 7, 7, 7, 7, 6, 5, 3, 2, 15, 12, 6, -1, -8, -13, -15, -14, -9, -3, 4, 11, 15, 15, 13, 7, 6, 5, 3, 0, -2, -4, -6, -7, -7, -6, -5,
 -3, -1, 1, 4, 5, 7, 7, 7, 6, 4, 2, -1, -3, -5, -6, -7, -7, -6, -4, -2, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8,
 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 9, 9, 8, 8, 7, 6, 5, 3, 2, 0, -1, -2, -4,
 -5, -6, -7, -8, -9, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -4, -3, -1, -15, -8, 1, 11, 15, 13, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6,
 -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, -15, -8, 1, 11, 15, 13, 6, -4, -13, -15, -12, -3, 7, 14, 15, 9, 8, 6, 4, 0, -3, -6, -8,
 -9, -10, -9, -7, -5, -1, 2, 5, 7, 9, 10, 9, 8, 5, 2, -1, -4, -7, -9, -10, -9, -8, -6, -3, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 6, 7, 7,
 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 7, 9, 10, 12, 13, 14, 15, 15, 15,
 15, 15, 14, 13, 11, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10,
 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 10, 13, 15, 15, 14, 11, 7, 2,
 -4, -9, -12, -15, -15, -14, -12, -8, -3, 2, 7, 11, 14, 15, 15, 13, 9, 4, -1, -6, -10, -13, -15, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 3, 4, 4, 5,
 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, -1, -1, -2, -2, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 4, 7, 9, 10, 12,
 13, 14, 15, 15, 15, 15, 15, 14, 13, 11, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, -1, -2, -3, -3, -4, -4, -3, -3, -2, -1, 0, 1, 2, 3, 3, 4, 7,
 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, -1, -3, -4, -5, -5, -5, -5, -4, -3, -1, 0, 1,
 3, 4, 5, 5, 10, 13, 15, 15, 14, 11, 7, 2, -4, -9, -12, -15, -15, -14, -12, -8, -3, 2, 7, 11, 14, 15, 15, 13, 9, 4, -1, -6, -10, -13, -15, 0, 1, 1, 1, 0, 0, 0, 0, -1, -1,
 -1, 0, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 1, 2, 2, 2, 2, 1, 0, -1, -2, -2, -2,
 -1, 0, 1, 2, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 15, 15, 14, 14, 13, 11, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, 2, 3, 4, 3, 3, 1, -1, -2, -3,
 -4, -3, -2, 0, 1, 3, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 2, 4, 5, 5, 4, 1,
 -1, -3, -5, -5, -5, -3, 0, 2, 4, 5, 10, 13, 15, 15, 14, 11, 6, 2, -4, -8, -12, -14, -15, -14, -12, -8, -3, 2, 7, 11, 14, 15, 15, 13, 9, 4, -1, -6, -10, -13, -15, 0, -1,
 -1, 0, 0, 1, 1, 1, 0, 0, -1, -1, 0, 0, 0, 1, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, -1, -2, -2,
 -1, 0, 2, 2, 2, 1, -1, -2, -2, -2, 0, 1, 2, 4, 6, 8, 10, 11, 13, 14, 14, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 3, 1, -1, -3, -5, -7, -9, -10, -12, -13, -14, -14, -15, -2,
 -4, -3, -2, 1, 3, 4, 3, 1, -1, -3, -4, -3, 0, 2, 3, 7, 10, 12, 14, 15, 15, 14, 12, 10, 7, 3, 0, -3, -7, -10, -12, -14, -15, -15, -14, -12, -10, -7, -3, 0, 3, 7, 10, 12,
 14, 15, -3, -5, -5, -2, 1, 4, 5, 4, 1, -2, -5, -5, -4, 0, 3, 5, 9, 13, 14, 15, 13, 10, 6, 1, -3, -8, -12, -14, -15, -14, -11, -8, -3, 2, 7, 11, 14, 15, 14, 12, 9, 4, -1,
 -6, -10, -13, -15, 1,1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 4, 4, 4, 3, 3, 3, 3, 3, 3,
 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1,
 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1,
 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4, -3, -3, -2, -1, 0, 1, 2, 3, 3, 4, 4, 3, 3, 2, 1, 2, 3, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3,
 -4, -4, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 4, -5, -5, -4, -3, -1, 0, 1, 3, 4, 5, 5, 5, 5, 4, 3, 1, 3, 4, 4, 4, 4, 3, 2, 0, -1, -2, -4, -4, -4, -4, -3, -2, -1,
 1, 2, 3, 4, 4, 4, 4, 3, 1, 0, -2, -3, -4, -4, 1, 1, 0, 0, 0, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7,
 7, 7, 7, 7, 7, 7, 7, 2, 2, 1, 0, -1, -2, -2, -2, -1, 0, 1, 2, 2, 2, 2, 1, 2, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 6, 6, 5, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -6, -7,
 -7, -7, 4, 3, 1, 0, -2, -3, -4, -3, -2, -1, 1, 3, 3, 4, 3, 2, 3, 5, 6, 7, 7, 7, 7, 6, 5, 3, 2, 0, -2, -3, -5, -6, -7, -7, -7, -7, -6, -5, -3, -2, 0, 2, 3, 5, 6, 7, 7, 5,
 4, 2, 0, -3, -5, -5, -5, -3, -1, 1, 4, 5, 5, 4, 2, 4, 6, 7, 7, 6, 5, 3, 1, -2, -4, -6, -7, -7, -7, -5, -4, -1, 1, 3, 5, 6, 7, 7, 6, 4, 2, 0, -3, -5, -6, -7, -1, 0, 0, 0,
 1, 1, 0, 0, -1, -1, -1, 0, 0, 1, 1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, -2, -1, 0, 2, 2, 2, 1, -1, -2, -2,
 -2, 0, 1, 2, 2, 1, 3, 4, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -5, -6, -7, -8, -8, -9, -9, -10, -3, -2, 0, 3, 4, 3, 1, -1, -3, -4, -3, -1,
 2, 3, 4, 2, 5, 6, 8, 9, 10, 10, 9, 8, 6, 5, 2, 0, -2, -5, -6, -8, -9, -10, -10, -9, -8, -6, -5, -2, 0, 2, 5, 6, 8, 9, 10, -5, -3, 0, 4, 5, 5, 2, -1, -4, -5, -4, -1, 2,
 5, 5, 3, 6, 8, 9, 10, 9, 7, 4, 1, -2, -5, -8, -9, -10, -9, -7, -5, -2, 1, 5, 7, 9, 10, 9, 8, 6, 3, 0, -4, -6, -8, -9, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5,
 -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11, -10, -8, -5, -2, 2, 5, 8, 10, 11, 11,
 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8,
 -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12,
 -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 15,
 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -16, -15, -15, -14, -14, -13, -12, -11,
 -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15,
 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10,
 -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9,
 -11, -12, -13, -14, -14, -15, -15, -16, -16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14,
 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2,
 -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7,
 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5, -8, -10, -11, -11, -11, -9, -7, -4, 0, 3, 6,
 8, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -14, -12, -8, -4, 1, 5, 9,
 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5,
 -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1,
 4, 8, 12, 14, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -14, -12, -8, -4,
 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7,
 -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 0,0,0,1,1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9,
 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 15, 15, 14, 13, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -12, -13, -14, -15, -15, -16, -16,
 -15, -14, -14, -12, -11, -9, -7, -5, -2, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 15, 13, 11,
 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 1, 2, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11,
 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 11, 6, 1, -5, -9, -13, -15, -16, -14, -12, -7, -2, 3, 8, 12, 15, 16, 15, 13, 9, 4, -2, -7, -11, -14, -16,
 -15, -14, -10, -5, -2, -5, -7, -9, -11, -12, -14, -14, -15, -16, -16, -15, -15, -14, -13, -12, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 13, 14, 15, 15, 16, 16, 16, 15,
 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, -2, -5, -7, -9, -11, -12, -14, -14, -15, -16, -16, -15, -15, -14, -13,
 -12, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 13, 14, 15, 15, 16, 15, 15, 14, 13, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -12, -13, -14, -15, -15, -16, -16, -15, -14,
 -14, -12, -11, -9, -7, -5, -2, -2, -5, -7, -9, -10, -12, -13, -14, -15, -15, -16, -15, -15, -14, -13, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 13, 14, 15, 15, 16,
 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, -2, -4, -7, -9, -10, -12, -13, -14, -15, -15,
 -15, -15, -15, -14, -13, -11, -10, -8, -6, -4, -2, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 14, 11, 6, 1, -5, -9, -13, -15, -16, -14, -12, -7, -2, 3, 8, 12, 15, 16, 15,
 13, 9, 4, -2, -7, -11, -14, -16, -15, -14, -10, -5, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11,
 13, 15, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 0,
 -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 15, 15, 14, 13, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11, -13, -14, -15, -15, -16,
 -15, -15, -14, -13, -12, -10, -9, -7, -5, -2, 4, 7, 10, 13, 15, 15, 15, 15, 13, 10, 7, 4, 0, -4, -7, -10, -13, -15, -15, -15, -15, -13, -10, -7, -4, 0, 4, 7, 10, 13, 15,
 15, 15, 13, 10, 7, 4, 0, -4, -7, -10, -13, -15, -15, -15, -15, -13, -10, -7, -4, 0, 4, 7, 10, 13, 15, 15, 15, 15, 13, 10, 7, 4, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7,
 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 14, 10, 6, 1, -5, -9, -13, -15, -16, -14, -11, -7, -2, 3, 8, 12, 15, 16, 15, 13,
 9, 4, -2, -7, -11, -14, -15, -15, -13, -10, -5, -5, -10, -14, -15, -16, -14, -11, -7, -2, 4, 9, 13, 15, 16, 15, 12, 8, 3, -2, -7, -12, -14, -16, -15, -13, -9, -5, 1, 6,
 11, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 2, 1, -5, -10, -14, -15, -16, -14, -11, -7, -2, 4,
 9, 13, 15, 16, 15, 12, 8, 3, -2, -7, -12, -14, -16, -15, -13, -9, -5, 1, 6, 11, 14, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11, -13, -14, -15, -15,
 -15, -15, -15, -14, -13, -12, -10, -9, -7, -4, -2, -5, -10, -13, -15, -15, -14, -11, -7, -2, 4, 9, 13, 15, 16, 15, 12, 8, 3, -2, -7, -11, -14, -16, -15, -13, -9, -5, 1,
 6, 10, 14, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, -5, -10, -13, -15, -15, -14,
 -11, -7, -2, 4, 9, 12, 15, 15, 14, 12, 8, 3, -2, -7, -11, -14, -15, -15, -13, -9, -4, 1, 6, 10, 13, 15, 13, 10, 6, 1, -4, -9, -13, -15, -15, -14, -11, -7, -2, 3, 8, 12,
 14, 15, 15, 12, 9, 4, -2, -7, -11, -14, -15, -15, -13, -10, -5, 1,16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4,
 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12,
 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 16, 16, 15,
 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0,
 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16,
 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -16, -15, -15, -14, -13, -11, -9, -7,
 -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16, 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -1, -1, -1, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 11, 13, 14, 15, 15, 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 2, 2, 2, 2, 1,
 1, 0, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 11, 13,
 14, 15, 15, 15, 15, 15, 14, 13, 12, 10, 9, 7, 4, 2, 2, 2, 1, 0, -1, -1, -2, -2, -2, -2, -2, -1, 0, 0, 1, 2, 2, 2, 2, 2, 1, 1, 0, -1, -2, -2, -2, -2, -2, -1, -1, 16, 15,
 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7,
 4, 4, 3, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 15, 15, 13, 10, 7, 4, 0, -4, -7, -10, -13, -15, -15,
 -15, -15, -13, -10, -7, -4, 0, 4, 7, 10, 13, 15, 15, 15, 15, 13, 10, 7, 4, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -4, -4, -3, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 4, 4, 3,
 3, 2, 2, 1, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15, -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 3, 2, 1, 0, -1, -2, -3, -4,
 -4, -3, -3, -2, -1, 1, 2, 3, 3, 4, 3, 3, 2, 1, 0, -2, -3, -3, -4, -4, -3, -2, -1, -15, -14, -11, -6, -1, 5, 9, 13, 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13,
 -9, -4, 2, 7, 11, 14, 16, 15, 14, 10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, -15, -14, -11, -6, -1, 5, 9, 13,
 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 16, 15, 14, 10, 5, 5, 5, 5, 4, 4, 3, 2, 2, 1, 0, -1, -1, -2, -3, -3, -4, -4, -5, -5, -5, -5,
 -5, -5, -5, -5, -4, -4, -3, -2, -2, -1, -15, -14, -10, -6, -1, 5, 9, 13, 15, 16, 14, 11, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 15, 15, 13, 10, 5,
 5, 4, 4, 2, 1, 0, -1, -2, -4, -4, -5, -5, -5, -5, -4, -4, -2, -1, 0, 1, 2, 4, 4, 5, 5, 5, 5, 4, 4, 2, 1, -15, -13, -10, -6, -1, 4, 9, 13, 15, 15, 14, 11, 7, 2, -3, -8,
 -12, -14, -15, -15, -12, -9, -4, 2, 7, 11, 14, 15, 15, 13, 10, 5, 5, 4, 2, 0, -2, -3, -4, -5, -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -5,
 -3, -2, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13,
 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 7, 9, 11, 12, 14,
 14, 15, 16, 16, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 0, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 10, 14, 15, 16, 14, 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 12, 14, 16,
 15, 13, 9, 5, -1, -6, -11, -14, -15, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7,
 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0,
 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 5, 7, 9, 11, 12, 14, 14, 15, 16, 16, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, -1, -1, -2,
 -2, -2, -3, -3, -3, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 3, 4, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13,
 -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, -1, -2, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -1, -1, 0, 1, 2, 2, 3, 4, 4,
 5, 5, 5, 5, 10, 14, 15, 16, 14, 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 12, 14, 16, 15, 13, 9, 5, -1, -6, -11, -14, -15, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0,
 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15,
 15, 15, 16, 16, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, 5, 7, 9, 10, 12, 13, 14, 15, 15, 16, 15, 15,
 14, 13, 11, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, 1, 2, 2, 3, 3, 4, 4, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -4, -4, -3, -3, -2, -2, -1, 0,
 1, 2, 2, 3, 3, 4, 7, 10, 13, 15, 15, 15, 15, 13, 10, 7, 4, 0, -4, -7, -10, -13, -15, -15, -15, -15, -13, -10, -7, -4, 0, 4, 7, 10, 13, 15, 15, 1, 2, 4, 4, 5, 5, 5, 5, 4,
 4, 2, 1, 0, -1, -2, -4, -4, -5, -5, -5, -5, -4, -4, -2, -1, 0, 1, 2, 4, 4, 5, 5, 10, 13, 15, 15, 14, 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 11, 14, 16,
 15, 13, 9, 5, -1, -6, -10, -14, -15, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 6, 7, 7,
 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, -1, -1, -2, -2, -2, -2, -2, -1, 0, 1, 1, 2, 2, 2, 2, 2, 1, 0, 0, -1, -2, -2, -2, -2,
 -2, -1, -1, 0, 1, 2, 2, 2, 4, 7, 9, 10, 12, 13, 14, 15, 15, 15, 15, 15, 14, 13, 11, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, -1, -2, -3, -4, -4,
 -3, -3, -2, 0, 1, 2, 3, 3, 4, 3, 3, 2, 1, -1, -2, -3, -3, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 7, 10, 13, 14, 15, 15, 14, 13, 10, 7, 4, 0, -4, -7, -10, -13, -14, -15, -15,
 -14, -13, -10, -7, -4, 0, 4, 7, 10, 13, 14, 15, -2, -3, -5, -5, -5, -5, -4, -2, -1, 1, 3, 4, 5, 5, 5, 4, 3, 1, -1, -2, -4, -5, -5, -5, -4, -3, -2, 0, 2, 4, 5, 5, 10, 13,
 15, 15, 14, 11, 7, 2, -4, -9, -12, -15, -15, -14, -12, -8, -3, 2, 7, 11, 14, 15, 15, 13, 9, 4, -1, -6, -10, -13, -15, 1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 4,
 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0,
 0, 0, 0, 0, 0, 1, 1, 1, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -4, -4, -3, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2,
 2, 3, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, -5, -5, -5,
 -5, -4, -4, -3, -2, -2, -1, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 2, 1, 1, 2, 2, 2, 2, 2, 1, 0, -1, -1, -2, -2, -2, -2, -2, -1, 0, 0, 1, 2, 2, 2,
 2, 2, 1, 1, 0, -1, -2, -2, -2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 2, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1,
 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -3, -4, -4, 4, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -4, -4, -3, -3, -2, -2, -1,
 0, 1, 2, 2, 3, 3, 4, 4, 3, 3, 2, 2, 1, 2, 2, 3, 3, 4, 4, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -4, -4, -3, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 4, 5, 5, 4, 4, 2, 1, 0, -1,
 -2, -4, -4, -5, -5, -5, -5, -4, -4, -2, -1, 0, 1, 2, 4, 4, 5, 5, 5, 5, 4, 4, 2, 1, 2, 3, 4, 4, 3, 3, 2, 0, -1, -2, -3, -3, -4, -3, -3, -2, -1, 1, 2, 3, 3, 4, 4, 3, 2, 1,
 0, -1, -2, -3, -4, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4,
 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -2, -2, -2, -1, 0, 1, 1, 2, 2, 2, 2, 2, 1, 0, 0, -1, -2, -2, -2, -2, -2, -1, -1, 0, 1, 2, 2, 2, 2, 2, 1, 1, 2, 2, 3, 4, 4,
 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 3, 3, 2, 1, -1, -2, -3, -3, -4, -3, -3, -2, -1,
 0, 2, 3, 3, 4, 4, 3, 2, 1, 2, 4, 4, 5, 5, 5, 5, 4, 4, 2, 1, 0, -1, -2, -4, -4, -5, -5, -5, -5, -4, -4, -2, -1, 0, 1, 2, 4, 4, 5, 5, -5, -5, -4, -2, 0, 2, 3, 4, 5, 5, 5,
 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2, 4, 5, 5, 5, 5, 3, 2, 3, 5, 5, 5, 5, 4, 2, 1, -1, -3, -4, -5, -5, -5, -4, -3, -1, 1, 2, 4, 5, 5, 5, 4, 3, 2, 0, -2, -4,
 -5, -5, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5,
 -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11,
 -10, -8, -5, -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6,
 -6, -5, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14,
 -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14,
 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6, -7, -8, -10, -11, -12, -13, -14, -15, -15, -16, -16, -16, -16,
 -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7,
 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12,
 -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5,
 -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 12, 8, 4, -1, -6, -10, -13,
 -15, -16, -16, -14, -11, -7, -2, 2, 7, 11, 14, 16, 16, 15, 13, 10, 6, 1, -4, -8, -12, -15, -16, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5,
 -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8,
 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4,
 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16,
 -16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13,
 -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15,
 -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 14, 12, 8, 4, -1,
 -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7,
 9, 10, 10, 11, 11, 11, 11, 11, 10, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9,
 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -15, -14, -12, -11,
 -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14,
 -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14,
 -14, -15, -15, -16, -16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4, -6, -8,
 -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12,
 -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16,
 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15,
 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 0,0,0,0,1,16, 16, 16, 16, 16, 16, 16,
 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5,
 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3,
 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 2, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14,
 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5,
 -4, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7,
 -6, -5, -4, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4,
 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, -16, -16, -15, -15, -15, -14, -14, -13, -13, -12, -11,
 -10, -10, -9, -8, -7, -6, -5, -3, -2, -1, 0, 1, 2, 3, 5, 6, 7, 8, 9, 10, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10,
 10, 9, 8, 7, 6, 5, 3, 2, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, -15, -15, -15, -15, -15, -14, -14,
 -13, -13, -12, -11, -10, -9, -9, -8, -7, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14,
 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 16, 16, 15, 14,
 14, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7, 9, 10,
 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7, 6, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 16,
 16, 15, 14, 14, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5,
 7, 9, 10, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7, 6, 4, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -1, -1, -1, -1, -1, 0, 16, 15, 15, 14, 13, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -14, -15, -15, -16, -16, -15, -15, -14, -13, -12, -11, -9,
 -8, -6, -4, -2, 0, 2, 3, 5, 7, 9, 10, 11, 13, 14, 15, 15, 15, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7, 6, 4, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -2, -2, -2, -1, -1,
 0, 0, 0, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 0, 15, 15, 15, 14, 13, 12, 11, 9, 8, 6, 4, 3, 1, -1, -3, -5, -7, -8, -10, -11, -12, -13, -14, -15, -15, -15, -15, -15, -15, -14,
 -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 3, 5, 7, 9, 10, 11, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 13, 12, 10, 9, 7, 6, 4, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -2, -1,
 -1, 0, 0, 1, 1, 2, 2, 2, 2, 1, 0, 0, -1, -1, -2, -2, -2, -2, -1, -1, -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 4, 6, 8, 11, 12, 14, 15, 16, 16, 16, 15, 14, 12, 10, 8,
 6, 3, 0, -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 16, 16, 15, 15, 14, 12, 10, 8, 5, 3, 3, 3, 3, 3,
 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 4, 6, 8, 11, 12, 14, 15, 16, 16, 16, 15, 14,
 12, 10, 8, 6, 3, 0, -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 16, 16, 15, 15, 14, 12, 10, 8, 5, 3,
 3, 2, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 3, 6, 8, 10,
 12, 14, 15, 15, 16, 15, 15, 14, 12, 10, 8, 6, 3, 0, -2, -5, -7, -10, -11, -13, -14, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 15,
 16, 15, 15, 13, 12, 10, 8, 5, 3, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 3, 3, 2, 2, 2, 1, 1, -15, -15, -14, -13, -11, -9,
 -7, -4, -2, 1, 3, 6, 8, 10, 12, 14, 15, 15, 15, 15, 15, 13, 12, 10, 8, 6, 3, 0, -2, -5, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4,
 6, 9, 11, 12, 14, 15, 15, 15, 15, 14, 13, 12, 10, 8, 5, 3, 2, 2, 1, 0, -1, -2, -2, -3, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 2, 2, 1, 1, 0, -1, -2, -2, -3, -3, -2, -2, -1,
 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 7,
 9, 11, 12, 14, 14, 15, 16, 16, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 7, 11, 13, 15, 16, 16, 15,
 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 10, 14, 15, 16, 14, 11, 7, 2, -4, -9, -13, -15,
 -16, -15, -12, -8, -3, 2, 7, 12, 14, 16, 15, 13, 9, 5, -1, -6, -11, -14, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 11,
 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 5, 7, 9, 11, 12, 14, 14, 15, 16, 16, 15,
 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4,
 -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 7, 11, 13, 15, 16, 16,
 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 10,
 14, 15, 16, 14, 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 12, 14, 16, 15, 13, 9, 5, -1, -6, -11, -14, -15, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4,
 5, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1,
 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 5, 7, 9, 11, 12, 14, 14,
 15, 16, 16, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1,
 0, 0, -1, -1, -1, -2, -2, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 7, 11, 13, 15, 16, 16,
 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 1, 1, 2, 2, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2,
 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 10, 14, 15, 16, 14,
 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 12, 14, 16, 15, 13, 9, 5, -1, -6, -11, -14, -15, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6,
 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1,
 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 5, 7, 9, 11, 12, 14, 14,
 15, 16, 16, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, -1, -3, -5, -7, -9, -11, -13, -14, -15, -15, -16, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -3, -3, -3, -2, -1, -1, 0,
 0, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 3, 4, 4, 7, 11, 13, 15,
 16, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, -1, -2, -3, -3, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3,
 -2, -1, 0, 1, 1, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -1, 0, 1, 2, 2, 3, 4, 4, 5, 5, 5, 10,
 14, 15, 16, 14, 11, 7, 2, -4, -9, -13, -15, -16, -15, -12, -8, -3, 2, 7, 12, 14, 16, 15, 13, 9, 5, -1, -6, -11, -14, -15, 1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4,
 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, -5,
 -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2,
 -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, 4, 4, 4, 3, 3, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4,
 -4, -4, -4, -3, -3, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 1, 1, 2, 2, 2, 2, 2, 2, 1, 1, 0, 0, 0,
 -1, -1, -2, -2, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5,
 -4, -4, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 2, 2, 1, 1, 1, 2, 2, 2, 2, 1, 1, 0, 0, -1, -2, -2, -2, -2, -1, -1, 0,
 0, 1, 1, 2, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3,
 3, 3, 3, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1,
 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -2, -2, -3, -3, -4, -4,
 -3, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3,
 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 2, 2, 2, 3, 3, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 3, -5, -5, -5, -4, -4, -3,
 -2, -2, -1, 0, 1, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 4, 3, 3, 2, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 4, 4, 5, 5, 5,
 5, 5, 5, 5, 4, 3, 3, 2, 1, 2, 2, 3, 3, 2, 2, 1, 0, -1, -1, -2, -2, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 3, 2, 2, 1, 0, -1, -2, -2, -3, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2,
 -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5,
 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 10, 8, 6, 3, 0, -4, -7, -9, -11, -11, -11, -10, -8, -5, -2, 2, 5, 8, 10, 11, 11, 11, 9, 7, 4, 1, -3, -6, -8, -10, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6,
 -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7,
 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15,
 -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4,
 -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6,
 -7, -8, -10, -11, -12, -13, -14, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14,
 -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2,
 -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14,
 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8,
 -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13,
 -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1,
 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15,
 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 12, 8, 4, -1, -6, -10, -13, -15, -16, -16, -14, -11, -7, -2, 2, 7, 11, 14, 16, 16, 15, 13, 10, 6, 1, -4, -8,
 -12, -15, -16, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6,
 -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8,
 -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14,
 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14,
 -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6,
 -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12,
 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10,
 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15,
 -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0,
 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15,
 -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5,
 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16,
 -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7,
 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16, -15, -14, -11, -7, -2, 2, 7, 11, 14, 15,
 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8,
 8, 9, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9,
 -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9,
 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16,
 -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16,
 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16,
 -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16,
 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -15, -14,
 -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14,
 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15,
 -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 15, 14, 12, 10, 8,
 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11,
 -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8,
 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12,
 -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 14, 12, 8, 4, -1, -5, -9, -13, -15, -16,
 -15, -14, -11, -7, -2, 2, 7, 11, 14, 15, 16, 15, 13, 9, 5, 1, -4, -8, -12, -14, -16, 0,0,0,0,1,16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12,
 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4,
 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13,
 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6,
 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9,
 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11,
 11, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2,
 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15,
 -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4,
 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9,
 -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7,
 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -15, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2,
 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15,
 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10,
 -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13,
 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9,
 9, 8, 8, 7, 7, 6, 6, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 16, 16, 16, 15, 15, 15,
 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15,
 -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9,
 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0,
 -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13,
 -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16,
 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, 0, 0, 0, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -10, -11, -12,
 -12, -13, -13, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -3, -2,
 -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 10, 10, 9, 8, 7,
 6, 6, 5, 4, 3, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12,
 11, 10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15,
 -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14,
 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0,
 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10,
 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 12, 12, 11, 10, 8, 7, 6, 5, 4, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16,
 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -16,
 -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14,
 13, 12, 12, 11, 10, 8, 7, 6, 5, 4, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12,
 -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, -16, -16, -15, -15, -15, -14, -13, -13, -12, -11, -10,
 -9, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 10, 10, 8, 7, 6, 5, 3, 2, 1, -1, -2,
 -3, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4,
 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 4, 3, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11,
 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15,
 -15, -15, -15, -15, -15, -15, -15, -15, -14, -13, -13, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 12, 12, 13, 14, 14, 15, 15, 15, 15, 15,
 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 4, 3, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0,
 0,0,1,1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2,
 -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -4,
 -4, -4, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0,
 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2,
 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3,
 -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3,
 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5,
 -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1,
 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3,
 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1,
 -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4,
 -4, -4, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 1, 1, 1, 1, 1, 1,
 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10,
 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10,
 -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7,
 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10, -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5,
 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -10, -11, -11, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, -15, -6, 6, 15,
 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -10,
 -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -6, 6, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5,
 -6, -7, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13,
 13, 14, 14, 15, 15, -15, -6, 6, 15, 14, 14, 13, 12, 10, 8, 7, 5, 2, 0, -2, -4, -6, -8, -10, -11, -12, -13, -14, -15, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3, -1, 1,
 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 15, 14, 13, 12, 11, 10, 8, 6, 4, 2, 0, -2, -5, -7, -8, -10, -12, -13, -14, -14, -15, 8, -8, -8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, -11, -11,
 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7,
 -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, -11, -11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5,
 -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11,
 11, 11, -11, -11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10, -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9,
 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -10, -11, -11, -4, 10, -10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -6, 15, -15, 6, 6, 6, 6, 6, 6, 6, 6,
 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -6, -6, -6, -6,
 -6, -6, -6, -6, -6, -6, -6, -6, 15, -15, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6,
 -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, -6, 15, -15, 6, 6, 6, 6, 5, 4, 4, 3, 2, 1, 0, -1, -2, -3, -3, -4, -5, -5,
 -6, -6, -6, -6, -6, -6, -6, -5, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 6, 6, 6, 6, 6, 6, 5, 5, 4, 3, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -6, -6, -6, 0,0,0,0,
1,15, 13, 10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 15, 13, 10, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 15, 13, 10, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1,
 -2, -2, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1,
 15, 13, 10, 5, 5, 5, 4, 4, 3, 2, 1, 1, 0, -1, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 1, 0,
 -1, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -2, -1, -13, 0, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11,
 11, 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, -13, 0, 13, 13, 13, 13, 13, 12,
 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13,
 -12, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, -13, 0, 13, 13, 13, 12, 12, 11, 10, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11,
 -12, -13, -13, -13, -13, -13, -12, -12, -11, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 4, 6, 7, 9, 10, 11, 11, 12, 13, 13, 13, 13, 13, 12, 12, 11, 10, 9, 8, 6, 5, 3, 2, -13,
 0, 13, 13, 13, 12, 11, 9, 8, 6, 4, 2, -1, -3, -5, -7, -9, -10, -11, -12, -13, -13, -13, -12, -11, -10, -9, -7, -5, -2, 0, 2, 4, 6, 8, 10, 11, 12, 13, 13, 13, 13, 12, 11,
 9, 8, 5, 3, 1, -1, -3, -5, -7, -9, -10, -12, -12, -13, -13, -13, -12, -11, -10, -8, -6, -4, -2, 10, -13, -5, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0,
 10, -13, -5, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14,
 -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 10, -13, -5, 15, 15, 14, 13, 13, 12, 10, 9, 8, 6, 4,
 2, 1, -1, -3, -5, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -7, -6, -4, -2, 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15,
 15, 14, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, 10, -13, -5, 15, 14, 13, 12, 11, 9, 7, 4, 2, -1, -3, -6, -8, -10, -12, -13, -14, -15, -15, -15, -14, -13, -11, -10, -8, -5,
 -3, 0, 2, 5, 7, 9, 11, 12, 14, 14, 15, 15, 14, 13, 12, 10, 9, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -14, -15, -15, -14, -14, -13, -11, -9, -7, -5, -2, -5, 13, -15, 10,
 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2,
 2, 2, 2, 1, 1, 1, 1, 0, 0, -5, 13, -15, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9,
 -9, -9, -9, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, -5, 13, -15, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 0,
 -1, -2, -3, -4, -5, -6, -7, -8, -8, -9, -9, -10, -10, -10, -9, -9, -9, -8, -7, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, 10, 9, 9, 9, 8, 7, 6,
 6, 5, 3, 2, 1, -5, 13, -15, 10, 9, 9, 8, 7, 6, 4, 3, 1, 0, -2, -4, -5, -6, -8, -8, -9, -10, -10, -10, -9, -8, -7, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 8, 9, 9, 10, 10, 9,
 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -3, -2, 0,1,0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7,
 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15,
 0, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2,
 -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, 1, 1, 2, 2, 4, 5, 7, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 14, 14, 13, 12, 11,
 10, 8, 7, 5, 3, 1, 0, -2, -4, -6, -7, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -5, -3, -1, 1, 2, 4, 6, 8, 9, 10, 12, 13, 13,
 14, 15, 15, 1, 2, 2, 2, 5, 7, 9, 11, 13, 14, 14, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -9, -10, -12, -13, -14, -15, -15, -14, -14, -12, -11, -9, -7, -5, -2, 0,
 3, 5, 8, 10, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -12, -13, -14, -15, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5,
 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13,
 -1, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2,
 -3, -4, -5, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -12, -12, -13, -13, -13, -13, -2, -2, 0, 2, 3, 5, 6, 8, 9, 10, 11, 12, 12, 13, 13, 13, 13, 13, 12, 11, 11, 10, 9,
 7, 6, 4, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -11, -12, -12, -13, -13, -13, -13, -13, -12, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 10, 11, 12, 12,
 13, 13, -2, -2, 0, 2, 4, 6, 8, 10, 11, 12, 13, 13, 13, 12, 12, 10, 9, 7, 5, 3, 1, -1, -3, -5, -8, -9, -11, -12, -13, -13, -13, -13, -12, -11, -10, -8, -6, -4, -2, 0, 2,
 5, 7, 9, 10, 11, 12, 13, 13, 13, 12, 11, 10, 9, 7, 5, 3, 1, -2, -4, -6, -8, -9, -11, -12, -13, -13, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5,
 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 1, 0, -1, 1, 1, 2, 3, 3, 4,
 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7, -7, -8,
 -8, -9, -9, -9, -9, -10, -10, -10, 2, -1, -2, 1, 2, 3, 5, 6, 6, 7, 8, 9, 9, 9, 10, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -7, -8, -9, -9,
 -9, -10, -10, -10, -9, -9, -8, -8, -7, -6, -5, -4, -3, -2, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, 2, -1, -2, 2, 3, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 5, 4, 2,
 1, -1, -3, -4, -6, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -3, -1, 0, 2, 3, 5, 6, 7, 8, 9, 10, 10, 10, 9, 8, 8, 6, 5, 4, 2, 0, -1, -3, -4, -6, -7, -8, -9, -9,
 -10, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -1, 1, -1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1,
 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -1, 2, -2, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2,
 1, 1, 0, -1, -1, -2, -2, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 4, 5, 5, 5, 5, -2, 2, -2, 1, 2, 2, 3,
 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -2, -1, 0, 1, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1,
 0, -1, -1, -2, -3, -4, -4, -5, -5, -5, 1,0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4,
 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, 2, 2, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5,
 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 4,
 5, 5, 5, 5, 2, 2, 2, 1, 2, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -2, -1, 0, 1, 2, 3, 3, 4, 4,
 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -5, -5, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9,
 9, 9, 9, 9, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, -1, 0, 1, 1, 2, 3, 4, 5, 5, 6, 7,
 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -9,
 -10, -11, -11, -12, -12, -12, -13, -13, -13, -13, -2, 0, 2, 2, 3, 5, 6, 8, 9, 10, 11, 12, 12, 13, 13, 13, 13, 13, 12, 11, 11, 10, 9, 7, 6, 4, 3, 1, 0, -2, -3, -5, -6,
 -8, -9, -10, -11, -12, -12, -13, -13, -13, -13, -13, -12, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 10, 11, 12, 12, 13, 13, -2, 0, 2, 2, 4, 6, 8, 10,
 11, 12, 13, 13, 13, 12, 12, 10, 9, 7, 5, 3, 1, -1, -3, -5, -8, -9, -11, -12, -13, -13, -13, -13, -12, -11, -10, -8, -6, -4, -2, 0, 2, 5, 7, 9, 10, 11, 12, 13, 13, 13,
 12, 11, 10, 9, 7, 5, 3, 1, -2, -4, -6, -8, -9, -11, -12, -13, -13, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10,
 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8,
 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10,
 -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, 1, -2, -1, 2, 4, 5, 7, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 5, 3, 1, 0, -2, -4,
 -6, -7, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -5, -3, -1, 1, 2, 4, 6, 8, 9, 10, 12, 13, 13, 14, 15, 15, 2, -2, -1, 2, 5,
 7, 9, 11, 13, 14, 14, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -9, -10, -12, -13, -14, -15, -15, -14, -14, -12, -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 11, 13, 14,
 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -12, -13, -14, -15, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6,
 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 0, 1, -1, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7,
 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9,
 -9, -10, -10, -10, -1, 2, -2, 1, 2, 3, 5, 6, 6, 7, 8, 9, 9, 9, 10, 10, 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -7, -8, -9, -9, -9, -10, -10, -10,
 -9, -9, -8, -8, -7, -6, -5, -4, -3, -2, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9, 10, 10, -1, 2, -2, 2, 3, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 5, 4, 2, 1, -1, -3, -4, -6,
 -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -3, -1, 0, 2, 3, 5, 6, 7, 8, 9, 10, 10, 10, 9, 8, 8, 6, 5, 4, 2, 0, -1, -3, -4, -6, -7, -8, -9, -9, -10, 1,8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0,
 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7,
 -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9,
 -10, -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9,
 -10, -10, -11, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -13, -9, -3, 3, 9, 13, 16, 16,
 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10,
 -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1,
 -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9,
 11, 12, 13, 14, 14, 15, 15, 16, 16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -6, -8, -10, -12, -13, -14, -15, -16, -16, -16, -15, -14,
 -13, -11, -10, -8, -6, -3, -1, 1, 3, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 14, 13, 12, 10, 8, 6, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, 10, 5, -4,
 -10, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 15, 6, -6, -15, -15, -6, 6, 15, 15, 15, 15, 14, 14, 14, 14, 13,
 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12,
 -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, 15, 6, -6, -15, -15, -6, 6, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -7, -9,
 -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -7, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 14,
 15, 15, 15, 6, -6, -15, -15, -6, 6, 15, 14, 14, 13, 12, 10, 8, 7, 5, 2, 0, -2, -4, -6, -8, -10, -11, -12, -13, -14, -15, -15, -15, -14, -13, -12, -11, -9, -7, -5, -3,
 -1, 1, 3, 5, 7, 9, 11, 12, 13, 14, 15, 15, 15, 14, 13, 12, 11, 10, 8, 6, 4, 2, 0, -2, -5, -7, -8, -10, -12, -13, -14, -14, -15, -9, 2, 11, 6, -6, -11, -2, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3,
 -4, -4, -5, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 12,
 12, 11, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6,
 -5, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, -13, 3, 16, 9, -9, -16, -3, 13, 13, 12, 12, 10, 9, 8, 6, 4, 2, 0, -2, -4, -5, -7, -9, -10, -11,
 -12, -13, -13, -13, -13, -13, -12, -11, -10, -8, -6, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 13, 13, 13, 13, 13, 12, 11, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6, -8, -9, -10,
 -12, -12, -13, -13, 0,0,0,0,1,15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 15,
 15, 14, 12, 11, 8, 6, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1,
 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 15, 15, 14, 12, 11, 8, 6, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -3, -3,
 -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -2, -2, -1, -1, 0, -15, -11, -3, 6,
 12, 15, 14, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3,
 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, -15, -11, -3, 6, 12, 15, 14, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5,
 -5, -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, -15, -11, -3, 6, 12, 15, 14, 8, 8, 8,
 7, 7, 6, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -7, -8, -8, -8, -8, -8, -8, -7, -7, -6, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 8,
 8, 8, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, -15, -11, -3, 6, 12, 15, 14, 8, 8, 7, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -8, -8, -8, -7, -6, -5, -4, -3, -2, 0,
 1, 3, 4, 5, 6, 7, 7, 8, 8, 8, 8, 7, 7, 6, 5, 3, 2, 1, -1, -2, -3, -4, -6, -6, -7, -8, -8, -8, -8, -8, -7, -6, -5, -4, -3, -1, 14, 3, -11, -15, -8, 6, 15, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4,
 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 14, 3, -11, -15, -8, 6, 15, 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6,
 -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -4, -3, -2, -1, 14, 3,
 -11, -15, -8, 6, 15, 12, 12, 12, 11, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -12, -12, -12, -12, -12, -12, -12, -11, -11, -10, -9, -7, -6, -5,
 -3, -2, 0, 1, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 12, 12, 12, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 14, 3, -11, -15, -8, 6, 15, 12, 12, 11, 10, 9, 7, 6, 4, 2, -1, -3, -5,
 -7, -8, -10, -11, -12, -12, -12, -12, -12, -11, -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 9, 11, 12, 12, 12, 12, 12, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -12,
 -12, -12, -12, -12, -11, -9, -8, -6, -4, -2, -12, 6, 15, 3, -14, -11, 8, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13,
 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, -12, 6, 15, 3, -14, -11, 8, 15, 15,
 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, -12, 6, 15, 3, -14, -11, 8, 15, 15, 14, 14, 13, 12, 10, 9, 8, 6, 4, 2, 1,
 -1, -3, -5, -6, -8, -9, -11, -12, -13, -14, -14, -15, -15, -15, -15, -14, -13, -13, -11, -10, -9, -7, -6, -4, -2, 0, 1, 3, 5, 7, 8, 10, 11, 12, 13, 14, 14, 15, 15, 15,
 15, 14, 13, 12, 11, 10, 9, 7, 5, 4, 2, -12, 6, 15, 3, -14, -11, 8, 15, 14, 14, 12, 11, 9, 7, 4, 2, -1, -3, -6, -8, -10, -12, -13, -14, -15, -15, -15, -14, -13, -11, -10,
 -8, -5, -3, 0, 2, 5, 7, 9, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 9, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -14, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0,1,0,
 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13,
 13, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -13, -14, -14, -15,
 -15, -15, -15, 0, 1, 1, 1, 2, 2, 2, 2, 4, 5, 7, 9, 10, 11, 13, 13, 14, 15, 15, 15, 15, 15, 14, 13, 12, 11, 10, 8, 7, 5, 3, 2, 0, -2, -4, -6, -7, -9, -10, -12, -13, -14,
 -14, -15, -15, -15, -15, -15, -14, -13, -12, -11, -10, -8, -6, -5, -3, -1, 1, 3, 4, 6, 8, 9, 11, 12, 13, 14, 14, 15, 15, 0, 1, 1, 2, 2, 2, 2, 3, 5, 7, 10, 11, 13, 14,
 15, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1, -2, -4, -6, -9, -11, -12, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13,
 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -13, -14, -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10,
 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, -1, -1, -1, -1, 0, 0, 1, 1,
 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5,
 -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, -1, -2, -2, -2, -1, 0, 1, 2, 4, 5, 7, 9, 10, 11, 12, 13, 14, 15, 15, 15, 15, 14, 14, 13, 12,
 11, 10, 8, 7, 5, 3, 1, 0, -2, -4, -6, -7, -9, -10, -11, -13, -13, -14, -15, -15, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -5, -3, -1, 1, 2, 4, 6, 8, 9, 10, 12, 13,
 14, 14, 15, 15, -1, -2, -3, -2, -1, 0, 2, 2, 5, 7, 9, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -9, -10, -12, -13, -14, -15, -15, -15, -14, -13,
 -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -12, -14, -14, -15, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2,
 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13,
 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 1, 1, 0, -1, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12,
 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -13, -13, -13, -14, -14, -14, 2, 2, 1, -1, -2, -1, 0, 2, 3,
 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 3, 1, 0, -2, -4, -5, -7, -8, -9, -11, -12, -12, -13, -13, -14, -14, -14, -13, -13, -12, -11,
 -10, -9, -7, -6, -4, -3, -1, 1, 2, 4, 5, 7, 8, 10, 11, 12, 12, 13, 14, 14, 2, 2, 1, -1, -3, -2, 0, 2, 5, 7, 9, 10, 12, 13, 13, 14, 14, 13, 12, 11, 9, 8, 5, 3, 1, -1, -4,
 -6, -8, -10, -11, -12, -13, -14, -14, -13, -13, -12, -10, -8, -6, -4, -2, 0, 3, 5, 7, 9, 11, 12, 13, 14, 14, 14, 13, 12, 11, 9, 7, 5, 3, 1, -2, -4, -6, -8, -10, -11,
 -12, -13, -14, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 11, 11,
 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, -1, -1, 1, 1, 0, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12,
 -12, -12, -12, -12, -2, -1, 1, 2, 0, -2, -1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 12, 12, 12, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 1, 0, -2, -3, -5, -6, -7, -9, -10, -11,
 -11, -12, -12, -12, -12, -12, -12, -12, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 11, 12, 12, 12, -2, -1, 2, 2, 0, -3, -1, 2, 4, 6, 8, 9, 11,
 12, 12, 12, 12, 12, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -12, -12, -12, -12, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10, 11, 12, 12, 12, 12, 12, 11,
 10, 8, 7, 5, 3, 1, -2, -4, -6, -7, -9, -10, -11, -12, -12, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3,
 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2,
 -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3,
 -3, -3, -3, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, -1, -1, 0, 0, 1, 1, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1,
 -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -8, -2, -1, 0, 1, 2, 2, 2, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 8, 8, 8, 8, 8, 7, 7, 6, 5, 4, 4,
 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -6, -7, -7, -8, -8, -8, -8, -8, -8, -7, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 6, 7, 7, 8, 8, 8, -2, -2, 0, 1, 2,
 3, 2, 1, 3, 4, 5, 6, 7, 8, 8, 8, 8, 8, 7, 6, 6, 4, 3, 2, 1, -1, -2, -3, -5, -6, -7, -7, -8, -8, -8, -8, -7, -7, -6, -5, -4, -3, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8, 8, 8, 8,
 7, 6, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -7, -8, -8, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8,
 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 1, 0, -1, -1, -1, 0, 1, 1, 2, 3, 4, 4,
 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -5, -6, -7, -8,
 -8, -9, -10, -10, -11, -11, -12, -12, -12, -12, -12, -12, 2, 0, -1, -2, -1, 1, 2, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 12, 12, 12, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 1,
 0, -2, -3, -5, -6, -7, -9, -10, -11, -11, -12, -12, -12, -12, -12, -12, -12, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 11, 12, 12, 12, 2, 0, -2,
 -3, -1, 1, 2, 2, 4, 6, 8, 9, 11, 12, 12, 12, 12, 12, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -12, -12, -12, -12, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 6,
 8, 10, 11, 12, 12, 12, 12, 12, 11, 10, 8, 7, 5, 3, 1, -2, -4, -6, -7, -9, -10, -11, -12, -12, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6,
 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, -1, 0, 1, 0, -1, -1, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4,
 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, -2, 1, 2, 0, -2, -1, 1, 2, 4, 5, 7, 9, 10, 11, 12, 13, 14,
 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 5, 3, 1, 0, -2, -4, -6, -7, -9, -10, -11, -13, -13, -14, -15, -15, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -5, -3,
 -1, 1, 2, 4, 6, 8, 9, 10, 12, 13, 14, 14, 15, 15, -2, 1, 3, 0, -2, -2, 1, 2, 5, 7, 9, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -9, -10, -12, -13,
 -14, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -12, -14, -14, -15, 1,8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10,
 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10,
 -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5,
 -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10, -10, -11, -11, -11, -11, -11, -10, -9,
 -8, -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -10, -11, -11, -11, -11, -10, -9, -7, -5,
 -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10,
 12, 14, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8,
 -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16,
 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12,
 -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 15, 15, 14,
 12, 11, 9, 7, 5, 3, 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15,
 15, 14, 12, 10, 8, 7, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12,
 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15,
 -15, -15, -15, -16, -16, -16, -16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5,
 -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14,
 14, 15, 15, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -6, -8, -10, -12, -13, -14, -15, -16, -16,
 -16, -15, -14, -13, -11, -10, -8, -6, -3, -1, 1, 3, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 14, 13, 12, 10, 8, 6, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15,
 -16, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -15, -10, -2, 8,
 14, 16, 12, 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 7, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1,
 -2, -3, -3, -4, -5, -6, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -10, -2, 8, 14, 16, 12,
 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 6, 5, 4, 2, 1, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -15,
 -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 1, 2, 4, 5, 6, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, -15, -10, -2, 8, 14, 16, 12, 4, -4, -12,
 -16, -14, -8, 2, 10, 15, 15, 14, 13, 12, 11, 9, 7, 5, 3, 0, -2, -4, -6, -8, -10, -12, -13, -14, -15, -15, -15, -15, -15, -14, -12, -11, -10, -7, -6, -3, -1, 1, 3, 6, 7,
 10, 11, 12, 14, 15, 15, 15, 15, 15, 14, 13, 12, 10, 8, 6, 4, 2, 0, -3, -5, -7, -9, -11, -12, -13, -14, -15, -15, 0,0,0,0,1,15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7,
 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 15, 15, 15, 15, 14, 14,
 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 6, 4, 3, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, 0, 0, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4,
 -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, -15, -14, -12, -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3,
 -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, -15, -14, -12,
 -8, -4, 0, 4, 8, 12, 14, 15, 15, 14, 12, 8, 4, 4, 4, 4, 3, 3, 2, 1, 1, 0, -1, -2, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, 0, 1, 1, 2, 3, 3, 4, 4, 4,
 4, 4, 4, 4, 4, 3, 3, 2, 1, 0, 0, -1, -2, -2, -3, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -1, -1, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1,
 1, 1, 1, 1, 0, 0, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3,
 -3, -4, -4, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -2, -1, -1, 15, 12, 6, -1, -8, -14,
 -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 7, 6, 6, 6, 5, 4, 4, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -6, -6, -6, -7, -7, -7, -7, -7, -7, -6, -6, -5, -5, -4, -3, -3,
 -2, -1, 0, 1, 2, 2, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 15, 12, 6, -1, -8, -14, -15, -14, -10, -3, 4, 11, 15, 15, 13, 7, 7, 6, 6, 5, 4, 3,
 2, 1, 0, -2, -3, -4, -5, -6, -6, -7, -7, -7, -7, -7, -6, -5, -5, -4, -2, -1, 0, 1, 2, 3, 4, 5, 6, 6, 7, 7, 7, 7, 6, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -6, -7, -7,
 -7, -7, -7, -6, -5, -4, -3, -2, -1, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 8,
 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15,
 -12, -3, 7, 14, 15, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10,
 -10, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -3, -3, -2, -1, -1, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 10, 9,
 9, 8, 8, 7, 6, 5, 4, 3, 2, 0, -1, -2, -3, -4, -5, -6, -7, -8, -8, -9, -9, -10, -10, -10, -9, -9, -9, -8, -7, -7, -6, -5, -4, -3, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 9,
 10, 10, 10, 9, 9, 9, 8, 7, 6, 6, 5, 3, 2, 1, -15, -8, 1, 11, 15, 14, 6, -4, -13, -15, -12, -3, 7, 14, 15, 10, 9, 9, 8, 7, 6, 4, 3, 1, 0, -2, -4, -5, -6, -8, -8, -9, -10,
 -10, -10, -9, -8, -7, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -9, -10, -10, -9, -9, -8, -7, -6, -5, -3,
 -2, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12,
 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4,
 6, 7, 8, 9, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -7, -8,
 -9, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 4, 6, 7, 9, 10, 12, 13, 14, 15, 15, 15, 15, 15, 15,
 14, 14, 13, 11, 10, 9, 7, 5, 3, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -15, -15, -15, -14, -13, -12, -11, -10, -8, -7, -5, -3, -1, 1, 3, 4, 6, 8,
 9, 11, 12, 13, 14, 15, 15, 15, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 5, 8, 10, 12, 13, 14, 15, 15, 15, 15, 14, 12, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13,
 -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 6, 8, 10, 12, 13, 15, 15, 15, 15, 15, 14, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -13, -14, -15, -15, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12,
 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 3, 4,
 6, 7, 8, 9, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -7, -8,
 -9, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -1, -1, -1, -2, -2, -2, -2, -1, -1, -1, 0, 1, 1, 1, 2, 2, 4, 6, 7, 9, 10, 12, 13, 14, 15, 15, 15, 15,
 15, 15, 14, 14, 13, 11, 10, 9, 7, 5, 3, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -15, -15, -15, -14, -13, -12, -11, -10, -8, -7, -5, -3, -1, 1, 3, 4,
 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, -1, -1, -2, -2, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 5, 8, 10, 12, 13, 14, 15, 15, 15, 15, 14, 12, 11, 9, 6, 4, 1, -2, -4, -7,
 -9, -11, -13, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 6, 8, 10, 12, 13, 15, 15, 15, 15, 15, 14, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -13, -14,
 -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12,
 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1,
 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5,
 -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, 1, 2, 2, 2, 1, 1, 0, -1, -2, -2, -2, -1, 0, 1, 1, 2, 4, 5, 7, 9, 10, 12, 13, 14, 14, 15, 15,
 15, 15, 15, 14, 13, 12, 11, 10, 8, 7, 5, 3, 2, 0, -2, -4, -6, -7, -9, -11, -12, -13, -14, -14, -15, -15, -15, -15, -15, -14, -13, -12, -11, -10, -8, -6, -5, -3, -1, 1,
 3, 4, 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, 1, 2, 3, 2, 2, 1, 0, -2, -2, -3, -2, -1, 0, 1, 2, 3, 5, 7, 10, 12, 13, 14, 15, 15, 15, 15, 14, 12, 11, 8, 6, 4, 1, -2, -4, -6,
 -9, -11, -12, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 12, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -13, -14,
 -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11,
 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, -1, -1, -1, -1, 0, 1, 1, 1, 0, 0, -1, -1, -1, 0, 1,
 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4,
 -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -14, -15, -15, -15, -1, -2, -2, -1, 0, 1, 2, 2, 1, -1, -2, -2, -1, 0, 1, 2, 4, 5, 7, 9, 10, 11, 12, 13, 14,
 15, 15, 15, 15, 14, 14, 13, 12, 11, 10, 8, 7, 5, 3, 1, 0, -2, -4, -6, -7, -9, -10, -11, -13, -13, -14, -15, -15, -15, -15, -14, -14, -13, -12, -11, -9, -8, -6, -5, -3,
 -1, 1, 2, 4, 6, 8, 9, 10, 12, 13, 14, 14, 15, 15, -2, -3, -2, -1, 0, 2, 3, 2, 1, -1, -2, -3, -2, 0, 1, 2, 5, 7, 9, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 4, 1,
 -1, -4, -6, -9, -10, -12, -13, -14, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 5, 8, 10, 11, 13, 14, 15, 15, 15, 14, 13, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9,
 -11, -12, -14, -14, -15, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3,
 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4,
 -4, -4, -4, -2, -2, -1, -1, -1, 0, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3,
 -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 2, 2, 1, 1, 1, 2,
 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 2, 2, 1, 0, 0, -1, -2, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -2, -1, -1, 0, 1, 2, 2, 3, 3, 4, 4, 4, 4, 4, 4, 4, 3, 3, 2, 2, 1,
 0, -1, -1, -2, -3, -3, -4, -4, -4, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4,
 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2,
 3, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6,
 -6, -6, -6, -7, -7, -7, -7, -7, 2, 1, 1, 0, -1, -2, -2, -2, -1, 0, 1, 1, 2, 2, 2, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7, 7, 6, 6, 5, 5, 4, 3, 2, 2, 1, 0, -1, -2,
 -3, -3, -4, -5, -5, -6, -6, -7, -7, -7, -7, -7, -7, -6, -6, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 4, 4, 5, 6, 6, 6, 7, 7, 7, 3, 2, 1, 0, -1, -2, -3, -2, -2, 0, 1,
 2, 2, 3, 2, 1, 2, 3, 4, 5, 6, 7, 7, 7, 7, 7, 6, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -6, -7, -7, -7, -7, -6, -6, -5, -4, -3, -2, -1, 0, 1, 3, 4, 5, 5, 6, 7, 7, 7,
 7, 7, 6, 6, 5, 4, 3, 2, 0, -1, -2, -3, -4, -5, -6, -6, -7, -7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5,
 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, -1, -1, 0, 1, 1, 1, 0, 0, -1,
 -1, -1, 0, 1, 1, 1, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2,
 -3, -3, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -2, -1, 0, 1, 2, 2, 1, -1, -2, -2, -1, 0, 1, 2, 2, 1, 2, 3, 5, 6, 6, 7, 8, 9, 9, 9, 10, 10,
 10, 9, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -7, -8, -9, -9, -9, -10, -10, -10, -9, -9, -8, -8, -7, -6, -5, -4, -3, -2, -1, 0, 2, 3, 4, 5, 6, 7, 8, 8,
 9, 9, 10, 10, -2, -1, 0, 2, 3, 2, 1, -1, -2, -3, -2, 0, 1, 2, 3, 2, 3, 5, 6, 7, 8, 9, 9, 10, 10, 9, 9, 8, 7, 5, 4, 2, 1, -1, -3, -4, -6, -7, -8, -9, -9, -10, -10, -9,
 -9, -8, -7, -6, -5, -3, -1, 0, 2, 3, 5, 6, 7, 8, 9, 10, 10, 10, 9, 8, 8, 6, 5, 4, 2, 0, -1, -3, -4, -6, -7, -8, -9, -9, -10, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3,
 -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10,
 -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10,
 -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5, -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10,
 -10, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -15, -15, -14, -13, -12, -11, -10, -8, -7, -6, -4, -2, -1, 1, 2, 4, 6, 7, 8, 10, 11, 12, 13, 14, 15,
 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8,
 -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4,
 -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14,
 -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, -16, -16, -15,
 -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -7, -8,
 -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 8, 7, 4, 2, 0, -3, -5,
 -7, -9, -11, -12, -14, -15, -15, -16, 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10,
 -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3,
 -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4,
 -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7,
 -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14,
 15, 15, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 15, 15, 14, 12, 11,
 9, 7, 5, 3, 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14,
 12, 10, 8, 7, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5, -8, -10, -11, -11, -11, -9, -7,
 -4, 0, 3, 6, 8, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15, -12, -8, -4, 1, 6, 10, 13, 15, 16, 16, 14, 11, 7,
 2, -2, -7, -11, -14, -16, -16, -15, -13, -10, -6, -1, 4, 8, 12, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4,
 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16,
 -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5,
 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7,
 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, -16, -14, -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14,
 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15,
 16, 16, 16, 15, 15, 14, 12, 10, 8, 7, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, 0,0,0,0,1,16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12,
 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5,
 5, 4, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3,
 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16, 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
 0, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16, 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1,
 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16, 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 2, 2, 1,
 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 1, 1, 1, 1, 0, -16, -15, -15, -14, -13, -11, -9, -7, -5, -3, -1, 2, 4, 6, 8, 10, 12, 13, 14, 15, 15, 16, 16, 15, 14, 14, 12, 11, 9, 7, 5, 2, 2, 2, 2, 2, 1, 1, 1,
 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -1, -1, -1, 0, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0,
 0, 0, 0, 0, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3,
 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -2, -2,
 -2, -2, -2, -1, -1, -1, -1, 0, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 4, 4, 3,
 3, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4,
 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 16, 15, 13, 11, 7, 4, 0, -4, -7, -11, -13, -15, -16, -16, -15, -13, -11, -7, -4, 0, 4, 7, 11, 13, 15, 16, 16, 15, 13, 11, 7, 4, 4, 3,
 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -3, -3,
 -3, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, -15, -14, -11, -6, -1, 5, 9, 13, 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 16, 15, 14,
 10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1,
 1, 1, 1, 1, 1, 1, 0, 0, 0, -15, -14, -11, -6, -1, 5, 9, 13, 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 16, 15, 14, 10, 5, 5, 5, 5, 5,
 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4,
 -4, -4, -4, -3, -3, -3, -2, -2, -2, -1, -1, 0, -15, -14, -11, -6, -1, 5, 9, 13, 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 16, 15, 14,
 10, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -3, -3, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 4, 4,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 2, 2, 1, 1, -15, -14, -11, -6, -1, 5, 9, 13, 15, 16, 14, 12, 7, 2, -3, -8, -12, -15, -16, -15, -13, -9, -4, 2, 7, 11, 14, 16, 15,
 14, 10, 5, 5, 5, 4, 4, 3, 2, 2, 1, 0, -1, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 4, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 1, 0, -1,
 -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -5, -4, -3, -3, -2, -1, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9,
 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10,
 -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 7, 9,
 11, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -11,
 -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 14, 14, 15, 16, 16, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3,
 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16, -16, -15, -15, -13, -12, -10, -7, -5, -2, 0, 3, 6, 8, 10, 12, 14,
 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0,
 -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1,
 -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 4, 6, 7, 9, 11, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13,
 -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 14, 14, 15, 16, 16, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3,
 -3, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 2, 2, 3, 3, 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15,
 -16, -16, -15, -15, -13, -12, -10, -7, -5, -2, 0, 3, 6, 8, 10, 12, 14, 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10,
 10, 10, 11, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 15,
 15, 15, 14, 14, 13, 13, 12, 11, 10, 10, 9, 8, 7, 6, 5, 3, 2, 1, 0, -1, -2, -3, -5, -6, -7, -8, -9, -10, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -16, -16, 0, 1,
 1, 2, 2, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 4, 6, 7, 9, 10, 12, 13, 14, 15, 15, 16, 16, 15, 15, 15, 14, 13, 11,
 10, 9, 7, 5, 3, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -16, -16, -15, -15, -14, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 13,
 14, 15, 15, 16, 1, 1, 2, 2, 2, 3, 3, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 3, 5, 8, 10, 12, 13, 15, 15, 16, 15, 15, 14, 13,
 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16, -16, -15, -14, -13, -11, -10, -7, -5, -2, 0, 3, 6, 8, 10, 12, 14, 15, 15, 16, 15, 15, 14, 12, 10, 8, 6, 3, 1,
 -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4,
 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 3, 4, 6, 7, 8, 9, 9, 10, 11, 12,
 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -7, -8, -9, -9, -10, -11, -12, -13,
 -13, -14, -14, -15, -15, -15, -15, -15, -1, -1, -2, -2, -2, -2, -1, -1, 0, 0, 1, 2, 2, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 4, 6, 7, 9, 10,
 12, 13, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 11, 10, 9, 7, 5, 3, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -15, -15, -15, -15, -14, -13, -12, -11, -10,
 -8, -7, -5, -3, -1, 1, 3, 4, 6, 8, 9, 11, 12, 13, 14, 15, 15, 15, -1, -2, -2, -3, -3, -2, -2, -1, 0, 1, 1, 2, 2, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -3, -2, -2, -1, 0, 1,
 2, 2, 3, 5, 8, 10, 12, 13, 14, 15, 15, 15, 15, 14, 12, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -15, -15, -15, -14, -13, -11, -9, -7, -5, -2, 0, 3, 6, 8, 10,
 12, 13, 15, 15, 15, 15, 15, 14, 12, 10, 8, 6, 3, 1, -2, -4, -7, -9, -11, -13, -14, -15, -15, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, -3, -3,
 -2, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2,
 -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1,
 -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 2, 2, 2, 1,
 1, 0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2,
 -1, -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 3, 2, 2, 2, 1, 1, 0, -1, -1, -2, -2, -2, -3, -3, -2, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 3, 3, 2, 2, 2, 1, 1, 1, 2, 2, 3,
 3, 3, 4, 4, 4, 4, 3, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0,
 -1, -2, -2, -3, -3, -3, -4, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -1, -1, -1, 0, 0, 0,
 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -2, -2, -1, -1, 0, 1, 1, 2, 2, 2, 2, 1, 1,
 0, 0, -1, -1, -2, -2, -2, -2, -1, 0, 0, 1, 1, 2, 2, 2, 2, 1, 1, 1, 2, 2, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -4,
 -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 5, 5, -3, -2, -2, -1, 0, 1, 2, 2, 3, 3, 2, 2, 1, 0, 0, -1, -2,
 -2, -3, -2, -2, -1, -1, 0, 1, 2, 2, 3, 3, 2, 2, 1, 2, 3, 3, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 0, -1, -1, -2, -3, -4, -4, -5, -5, -5, -5, -5, -5, -4, -4, -3, -2, -2,
 -1, 0, 1, 2, 3, 3, 4, 5, 5, 5, 5, 5, 5, 5, 4, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -5, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0,
 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10, -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5, -4,
 -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10,
 -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10,
 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -15, -15, -15, -15,
 -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13,
 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2,
 -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15,
 -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12,
 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6, -7, -8, -10, -11, -12, -13, -14, -15, -15,
 -16, -16, -16, -16, -16, -16, -15, -15, -14, -13, -12, -11, -10, -8, -7, -6, -4, -2, -1, 1, 2, 4, 6, 7, 8, 10, 11, 12, 13, 14, 15, 15, 16, 16, 16, -16, -16, -16, -16,
 -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10,
 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 15, 14, 13, 11, 9, 7, 5, 3, 0, -2, -4, -7, -9, -10, -12, -14, -15, -15, -16, -16, -16, -15,
 -14, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 9, 7, 4, 2, 0, -3, -5, -7, -9, -11, -13, -14, -15, -16, -16, 11, 11,
 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5,
 -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 15,
 15, 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6, -7, -8, -10, -11, -12, -13, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -13, -12, -11, -10, -8,
 -7, -6, -4, -2, -1, 1, 2, 4, 6, 7, 8, 10, 11, 12, 13, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6,
 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16,
 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12,
 -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8,
 -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15,
 15, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14,
 -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -7, -8, -10, -12,
 -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 8, 7, 4, 2, 0, -3, -5, -7, -9,
 -11, -12, -14, -15, -15, -16, -11, -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11, 11, 11, 11, 10, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7,
 -8, -9, -10, -11, -11, -11, -11, -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, -16, -16, -15, -14, -13, -11, -9, -7, -5, -3, 0, 2, 4, 7, 9, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10,
 -12, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -9, -7, -4, -2, 0, 3, 5, 7, 9, 11, 13, 14, 15, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12,
 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15,
 -15, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4,
 -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7,
 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5,
 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1,
 -1, -4, -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3,
 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 8, 7,
 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, 0,0,0,0,1,16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13,
 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16,
 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6,
 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13,
 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16,
 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6,
 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2,
 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16,
 -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16,
 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -16, -16, -16, -15, -15, -15, -14,
 -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14,
 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6,
 -5, -4, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2,
 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 16, 16, 15, 14, 14, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -15, -15, -16, -16, -16, -15,
 -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7, 9, 10, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7, 6, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16,
 15, 14, 14, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7,
 9, 10, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7, 6, 4, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 16, 16, 15, 14, 14, 12, 11, 10, 8,
 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7, 9, 10, 12, 13, 14, 15, 15,
 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7, 6, 4, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1,
 -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 16, 16, 15, 14, 14, 12, 11, 10, 8, 6, 5, 3, 1, -1, -3, -5, -7, -8, -10, -11,
 -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -12, -11, -9, -8, -6, -4, -2, 0, 2, 4, 5, 7, 9, 10, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 11, 9, 7,
 6, 4, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1,
 -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 4, 6, 8, 11, 12, 14, 15, 16, 16, 16, 15, 14, 12, 10, 8, 6, 3, 0,
 -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 16, 16, 15, 15, 14, 12, 10, 8, 5, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0,
 -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 4, 6, 8, 11, 12, 14, 15, 16, 16, 16, 15, 14, 12, 10, 8, 6, 3, 0, -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14,
 -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 16, 16, 15, 15, 14, 12, 10, 8, 5, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, -16, -15, -14, -13,
 -11, -9, -7, -5, -2, 1, 4, 6, 8, 11, 12, 14, 15, 16, 16, 16, 15, 14, 12, 10, 8, 6, 3, 0, -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4,
 -2, 1, 4, 6, 9, 11, 13, 14, 15, 16, 16, 15, 15, 14, 12, 10, 8, 5, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -2,
 -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, -16, -15, -14, -13, -11, -9, -7, -5, -2, 1, 4, 6, 8, 11,
 12, 14, 15, 16, 16, 16, 15, 14, 12, 10, 8, 6, 3, 0, -2, -5, -7, -10, -12, -13, -15, -15, -16, -16, -15, -14, -13, -11, -9, -7, -4, -2, 1, 4, 6, 9, 11, 13, 14, 15, 16,
 16, 15, 15, 14, 12, 10, 8, 5, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2,
 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, 0, 0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7,
 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2,
 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 7, 9, 11, 12, 13, 14, 15, 15, 16, 16,
 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5,
 6, 8, 10, 11, 12, 14, 14, 15, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16, -16,
 -15, -15, -13, -12, -10, -7, -5, -2, 0, 3, 6, 8, 10, 12, 14, 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15,
 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14,
 -15, -15, -15, -16, -16, -16, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 6, 7, 9, 11, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7,
 5, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 14, 14, 15, 16,
 16, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16,
 -16, -15, -15, -13, -12, -10, -7, -5, -2, 0, 3, 6, 8, 10, 12, 14, 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15,
 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15,
 -15, -16, -16, -16, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 4, 6, 7, 9, 11, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6,
 -8, -9, -11, -12, -13, -14, -15, -15, -16, -16, -16, -15, -15, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 14, 14, 15, 16, 16, 0, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1,
 1, 2, 2, 2, 2, 2, 3, 3, 3, 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16, -16, -15, -15, -13, -12, -10, -7, -5,
 -2, 0, 3, 6, 8, 10, 12, 14, 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5,
 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16,
 16, 16, 16, 16, 16, 16, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15,
 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0, -1, -1, -1,
 -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1,
 -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 4, 6, 7, 9, 11, 12, 13, 14, 15, 15, 16, 16, 16, 15, 15, 14, 13, 12, 10, 9, 7, 5, 4, 2, 0, -2, -4, -6, -8, -9, -11, -12, -13, -14,
 -15, -15, -16, -16, -16, -15, -15, -14, -13, -11, -10, -8, -7, -5, -3, -1, 1, 3, 5, 6, 8, 10, 11, 12, 14, 14, 15, 16, 16, 0, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -2,
 -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -2, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2,
 2, 3, 3, 5, 8, 10, 12, 14, 15, 15, 16, 16, 15, 14, 13, 11, 9, 6, 4, 1, -2, -4, -7, -9, -11, -13, -14, -15, -16, -16, -15, -15, -13, -12, -10, -7, -5, -2, 0, 3, 6, 8, 10,
 12, 14, 15, 16, 16, 16, 15, 14, 12, 11, 8, 6, 4, 1, -2, -5, -7, -9, -11, -13, -14, -15, -16, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0,
 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, -3, -3, -3, -3, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1,
 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -3,
 -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2,
 -2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, -1,
 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2,
 -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, -3, -3, -2, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -2, -2,
 -2, -2, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1,
 -2, -2, -2, -2, -3, -3, -3, -3, -2, -2, -2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -2, -3, -3, 1,8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0,
 -1, -1, -2, -2, -3, -3, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 8,
 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2,
 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 9, 8, 7, 5, 4, 2, 0, -1, -3, -5, -6, -7, -9, -10, -10, -11, -11, -11, -11, -11, -10, -9, -8, -7, -5,
 -4, -2, -1, 1, 3, 4, 6, 7, 8, 9, 10, 11, 11, 11, 11, 11, 10, 10, 9, 7, 6, 5, 3, 1, 0, -2, -3, -5, -6, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -9, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5,
 -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8,
 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13,
 -13, -12, -12, -12, -12, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1,
 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15,
 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2,
 -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16,
 -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9,
 -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9,
 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15,
 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6, -7, -8, -10, -11, -12, -13, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -13, -12, -11, -10, -8, -7,
 -6, -4, -2, -1, 1, 2, 4, 6, 7, 8, 10, 11, 12, 13, 14, 15, 15, 16, 16, 16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14,
 -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3,
 -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13,
 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 14, 13, 11, 9, 7, 5, 3, 0, -2, -4, -7, -9, -10, -12, -14, -15, -15,
 -16, -16, -16, -15, -14, -13, -12, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 12, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 9, 7, 4, 2, 0, -3, -5, -7, -9, -11, -13, -14,
 -15, -16, -16, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6,
 -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8,
 -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8,
 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16,
 -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3,
 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11,
 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15,
 -15, -16, -16, -16, -16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5,
 -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14,
 -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13,
 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16,
 -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14,
 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13,
 -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7,
 -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 15, 15, 14, 12, 11, 9,
 7, 5, 3, 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4, 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12,
 10, 8, 7, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3,
 3, 4, 5, 6, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7,
 -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5,
 6, 6, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -15, -15,
 -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15,
 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16,
 -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16,
 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11,
 -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2,
 -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3,
 -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5,
 -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8,
 -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15,
 15, 16, 16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15,
 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13,
 14, 14, 15, 15, 15, 16, 16, 16, 15, 15, 14, 12, 11, 9, 7, 5, 3, 0, -2, -4, -7, -8, -10, -12, -14, -15, -15, -16, -16, -16, -15, -14, -13, -11, -10, -8, -6, -4, -1, 1, 4,
 6, 8, 10, 11, 13, 14, 15, 16, 16, 16, 15, 15, 14, 12, 10, 8, 7, 4, 2, 0, -3, -5, -7, -9, -11, -12, -14, -15, -15, -16, 0,0,0,0,1,16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13,
 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5,
 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12,
 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4,
 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11,
 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2,
 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 11, 11,
 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1,
 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -9, -8,
 -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6,
 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10,
 -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13,
 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9,
 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -14, -13, -13, -13, -12,
 -12, -12, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 12, 12,
 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14,
 -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 7, 7, 8,
 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14,
 14, 13, 13, 13, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12,
 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16,
 -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13,
 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15,
 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15,
 -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9,
 -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -10, -10, -9, -8, -7,
 -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12,
 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0,
 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2,
 1, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14,
 -14, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -4, -4, -2, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16,
 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, -16, -16, -15, -15, -15, -14,
 -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 12, 12, 11, 10, 8,
 7, 6, 5, 4, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7,
 -6, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15,
 15, 14, 13, 12, 12, 11, 10, 8, 7, 6, 5, 4, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14,
 -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7,
 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9,
 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 12, 12, 11, 10, 8, 7, 6, 5, 4, 2, 1, -1, -2, -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14,
 -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4, 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15,
 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10,
 -9, -8, -6, -5, -4, -2, -1, 0, 2, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 12, 12, 11, 10, 8, 7, 6, 5, 4, 2, 1, -1, -2,
 -3, -5, -6, -7, -8, -9, -11, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, 0, 1, 2, 4,
 5, 6, 7, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0,
 0, 0,0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, -3, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1,
 -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3,
 -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2,
 -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2,
 -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2,
 2, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2,
 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9,
 -9, -9, -10, -10, -10, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10,
 -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2,
 -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9,
 9, 8, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9,
 -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7,
 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -10, -4, 5, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, -15, -6, 6, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 6, 5,
 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -9, -10, -10,
 -10, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15,
 -6, 6, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8,
 -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -13, -13, -13, -12, -12, -12,
 -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14,
 14, 14, 15, 15, 15, -15, -6, 6, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -6, -6, -7, -8, -9, -10, -11, -12, -12,
 -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, 0, 0, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10,
 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 6, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11,
 -12, -12, -13, -13, -14, -14, -14, -15, -15, 8, -8, -8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0,
 -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10,
 -10, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9,
 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1,
 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, -11, -11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 6, 6,
 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7,
 -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3,
 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -4, 10, -10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, -6,
 15, -15, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2,
 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -5, -5, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, 15, -15, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5,
 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6,
 -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3,
 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, -6, 15, -15, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2,
 -3, -3, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -6, -5, -5, -5, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3,
 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -5, -5, -5, -5, -6,
 -6, -6, -6, -6, -6, -6, 0,0,0,0,0,0,1,0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7,
 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0,
 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4,
 -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, 0, 1,
 1, 1, 2, 3, 4, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 6, 5,
 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -13, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13,
 -13, -13, -12, -12, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15,
 15, 0, 1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, 0, -2, -3, -4, -5, -6, -8,
 -9, -10, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10,
 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14,
 -14, -14, -15, -15, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7,
 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4,
 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12,
 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -7,
 -8, -8, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -1, -1, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 8, 8,
 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4,
 -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -10, -10, -9, -9, -8, -7,
 -7, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 5, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, -1, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -12, -13, -13,
 -13, -13, -13, -13, -13, -12, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13,
 12, 12, 11, 11, 10, 9, 8, 8, 7, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -13, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2,
 -3, -3, -3, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, 1, 0, -1, 1, 1, 2,
 2, 3, 3, 4, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3,
 -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4,
 -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 1, 0, -1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10,
 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8,
 -8, -7, -6, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -2,
 -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3,
 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -1, 1, -1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4,
 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5,
 -5, -5, -5, -5, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, -1, 1,
 -1, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4,
 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3,
 -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -5, -5, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4,
 4, 5, 5, 5, 5, 5, 5, 5, 5, 1, 1, 1, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2,
 -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -2, -2, -1, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 5, 5,
 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1,
 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10,
 10, 11, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8,
 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -10, -11,
 -11, -11, -11, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -1, 0, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -11,
 -11, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 3,
 3, 4, 5, 5, 6, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, -1, 0, 1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12,
 12, 12, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -13, -12, -12, -11, -11,
 -10, -9, -9, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 9, 8, 8, 7, 5, 4, 3, 2, 1,
 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -13, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14,
 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4,
 4, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14,
 -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, 1, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14,
 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -13, -13, -13, -14, -14,
 -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -13, -12, -12, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8,
 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 14, 15, 15, 15, 1, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 14, 14, 14, 13, 12, 12,
 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, 0, -2, -3, -4, -5, -6, -8, -9, -10, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -11, -10,
 -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 0, -1,
 -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -14, -15, -15, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 0, 0, 0, 0, 1, 1,
 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9,
 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -7,
 -7, -7, -7, -8, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, 0, 1, -1, 1, 1, 2, 2, 3, 3, 4, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9,
 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -8, -8, -9,
 -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9, -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6,
 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 0, 1, -1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 1,
 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -6, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4,
 5, 6, 6, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10,
 -10, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1,
 -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10,
 -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9,
 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1,
 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9,
 8, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9,
 -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7,
 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -13, -9, -3,
 3, 9, 13, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9,
 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9,
 -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -16, -16, -16, -16, -16, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1,
 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15,
 -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10,
 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 10, 9, 8, 7, 6, 4, 3,
 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9,
 -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7,
 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -16, -16, 10, 5, -4, -10, -10, -4, 5, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 15, 6, -6, -15,
 -15, -6, 6, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 9, 8, 8,
 8, 7, 7, 7, 6, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -6, -7, -7, -7, -8, -8, -8,
 -9, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -14, -14, -14, -14, -15,
 -15, -15, -15, -15, -15, 15, 6, -6, -15, -15, -6, 6, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 2, 1, 0, 0,
 -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14,
 -14, -14, -14, -13, -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10,
 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 6, -6, -15, -15, -6, 6, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
 -1, -2, -3, -4, -6, -6, -7, -8, -9, -10, -11, -12, -12, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7,
 -6, -5, -4, -3, -2, 0, 0, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 6, 4, 3, 2,
 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -9, 2, 11, 6, -6, -11, -2, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
 9, 9, 9, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 9,
 9, 9, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -6, -6,
 -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13,
 -13, -13, -13, -13, -13, -13, -13, -13, -13, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4,
 3, 2, 2, 1, 0, 0, -1, -2, -2, -3, -4, -4, -5, -5, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13,
 -13, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -7, -6, -5, -5, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4, 4, 5, 5, 6, 7, 7,
 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, -13, 3, 16, 9, -9, -16, -3, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4,
 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, -13, -13, -13, -13, -12, -12, -12, -11, -11, -10, -9, -8, -8,
 -7, -6, -5, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 7, 7, 6, 5, 4,
 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -6, -7, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -13, -13, -13, 0,0,0,0,0,0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11,
 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9,
 10, 10, 11, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12,
 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10,
 -11, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, 0, 0, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10,
 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5,
 -6, -7, -8, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -9,
 -9, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 0, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 6,
 7, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -12, -13,
 -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 6, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -14, -15, -15, -15, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9,
 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, -1, 0, 0, 0, 0, 0, 1, 2, 2, 3, 3,
 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14,
 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6,
 -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -1, -1, -1, -1, 0, 0, 1,
 1, 2, 3, 4, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4,
 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13,
 -13, -12, -12, -11, -11, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, -1,
 -1, -1, -1, 0, 0, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, 0, -2, -3, -4, -5,
 -6, -8, -9, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 3, 4, 6, 7, 8,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13,
 -14, -14, -15, -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7,
 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 0, 0, 0,
 0, -1, 0, 0, 0, 1, 2, 2, 2, 3, 4, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 8, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 13, 13, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3,
 -4, -4, -5, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, 1,
 1, 0, -1, -1, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 10, 9, 9, 8,
 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -14, -14, -14, -13,
 -13, -13, -12, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 13, 13, 13,
 13, 14, 14, 14, 1, 1, 1, -1, -1, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0,
 -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2,
 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -5, -7, -8, -9, -9, -10, -11,
 -12, -12, -13, -13, -13, -14, -14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 6,
 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 0, 0, 0,
 0, 0, -1, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 6, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3,
 -4, -4, -5, -5, -5, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -1, -1, 1,
 1, 0, -1, 0, 1, 2, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5,
 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12,
 -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 12,
 12, -1, -1, 1, 1, 0, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3,
 -4, -5, -6, -7, -8, -9, -10, -10, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7,
 8, 8, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12,
 -12, -12, -12, -12, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3,
 -3, -3, -3, -3, -3, -3, -3, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2,
 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3,
 -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1, 1, 1, 0, 0, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3,
 -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1,
 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7,
 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 0, 0, 0, 0, 0,
 1, 0, 0, 1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7,
 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5,
 -5, -5, -6, -6, -6, -6, -6, -7, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 5, 6, 6, 6, 7, 7,
 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7,
 -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7, -7, -7, -7, -6, -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 5,
 5, 5, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, -1, -1, 0, 1, 1, 1, 1, 1, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2,
 1, 1, 0, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -7, -8, -8, -8, -8, -8, -8, -8, -8, -8, -7, -7, -7, -6, -6, -5, -5, -4, -4, -3, -2, -2, -1, 0, 0, 1, 2, 2, 3, 4,
 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 3, 3, 2, 1, 1, 0, -1, -1, -2, -3, -3, -4, -4, -5, -6, -6, -6, -7, -7, -8, -8, -8, -8, -8, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7,
 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 0, 0, 0, -1, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4,
 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 6, 6, 6, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3, -3, -4, -4, -5, -5, -5, -6, -6, -7, -7,
 -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, 1, 0, -1, -1, -1, 0, 1, 1, 2, 2, 3, 4, 4, 5,
 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3,
 -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -11, -10, -10, -9, -9, -8,
 -8, -7, -6, -6, -5, -4, -4, -3, -2, -1, -1, 0, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 1, 0, -1, -1, -1, 1, 1, 1, 2, 3, 4,
 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -10, -11,
 -11, -12, -12, -12, -12, -12, -12, -12, -12, -12, -11, -11, -10, -10, -9, -8, -7, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 12, 12, 12,
 12, 12, 12, 12, 12, 12, 11, 11, 10, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -11, -12, -12, -12, -12, -12, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9,
 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 1, 0, 0, 0, 0, 0, 1, 2, 2, 3, 3, 4, 4,
 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14,
 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7,
 -8, -8, -9, -9, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -1, 0, 1, 0, -1, -1, 1, 1, 2, 3,
 4, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6, 5, 4, 3, 2, 1,
 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13,
 -12, -12, -11, -11, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, -1, 1,
 1, 0, -1, -1, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 3, 2, 1, 0, -2, -3, -4, -5, -6,
 -8, -9, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -5, -4, -3, -2, 0, 1, 2, 3, 4, 6, 7, 8, 9,
 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14,
 -14, -15, -15, -15, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5,
 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7,
 -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1,
 -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11,
 -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10,
 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 2,
 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -5,
 -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0,
 -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -15,
 -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12,
 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5,
 -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15,
 -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15,
 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13,
 -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7,
 -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, -16, -15, -14, -12,
 -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12,
 -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10,
 -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 11, 9, 6, 2, -2, -6, -9, -11, -11, -9, -6, -2, 2, 6, 9, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9,
 -3, 3, 9, 13, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 10, 9,
 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9,
 -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -16, -16, -16, -16, -16, -16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 11, 10, 10, 9,
 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15,
 -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2,
 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 13, 9, 3, -3, -9, -13, -16, -16, -13, -9, -3, 3, 9, 13, 16,
 16, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -15, -16,
 -16, -16, -16, -15, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15,
 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -10, -11, -12, -13, -13, -14, -14, -15, -15,
 -15, -16, -16, -11, -7, -1, 5, 10, 11, 9, 3, -3, -9, -11, -10, -5, 1, 7, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -15, -10, -2, 8, 14, 16, 12, 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 13, 12, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4,
 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11,
 -11, -12, -12, -12, -12, -12, -13, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -10, -2,
 8, 14, 16, 12, 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 11, 10, 10, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, 0, -1,
 -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -10, -10, -11, -11, -12, -12, -13, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -15, -14, -14, -14, -13, -13, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8, 8, 9, 10, 10, 11,
 11, 12, 12, 13, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, -15, -10, -2, 8, 14, 16, 12, 4, -4, -12, -16, -14, -8, 2, 10, 15, 15, 15, 15, 15, 14, 14, 13, 12, 12, 11,
 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13,
 -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13,
 13, 12, 11, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, -15, 0,0,0,0,0,0,1,0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8,
 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0,
 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15,
 -15, -15, -15, -15, -15, -15, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14,
 -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,
 5, 6, 7, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14,
 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 12, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -13, -13, -14, -15, -15, -15, -15, -15,
 -15, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14,
 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10,
 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4,
 -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, 0,
 -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14,
 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15,
 -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 10,
 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 3, 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15,
 15, 15, 15, 15, 14, 14, 13, 12, 12, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -13, -13, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15,
 -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10,
 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10,
 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 1, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3,
 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 14, 14, 14, 14, 14, 13, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7,
 -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, 0, 1, 1, 1, 1, 0, 0, -1,
 -1, -1, -1, -1, 0, 0, 1, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11,
 11, 10, 9, 8, 7, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -8, -9, -10, -11, -11, -12, -13, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -13, -13, -12, -11, -11, -10, -9, -9, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14,
 14, 14, 15, 15, 15, 15, 15, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 1, 1, 1, 3, 4, 5, 6, 7, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14,
 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12,
 -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 6, 5, 4, 3,
 2, 0, -1, -2, -4, -5, -6, -7, -8, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11,
 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, -1, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7,
 7, 8, 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 13, 13,
 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9,
 -9, -10, -10, -11, -11, -11, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -1, -1, -1, 0, 0, 1, 1, 1, 0, 0, -1, -1, -1,
 0, 1, 1, 2, 3, 4, 4, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 13, 13, 13, 12, 12, 11, 10, 10, 9, 8, 7, 6, 6,
 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14,
 -14, -13, -13, -12, -12, -11, -11, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, -2, -1, 0, 1, 2, 3, 4, 5, 5, 6, 7, 8, 9, 9, 10, 11, 11, 12, 13, 13, 13, 14, 14, 14, 15, 15,
 15, 15, -1, -1, -1, -1, 0, 1, 1, 1, 0, -1, -1, -1, -1, 0, 1, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9,
 8, 7, 6, 5, 3, 2, 1, 0, -2, -3, -4, -5, -6, -8, -9, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7,
 -5, -4, -3, -2, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 7, 6, 5, 4, 3, 1, 0, -1, -2, -4, -5,
 -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -15, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -3,
 -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 1, 2, 2,
 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2,
 -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0,
 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -2, -2, -2, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -3,
 -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2,
 -2, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6,
 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 1, 0, 0, 0, 0, 0, -1, 0, 0,
 0, 0, 0, 0, 1, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
 7, 7, 7, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -4, -4,
 -4, -4, -4, -5, -5, -5, -5, -5, -5, -6, -6, -6, -6, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, 1, 1, 0, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 0, 1, 1, 2,
 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 3, 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -2, -3,
 -3, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6, -6, -6, -6, -6, -5, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1,
 -1, 0, 0, 1, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 1, 1, 1, 0, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6,
 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -5, -6, -6, -6, -7, -7, -7, -7, -7, -7, -7, -7, -7, -6,
 -6, -6, -5, -5, -5, -4, -4, -3, -3, -2, -1, -1, 0, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 5, 5, 5, 4, 4, 3, 2, 2, 1, 1, 0, -1, -1,
 -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -7, -7, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
 10, 0, 0, 0, 0, 1, 0, 0, 0, 0, -1, 0, 0, 0, 0, 1, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 6, 6, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 9, 10, 10,
 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -2, -2,
 -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -7, -7, -7, -7, -8, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -1, -1, 0, 1,
 1, 1, 0, 0, -1, -1, -1, 0, 0, 1, 1, 1, 1, 2, 2, 3, 3, 4, 5, 5, 6, 6, 6, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 8, 8, 7, 7, 6, 6, 5,
 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -2, -2, -3, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -9, -9, -9, -9, -9,
 -8, -8, -8, -7, -7, -6, -6, -5, -5, -4, -4, -3, -3, -2, -2, -1, 0, 0, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 9, 9, 10, 10, 10, -1, -1, 0, 1, 1, 1,
 1, 0, -1, -1, -1, 0, 1, 1, 1, 1, 2, 2, 3, 4, 5, 5, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 10, 9, 9, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 1, 0, -1, -2, -3, -3, -4, -5,
 -6, -6, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -6, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 6, 7, 8, 8, 8, 9, 9,
 9, 10, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -5, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, 1,8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10, 10,
 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3,
 -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5,
 -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9,
 -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9,
 8, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9,
 -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7,
 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6,
 -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -15, -15, -14, -13, -12, -11, -10, -8, -7, -6, -4, -2, -1, 1, 2, 4, 6, 7, 8,
 10, 11, 12, 13, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11,
 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7,
 -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15,
 -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14,
 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8,
 -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13,
 -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15,
 15, 15, 15, 16, 16, 16, 16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 15,
 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16,
 -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16,
 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16,
 11, 11, 10, 9, 7, 5, 3, 1, -1, -3, -5, -7, -9, -10, -11, -11, -11, -11, -10, -9, -7, -5, -3, -1, 1, 3, 5, 7, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8,
 -10, -12, -14, -15, -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14,
 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2,
 -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14,
 -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15,
 -16, -16, -15, -14, -12, -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4,
 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16,
 -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8,
 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 15, 14, 12, 10, 8, 4, 2, -2, -4, -8, -10, -12, -14, -15, -16, -16, -15, -14, -12,
 -10, -8, -4, -2, 2, 4, 8, 10, 12, 14, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12,
 -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10,
 -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -11, -10, -8, -6, -3, 1, 4, 7, 9, 11, 11, 11, 10, 8, 5, 2, -2, -5, -8, -10, -11, -11, -11, -9, -7, -4, 0, 3,
 6, 8, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, -16, -15, -12, -8, -4, 1, 6, 10, 13, 15, 16, 16, 14, 11, 7, 2, -2, -7, -11, -14, -16, -16, -15, -13, -10, -6, -1, 4, 8, 12, 15, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5,
 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11,
 -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -14,
 -12, -8, -4, 1, 5, 9, 13, 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13,
 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15,
 -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4,
 -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, -16, -14, -12, -8, -4, 1, 5, 9, 13,
 15, 16, 15, 14, 11, 7, 2, -2, -7, -11, -14, -15, -16, -15, -13, -9, -5, -1, 4, 8, 12, 14, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0,
 -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7,
 -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4,
 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0,0,0,0,0,0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8,
 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12,
 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11,
 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -12, -12, -12,
 -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9,
 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15,
 -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15,
 15, 15, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16,
 16, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16,
 -15, -15, -14, -14, -13, -12, -11, -11, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13,
 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8,
 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14,
 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0,
 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12,
 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 11,
 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -12, -12, -12, -13,
 -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0,
 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11,
 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16,
 -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14,
 15, 15, 15, 15, 16, 16, 16, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11,
 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15,
 -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -11, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16, 16,
 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6,
 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13,
 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 8,
 8, 9, 9, 10, 10, 10, 11, 11, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14,
 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9,
 -10, -10, -11, -11, -11, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
 -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 15, 15, 15, 15,
 15, 14, 14, 13, 13, 12, 11, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -15, -15, -15,
 -15, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9,
 10, 11, 11, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 3,
 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 15, 15, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12,
 -13, -14, -14, -15, -15, -15, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 10, 10, 11, 12, 13, 14,
 14, 15, 15, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -13, -13, -14, -15, -15, -15, -16,
 -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12,
 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4,
 5, 6, 6, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 12, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14,
 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8,
 -8, -9, -9, -10, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, 0, -1, -1, -1, -1, -1, -1, 0, 0,
 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 2, 3, 4, 5, 6, 6, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 11, 11, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -3, -4, -5, -6, -7, -8, -9, -9, -10, -11, -12, -12, -13, -13, -14, -14,
 -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -11, -10, -9, -9, -8, -7, -6, -5, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6,
 7, 7, 8, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0,
 1, 1, 1, 1, 3, 4, 5, 6, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 13, 12, 12, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -5, -7, -8, -9,
 -10, -11, -12, -13, -13, -14, -15, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 1, 2, 3, 5, 6, 7, 8, 9, 10, 11,
 12, 13, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -7, -9, -10, -11, -12, -12, -13, -14, -14, -15,
 -15, -15, -15, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, -1, -1, -1, -1,
 -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0,
 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 1, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
 0, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -4, 1, 1, 1,
 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4,
 -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3,
 4, 4, 4, 4, 4, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4, 4, 4,
 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -3, -3, -3, -3, -3, -3,
 -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2,
 -2, -2, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4,
 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
 5, 5, 5, 5, 5, 5, 5, -1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3,
 3, 4, 4, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 2, 2,
 2, 2, 2, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -4, -4, -4, -4, -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3,
 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3, -3, -3,
 -4, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -4, -3, -3, -3, -2, -2, -2, -2, -1, -1, -1, 0, 0, 0, 1, 1, 1,
 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, -1, -1, -1, -1, -1, -1, -1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0,
 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 0, 0, -1, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -5,
 -5, -5, -5, -5, -5, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4, 4, 4, 3,
 3, 3, 2, 2, 1, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -3, -4, -4, -4, -5, -5, -5, -5, -5, -5, -5, 1,8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10, 10,
 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3, -3,
 -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -10, -11, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9,
 -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4,
 -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7, -8, -8,
 -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6, 7, 7,
 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9, -9,
 -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -9, -9, -9, -8, -8, -7, -7, -6, -6, -5, -5, -5, -4, -3, -3, -2, -2,
 -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -15, -15, -15, -15,
 -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13,
 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12,
 12, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6,
 -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15,
 -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8,
 -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16,
 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12,
 -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -10,
 -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16,
 -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6,
 7, 7, 8, 8, 9, 10, 10, 11, 11, 12, 12, 13, 14, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1,
 -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -10, -9, -8, -7, -6,
 -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2,
 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, 11, 11, 11, 11, 10, 10, 9, 8, 8, 7, 6, 5, 4, 3, 2, 1, 0, -2, -3,
 -4, -5, -6, -7, -8, -8, -9, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -8, -8, -7, -6, -5, -4, -3, -2, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 10, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 16, 16, 16, 15, 15, 14, 13, 12, 11, 10, 8, 7, 6, 4, 2, 1, -1, -2, -4, -6, -7, -8, -10, -11, -12, -13, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15,
 -15, -14, -13, -12, -11, -10, -8, -7, -6, -4, -2, -1, 1, 2, 4, 6, 7, 8, 10, 11, 12, 13, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15,
 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0,
 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13,
 -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7,
 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5,
 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2,
 -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15,
 -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12,
 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 9, 8, 7, 5, 4, 2, 1, -1, -2, -4, -5, -7, -8, -9, -11, -12, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -9, -8, -7, -5, -4, -2, -1, 1, 2, 4, 5, 7, 8, 9, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 15, 15, 15,
 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16,
 -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16,
 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -11,
 -11, -10, -10, -9, -8, -6, -5, -3, -2, 0, 1, 3, 5, 6, 7, 9, 10, 10, 11, 11, 11, 11, 11, 10, 9, 8, 7, 6, 4, 3, 1, -1, -2, -4, -5, -7, -8, -9, -10, -11, -11, -11, -11,
 -11, -10, -10, -9, -7, -6, -5, -3, -1, 0, 2, 4, 5, 7, 8, 9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -15, -14, -13, -11, -9, -7, -5, -3, 0, 2, 4, 7, 9, 10, 12, 14, 15, 15, 16, 16,
 16, 15, 14, 13, 12, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -12, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -9, -7, -4, -2, 0, 3, 5, 7, 9, 11, 13, 14, 15, 16,
 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8,
 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10,
 -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16,
 -16, -16, -16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3, 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -11,
 -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7, -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12,
 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15,
 -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2,
 -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, -16, -15, -15, -14, -12, -11, -9, -7, -5, -3,
 0, 2, 4, 7, 8, 10, 12, 14, 15, 15, 16, 16, 16, 15, 14, 13, 11, 10, 8, 6, 4, 1, -1, -4, -6, -8, -10, -11, -13, -14, -15, -16, -16, -16, -15, -15, -14, -12, -10, -8, -7,
 -4, -2, 0, 3, 5, 7, 9, 11, 12, 14, 15, 15, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12,
 -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8,
 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10,
 -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0,0,0,0,0,0,1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5,
 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12,
 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14,
 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8,
 7, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14,
 -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16,
 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13,
 -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0,
 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16,
 16, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16,
 -15, -15, -14, -14, -13, -12, -11, -11, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13,
 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2,
 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11,
 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1,
 -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9,
 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14,
 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10,
 -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 9, 8, 7, 6, 5,
 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14,
 -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16,
 16, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9,
 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -11, -9, -8, -7, -6,
 -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6,
 -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5,
 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13,
 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16,
 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14,
 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7,
 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6, -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14,
 -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15,
 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13, 12, 12, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11,
 -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4,
 -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9,
 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14,
 -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -11, -11, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16,
 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5, 4, 3, 2, 0, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 5, 5, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10,
 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12, 12, 12, 12, 12, 12, 12, 12, 13, 13, 13, 13, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 15, 15,
 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2,
 3, 4, 4, 5, 5, 6, 6, 7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 13, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15,
 15, 15, 15, 15, 15, 14, 14, 14, 14, 13, 13, 12, 12, 12, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 0, 0, -1, -1, -2, -2, -3, -4, -4, -5, -5, -6,
 -6, -7, -7, -8, -8, -9, -9, -10, -10, -11, -11, -12, -12, -12, -13, -13, -13, -14, -14, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, 0, 0, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 3, 4, 5, 6, 7, 7, 8, 9, 10, 11, 11, 12, 13, 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 13,
 12, 12, 11, 10, 9, 9, 8, 7, 6, 5, 4, 3, 2, 1, 0, -1, -2, -2, -4, -4, -5, -6, -7, -8, -9, -10, -10, -11, -12, -12, -13, -14, -14, -14, -15, -15, -15, -16, -16, -16, -16,
 -16, -16, -16, -15, -15, -15, -14, -14, -14, -13, -12, -12, -11, -11, -10, -9, -8, -7, -6, -5, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 8, 9, 10, 11, 11, 12, 13,
 13, 14, 14, 15, 15, 15, 15, 16, 16, 16, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0,
 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 3, 4, 5, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 15, 15,
 15, 14, 13, 13, 12, 11, 10, 9, 7, 6, 5, 4, 2, 1, 0, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13,
 -12, -11, -11, -9, -8, -7, -6, -5, -3, -2, -1, 1, 2, 4, 5, 6, 7, 8, 10, 11, 12, 12, 13, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 5,
 4, 3, 2, 0, -1, -2, -4, -5, -6, -8, -9, -10, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, 1,0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0,
 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0,
 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1,
 -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1,
 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2,
 -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -3, -3, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1,
 -1, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3,
 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3,
 -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
 3, 3, 3, 3, 3, -1, -1, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 0, 0, 0,
 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, -3, -3, -3, -2, -2, -2, -2, -2, -2, -2, -2, -1, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2,
 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, -1, -1, -1, -1, -1, -2, -2, -2, -2, -2, -2, -2, -3, -3, -3, -3, 1,8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 10,
 10, 10, 10, 9, 9, 9, 9, 9, 9, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 6, 6, 6, 5, 5, 5, 5, 4, 4, 4, 3, 3, 3, 3, 2, 2, 2, 2, 1, 1, 1, 0, 0, 0, 0, -1, -1, -1, -1, -2, -2, -2, -3,
 -3, -3, -3, -4, -4, -4, -4, -5, -5, -5, -5, -6, -6, -6, -6, -7, -7, -7, -7, -7, -8, -8, -8, -8, -8, -9, -9, -9, -9, -9, -9, -10, -10, -10, -10, -10, -10, -10, -10, -11,
 -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3,
 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11,
 -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2, -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8,
 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 2, 2, 1, 0, -1, -2, -3, -3, -4, -5, -6, -6, -7,
 -8, -8, -9, -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -9, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 5, 5, 6,
 7, 7, 8, 9, 9, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 9, 9, 8, 8, 7, 6, 6, 5, 4, 3, 3, 2, 1, 0, -1, -1, -2, -3, -4, -5, -5, -6, -7, -7, -8, -9,
 -9, -10, -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -10, -10, -10, -10, -9, -9,
 -9, -9, -9, -9, -8, -8, -8, -8, -8, -7, -7, -7, -7, -7, -6, -6, -6, -6, -5, -5, -5, -5, -4, -4, -4, -4, -3, -3, -3, -3, -2, -2, -2, -1, -1, -1, -1, 0, 0, 0, 0, 1, 1, 1,
 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 6, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 9, 9, 9, 9, 9, 9, 10, 10, 10, 10, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13,
 -13, -13, -12, -12, -12, -12, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1,
 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15,
 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12,
 12, 12, 12, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5,
 -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15,
 -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -15, -15, -15, -14, -14, -14,
 -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6, -5, -5, -4, -4, -4, -3, -3, -2,
 -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12, 13, 13, 13, 13, 14, 14,
 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14, 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8, 7,
 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14, -14, -15, -15, -15, -15, -16, -16, -16, -16,
 -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -9, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4,
 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15,
 -15, -15, -15, -14, -14, -14, -14, -14, -14, -13, -13, -13, -13, -12, -12, -12, -12, -11, -11, -11, -11, -10, -10, -10, -9, -9, -9, -8, -8, -8, -7, -7, -7, -6, -6, -6,
 -5, -5, -4, -4, -4, -3, -3, -2, -2, -2, -1, -1, -1, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 8, 9, 9, 9, 10, 10, 10, 11, 11, 11, 11, 12, 12,
 12, 12, 13, 13, 13, 13, 14, 14, 14, 14, 14, 14, 15, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8,
 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12,
 -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11,
 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -16, -16, -16, -16, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10,
 10, 10, 9, 9, 9, 8, 8, 7, 7, 7, 6, 6, 5, 5, 4, 4, 3, 3, 2, 1, 1, 0, 0, -1, -1, -2, -2, -3, -3, -4, -5, -5, -6, -6, -6, -7, -7, -8, -8, -9, -9, -9, -9, -10, -10, -10,
 -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -10, -9, -9, -9, -9, -8, -8, -7, -7, -6, -6, -6, -5, -5, -4, -3, -3, -2, -2,
 -1, -1, 0, 0, 1, 1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 16, 16, 16, 16, 16, 15, 15, 15, 15, 14, 14,
 14, 13, 13, 12, 12, 11, 10, 10, 9, 9, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -9, -9, -10, -10, -11, -12, -12, -13, -13, -14, -14,
 -14, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -14, -14, -14, -13, -13, -12, -12, -11, -10, -10, -9, -9, -8, -7, -7, -6,
 -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 9, 9, 10, 10, 11, 12, 12, 13, 13, 14, 14, 14, 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12, 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4,
 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6, -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11,
 -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, 16, 16, 16,
 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11,
 -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10,
 -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16,
 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9,
 -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12,
 -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15,
 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7, 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7,
 -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13,
 -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6, 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15,
 15, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7, 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14,
 -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10, 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -11, -12,
 -13, -14, -14, -15, -15, -15, -16, -16, -16, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -7, -7, -6, -5, -5, -4, -3, -2, -1, -1, 0, 1, 2, 3, 3, 4, 5, 6, 6, 7, 8,
 8, 9, 9, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 10, 10, 10, 9, 9, 8, 7, 7, 6, 5, 5, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -9,
 -10, -10, -11, -11, -11, -11, -11, -11, -11, -11, -11, -11, -10, -10, -10, -9, -9, -8, -8, -7, -6, -6, -5, -4, -3, -3, -2, -1, 0, 1, 2, 2, 3, 4, 5, 6, 6, 7, 8, 8, 9, 9,
 10, 10, 10, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
 11, 11, 11, 11, 11, 11, 11, 11, 11, -16, -16, -16, -16, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11,
 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12,
 -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9,
 10, 11, 12, 12, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 15, 15, 15, 14, 14, 14, 14, 14, 14, 13, 13, 13, 13, 12, 12, 12,
 12, 11, 11, 11, 11, 10, 10, 10, 9, 9, 9, 8, 8, 8, 7, 7, 7, 6, 6, 6, 5, 5, 4, 4, 4, 3, 3, 2, 2, 2, 1, 1, 1, 0, 0, -1, -1, -1, -2, -2, -2, -3, -3, -4, -4, -4, -5, -5, -6,
 -6, -6, -7, -7, -7, -8, -8, -8, -9, -9, -9, -10, -10, -10, -11, -11, -11, -11, -12, -12, -12, -12, -13, -13, -13, -13, -14, -14, -14, -14, -14, -14, -15, -15, -15, -15,
 -15, -15, -15, -15, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7, -6, -4, -3, -2, -1, 0, 1,
 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1, -1, -2, -3, -4, -5,
 -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10, -9, -8, -7, -6, -5, -4, -2,
 -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 15, 15, 14, 14, 13, 13, 12, 12, 12, 11, 10, 10, 9, 8, 8, 7, 7,
 6, 5, 4, 4, 3, 2, 1, 0, 0, -1, -2, -3, -4, -4, -5, -6, -7, -7, -8, -8, -9, -10, -10, -11, -12, -12, -12, -13, -13, -14, -14, -15, -15, -15, -15, -15, -16, -16, -16, -16,
 -16, -16, -16, -16, -15, -15, -15, -15, -15, -14, -14, -13, -13, -12, -12, -12, -11, -10, -10, -9, -8, -8, -7, -7, -6, -5, -4, -4, -3, -2, -1, 0, 0, 1, 2, 3, 4, 4, 5, 6,
 7, 7, 8, 8, 9, 10, 10, 11, 12, 12, 12, 13, 13, 14, 14, 15, 15, 15, 15, 15, 16, 16, 16, 16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -11, -11, -10, -9, -8, -7,
 -6, -4, -3, -2, -1, 0, 1, 2, 4, 5, 6, 7, 8, 9, 10, 11, 12, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 14, 14, 13, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3,
 2, 1, -1, -2, -3, -4, -5, -6, -7, -8, -9, -10, -11, -12, -13, -13, -14, -14, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -15, -14, -14, -13, -12, -12, -11, -10,
 -9, -8, -7, -6, -5, -4, -2, -1, 0, 1, 2, 3, 4, 6, 7, 8, 9, 10, 11, 11, 12, 13, 14, 14, 15, 15, 15, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 11, 11, 10, 9, 8, 7,
 6, 4, 3, 2, 1, 0, -1, -2, -4, -5, -6, -7, -8, -9, -10, -11, -12, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, -16, -16, -16, -15, -15, -14, -14, -13, -13, -12, -11,
 -10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 13, 14, 14, 15, 15, 16, 16, 16, 16, 16, 16, 15, 15, 15, 14, 14, 13, 12, 12, 11, 10,
 9, 8, 7, 6, 5, 4, 2, 1, 0, -1, -2, -3, -4, -6, -7, -8, -9, -10, -11, -11, -12, -13, -14, -14, -15, -15, -15, -16, -16, -16, 0,0,0,0,0,0,0,0,
};
#endif
#endif

const char* nalUnitTypeToString(NalUnitType type)
{
  switch (type)
  {
  case NAL_UNIT_CODED_SLICE_TRAIL:      return "TRAIL";
  case NAL_UNIT_CODED_SLICE_STSA:       return "STSA";
  case NAL_UNIT_CODED_SLICE_RADL:       return "RADL";
  case NAL_UNIT_CODED_SLICE_RASL:       return "RASL";
  case NAL_UNIT_CODED_SLICE_IDR_W_RADL: return "IDR_W_RADL";
  case NAL_UNIT_CODED_SLICE_IDR_N_LP:   return "IDR_N_LP";
  case NAL_UNIT_CODED_SLICE_CRA:        return "CRA";
  case NAL_UNIT_CODED_SLICE_GDR:        return "GDR";
  case NAL_UNIT_DCI:                    return "DCI";
  case NAL_UNIT_VPS:                    return "VPS";
  case NAL_UNIT_SPS:                    return "SPS";
  case NAL_UNIT_PPS:                    return "PPS";
  case NAL_UNIT_PREFIX_APS:             return "Prefix APS";
  case NAL_UNIT_SUFFIX_APS:             return "Suffix APS";
  case NAL_UNIT_PH:                     return "PH";
  case NAL_UNIT_ACCESS_UNIT_DELIMITER:  return "AUD";
  case NAL_UNIT_EOS:                    return "EOS";
  case NAL_UNIT_EOB:                    return "EOB";
  case NAL_UNIT_PREFIX_SEI:             return "Prefix SEI";
  case NAL_UNIT_SUFFIX_SEI:             return "Suffix SEI";
  case NAL_UNIT_FD:                     return "FD";
  default:                              return "UNK";

  }
}

class ScanGenerator
{
private:
  uint32_t m_line, m_column;
  const uint32_t m_blockWidth, m_blockHeight;
  const uint32_t m_stride;
  const CoeffScanType m_scanType;

public:
  ScanGenerator(uint32_t blockWidth, uint32_t blockHeight, uint32_t stride, CoeffScanType scanType)
    : m_line(0), m_column(0), m_blockWidth(blockWidth), m_blockHeight(blockHeight), m_stride(stride), m_scanType(scanType)
  { }

  uint32_t GetCurrentX() const { return m_column; }
  uint32_t GetCurrentY() const { return m_line; }

  uint32_t GetNextIndex(uint32_t blockOffsetX, uint32_t blockOffsetY)
  {
    const uint32_t rtn = ((m_line + blockOffsetY) * m_stride) + m_column + blockOffsetX;

    //advance line and column to the next position
    switch (m_scanType)
    {
      //------------------------------------------------

      case SCAN_DIAG:

        if ((m_column == m_blockWidth - 1) || (m_line == 0)) //if we reach the end of a rank, go diagonally down to the next one
        {
          m_line += m_column + 1;
          m_column = 0;

          if (m_line >= m_blockHeight) //if that takes us outside the block, adjust so that we are back on the bottom row
          {
            m_column += m_line - (m_blockHeight - 1);
            m_line = m_blockHeight - 1;
          }
        }
        else
        {
          m_column++;
          m_line--;
        }
        break;

      case SCAN_TRAV_HOR:
        if (m_line % 2 == 0)
        {
          if (m_column == (m_blockWidth - 1))
          {
            m_line++;
            m_column = m_blockWidth - 1;
          }
          else
          {
            m_column++;
          }
        }
        else
        {
          if (m_column == 0)
          {
            m_line++;
            m_column = 0;
          }
          else
          {
            m_column--;
          }
        }
        break;

      case SCAN_TRAV_VER:
        if (m_column % 2 == 0)
        {
          if (m_line == (m_blockHeight - 1))
          {
            m_column++;
            m_line = m_blockHeight - 1;
          }
          else
          {
            m_line++;
          }
        }
        else
        {
          if (m_line == 0)
          {
            m_column++;
            m_line = 0;
          }
          else
          {
            m_line--;
          }
        }
        break;
      //------------------------------------------------

      default:

        THROW("ERROR: Unknown scan type \"" << m_scanType << "\"in ScanGenerator::GetNextIndex");
        break;
    }

    return rtn;
  }
};
const int8_t g_BcwLog2WeightBase = 3;
const int8_t g_BcwWeightBase = (1 << g_BcwLog2WeightBase);
const int8_t g_BcwWeights[BCW_NUM] = { -2, 3, 4, 5, 10 };
const int8_t g_BcwSearchOrder[BCW_NUM] = { BCW_DEFAULT, BCW_DEFAULT - 2, BCW_DEFAULT + 2, BCW_DEFAULT - 1, BCW_DEFAULT + 1 };
int8_t g_BcwCodingOrder[BCW_NUM];
int8_t g_BcwParsingOrder[BCW_NUM];

int8_t getBcwWeight(uint8_t bcwIdx, uint8_t uhRefFrmList)
{
  // Weghts for the model: P0 + w * (P1 - P0) = (1-w) * P0 + w * P1
  // Retuning  1-w for P0 or w for P1
  return (uhRefFrmList == REF_PIC_LIST_0 ? g_BcwWeightBase - g_BcwWeights[bcwIdx] : g_BcwWeights[bcwIdx]);
}

void resetBcwCodingOrder(bool bRunDecoding, const CodingStructure &cs)
{
  // Form parsing order: { BCW_DEFAULT, BCW_DEFAULT+1, BCW_DEFAULT-1, BCW_DEFAULT+2, BCW_DEFAULT-2, ... }
  g_BcwParsingOrder[0] = BCW_DEFAULT;
  for (int i = 1; i <= (BCW_NUM >> 1); ++i)
  {
    g_BcwParsingOrder[2 * i - 1] = BCW_DEFAULT + (int8_t)i;
    g_BcwParsingOrder[2 * i] = BCW_DEFAULT - (int8_t)i;
  }

  // Form encoding order
  if (!bRunDecoding)
  {
    for (int i = 0; i < BCW_NUM; ++i)
    {
      g_BcwCodingOrder[(uint32_t)g_BcwParsingOrder[i]] = i;
    }
  }
}

uint32_t deriveWeightIdxBits(uint8_t bcwIdx) // Note: align this with TEncSbac::codeBcwIdx and TDecSbac::parseBcwIdx
{
  uint32_t numBits = 1;
  uint8_t  bcwCodingIdx = (uint8_t)g_BcwCodingOrder[bcwIdx];

  if (BCW_NUM > 2 && bcwCodingIdx != 0)
  {
    uint32_t prefixNumBits = BCW_NUM - 2;
    uint32_t step = 1;
    uint8_t  prefixSymbol = bcwCodingIdx;

    // Truncated unary code
    uint8_t idx = 1;
    for (int ui = 0; ui < prefixNumBits; ++ui)
    {
      if (prefixSymbol == idx)
      {
        ++numBits;
        break;
      }
      else
      {
        ++numBits;
        idx += step;
      }
    }
  }
  return numBits;
}

uint32_t g_log2SbbSize[MAX_CU_DEPTH + 1][MAX_CU_DEPTH + 1][2] =
//===== luma/chroma =====
{
#if TU_256
  { { 0,0 },{ 0,1 },{ 0,2 },{ 0,3 },{ 0,4 },{ 0,4 },{ 0,4 },{ 0,4 },{ 0,4 } },
  { { 1,0 },{ 1,1 },{ 1,1 },{ 1,3 },{ 1,3 },{ 1,3 },{ 1,3 },{ 1,3 },{ 1,3 } },
  { { 2,0 },{ 1,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 3,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } }
#else
  { { 0,0 },{ 0,1 },{ 0,2 },{ 0,3 },{ 0,4 },{ 0,4 },{ 0,4 },{ 0,4 } },
  { { 1,0 },{ 1,1 },{ 1,1 },{ 1,3 },{ 1,3 },{ 1,3 },{ 1,3 },{ 1,3 } },
  { { 2,0 },{ 1,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 3,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } },
  { { 4,0 },{ 3,1 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 },{ 2,2 } }
#endif
};

#if TU_256
TMatrixCoeff g_trCoreDCT2P256[256][256];
TMatrixCoeff g_trCoreDCT8P256[256][256];
TMatrixCoeff g_trCoreDST7P256[256][256];
#endif
#if JVET_W0103_INTRA_MTS
TMatrixCoeff g_aiTr2[NUM_TRANS_TYPE][2][2];
TMatrixCoeff g_aiTr4[NUM_TRANS_TYPE][4][4];
TMatrixCoeff g_aiTr8[NUM_TRANS_TYPE][8][8];
TMatrixCoeff g_aiTr16[NUM_TRANS_TYPE][16][16];
TMatrixCoeff g_aiTr32[NUM_TRANS_TYPE][32][32];
TMatrixCoeff g_aiTr64[NUM_TRANS_TYPE][64][64];
TMatrixCoeff g_aiTr128[NUM_TRANS_TYPE][128][128];
TMatrixCoeff g_aiTr256[NUM_TRANS_TYPE][256][256];

const uint8_t g_aucIpmToTrSet[16][36] =
{
  //0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 MIP
  { 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4 }, //4x4
  { 5, 5, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 9 }, //4x8
  {10,10,11,11,11,11,11,11,11,11,11,11,11,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,14 }, //4x16
  {15,15,16,16,16,16,16,16,16,16,16,16,16,17,17,17,17,17,17,17,17,17,17,17,18,18,18,18,18,18,18,18,18,18,18,19 }, //4x32
  {20,20,21,21,21,21,21,21,21,21,21,21,21,22,22,22,22,22,22,22,22,22,22,22,23,23,23,23,23,23,23,23,23,23,23,24 }, //8x4
  {25,25,26,26,26,26,26,26,26,26,26,26,26,27,27,27,27,27,27,27,27,27,27,27,28,28,28,28,28,28,28,28,28,28,28,29 }, //8x8
  {30,30,31,31,31,31,31,31,31,31,31,31,31,32,32,32,32,32,32,32,32,32,32,32,33,33,33,33,33,33,33,33,33,33,33,34 }, //8x16
  {35,35,36,36,36,36,36,36,36,36,36,36,36,37,37,37,37,37,37,37,37,37,37,37,38,38,38,38,38,38,38,38,38,38,38,39 }, //8x32
  {40,40,41,41,41,41,41,41,41,41,41,41,41,42,42,42,42,42,42,42,42,42,42,42,43,43,43,43,43,43,43,43,43,43,43,44 }, //16x4
  {45,45,46,46,46,46,46,46,46,46,46,46,46,47,47,47,47,47,47,47,47,47,47,47,48,48,48,48,48,48,48,48,48,48,48,49 }, //16x8
  {50,50,51,51,51,51,51,51,51,51,51,51,51,52,52,52,52,52,52,52,52,52,52,52,53,53,53,53,53,53,53,53,53,53,53,54 }, //16x16
  {55,55,56,56,56,56,56,56,56,56,56,56,56,57,57,57,57,57,57,57,57,57,57,57,58,58,58,58,58,58,58,58,58,58,58,59 }, //16x32
  {60,60,61,61,61,61,61,61,61,61,61,61,61,62,62,62,62,62,62,62,62,62,62,62,63,63,63,63,63,63,63,63,63,63,63,64 }, //32x4
  {65,65,66,66,66,66,66,66,66,66,66,66,66,67,67,67,67,67,67,67,67,67,67,67,68,68,68,68,68,68,68,68,68,68,68,69 }, //32x8
  {70,70,71,71,71,71,71,71,71,71,71,71,71,72,72,72,72,72,72,72,72,72,72,72,73,73,73,73,73,73,73,73,73,73,73,74 }, //32x16
  {75,75,76,76,76,76,76,76,76,76,76,76,76,77,77,77,77,77,77,77,77,77,77,77,78,78,78,78,78,78,78,78,78,78,78,79 }, //32x32
};
const int8_t g_aiIdLut[3][3] =
{
  { 8, 6, 4 },{ 8, 8, 6 },{ 4, 2, -1 }
};
const uint8_t g_aucTrIdxToTr[25][2] =
{
    { DCT8, DCT8 },{ DCT8, DST7 },{ DCT8, DCT5 },{ DCT8, DST4 }, {DCT8, DST1},
    { DST7, DCT8 },{ DST7, DST7 },{ DST7, DCT5 },{ DST7, DST4 }, {DST7, DST1},
    { DCT5, DCT8 },{ DCT5, DST7 },{ DCT5, DCT5 },{ DCT5, DST4 }, {DCT5, DST1},
    { DST4, DCT8 },{ DST4, DST7 },{ DST4, DCT5 },{ DST4, DST4 }, {DST4, DST1},
    { DST1, DCT8 },{ DST1, DST7 },{ DST1, DCT5 },{ DST1, DST4 }, {DST1, DST1},
};
#if JVET_Y0142_ADAPT_INTRA_MTS
const uint8_t g_aucTrSet[80][6] =
{
  //T0:0,  1,  2,  3,  4,  5,
  { 18, 24, 17, 23, 8, 12},
  //T1:0,  1,  2,  3,  4,  5,
  { 18, 3, 7, 22,  0, 16},
  //T2:0,  1,  2,  3,  4,  5,
  { 18, 2, 17, 22, 3, 23},
  //T3:0,  1,  2,  3,  4,  5,
  { 18, 3, 15, 17, 12, 23},
  //T4:0,  1,  2,  3,  4,  5,
  { 18, 12, 3, 19, 10, 13},
  //T5:0,  1,  2,  3,  4,  5,
  { 18, 12, 19, 23, 13, 24},
  //T6:0,  1,  2,  3,  4,  5,
  { 18, 12, 17, 2, 3, 23},
  //T7:0,  1,  2,  3,  4,  5,
  { 18, 2, 17, 22, 12, 23},
  //T8:0,  1,  2,  3,  4,  5,
  { 18, 2, 11, 17, 22, 23},
  //T9:0,  1,  2,  3,  4,  5,
  { 18, 12, 19, 23, 3, 10},
  //T10:0,  1,  2,  3,  4, 5,
  { 16, 12, 13, 24, 7, 8},
  //T11:0,  1,  2,  3,  4,  5,
  { 16, 2, 11, 23, 12, 18},
  //T12:0,  1,  2,  3,  4,  5,
  { 13, 17, 2, 22, 12, 18},
  //T13:0,  1,  2,  3,  4,  5,
  { 17, 11, 2, 21, 12, 18},
  //T14:0,  1,  2,  3,  4,  5,
  { 16, 13, 19, 22, 3, 10},
  //T15:0,  1,  2,  3,  4,  5,
  { 18, 12, 13, 7, 14, 22},
  //T16:0,  1,  2,  3,  4,  5,
  { 16, 12, 11, 1, 18, 22},
  //T17:0,  1,  2,  3,  4,  5,
  { 17, 13,  3, 22, 12, 18},
  //T18:0,  1,  2,  3,  4,  5,
  {  6, 12,  1, 22, 13, 17},
  //T19:0,  1,  2,  3,  4,  5,
  { 16, 12, 13, 15, 2, 23},
  //T20:0,  1,  2,  3,  4,  5,
  { 18, 24, 23, 19, 12, 17},
  //T21:0,  1,  2,  3,  4,  5,
  { 18, 24, 2,  17,  0, 23},
  //T22:0,  1,  2,  3,  4,  5,
  { 17, 3,  4, 22, 2, 13},
  //T23:0,  1,  2,  3,  4,  5,
  { 18, 12, 19, 23, 3, 15},
  //T24:0,  1,  2,  3,  4,  5,
  { 18, 12, 19, 23, 3, 10},
  //T25:0,  1,  2,  3,  4,  5,
  {  6, 12, 18, 24, 13, 19},
  //T26:0,  1,  2,  3,  4,  5,
  {  6, 12,  2, 21, 13, 18},
  //T27:0,  1,  2,  3,  4,  5,
  { 17, 11,  1, 22, 2, 18},
  //T28:0,  1,  2,  3,  4,  5,
  { 16, 17, 3,  11, 12, 23},
  //T29:0,  1,  2,  3,  4,  5,
  {  8, 12, 19, 23, 11, 24},
  //T30:0,  1,  2,  3,  4,  5,
  { 16, 13, 7, 23, 12, 19},
  //T31:0,  1,  2,  3,  4,  5,
  {  6, 12, 1, 11, 18, 22},
  //T32:0,  1,  2,  3,  4,  5,
  { 17, 11, 1, 21, 12, 18},
  //T33:0,  1,  2,  3,  4,  5,
  {  6, 11, 17, 21, 12, 18},
  //T34:0,  1,  2,  3,  4,  5,
  {  8, 11, 14, 17, 12, 22},
  //T35:0,  1,  2,  3,  4,  5,
  {  6, 12, 11, 21, 14, 16},
  //T36:0,  1,  2,  3,  4,  5,
  {  6, 12, 11,  1, 17, 21},
  //T37:0,  1,  2,  3,  4,  5,
  {  6, 12, 11,  2, 17, 21},
  //T38:0,  1,  2,  3,  4,  5,
  {  6, 11, 21,  1, 12, 17},
  //T39:0,  1,  2,  3,  4,  5,
  {  16, 12, 11, 7, 1, 5},
  //T40:0,  1,  2,  3,  4,  5,
  {  8, 12, 19, 24, 11, 17},
  //T41:0,  1,  2,  3,  4,  5,
  { 18, 13, 1,  22, 2, 24},
  //T42:0,  1,  2,  3,  4,  5,
  {  6, 2, 17, 21, 19, 22},
  //T43:0,  1,  2,  3,  4,  5,
  { 16, 12, 11, 19, 8, 15},
  //T44:0,  1,  2,  3,  4,  5,
  {  8, 12, 17, 24, 13, 15},
  //T45:0,  1,  2,  3,  4,  5,
  {  6, 12, 19, 21, 17, 18},
  //T46:0,  1,  2,  3,  4,  5,
  {  6, 12, 13, 21, 2, 18},
  //T47:0,  1,  2,  3,  4,  5,
  { 16, 2,  17, 21, 1, 11},
  //T48:0,  1,  2,  3,  4,  5,
  {  6, 17, 19, 23, 12, 16},
  //T49:0,  1,  2,  3,  4,  5,
  {  6, 12, 14, 17, 8, 22},
  //T50:0,  1,  2,  3,  4,  5,
  {  6,  7, 11, 21, 9, 12},
  //T51:0,  1,  2,  3,  4,  5,
  { 16, 12, 11,  1, 7, 21},
  //T52:0,  1,  2,  3,  4,  5,
  {  6, 12, 11, 1, 17, 21},
  //T53:0,  1,  2,  3,  4,  5,
  {  6, 12, 11, 21, 1, 16},
  //T54:0,  1,  2,  3,  4,  5,
  {  8,  7,  9, 11, 12, 21},
  //T55:0,  1,  2,  3,  4,  5,
  {  6, 12, 7,  11, 14, 21},
  //T56:0,  1,  2,  3,  4,  5,
  {  6, 12, 7,  11, 1, 21},
  //T57:0,  1,  2,  3,  4,  5,
  { 16, 12, 11, 1, 2, 21},
  //T58:0,  1,  2,  3,  4,  5,
  {  6, 11, 17, 21, 1, 12},
  //T59:0,  1,  2,  3,  4,  5,
  {  6, 12, 7,  11, 9, 21},
  //T60:0,  1,  2,  3,  4,  5,
  { 18, 12, 14, 21, 6, 21},
  //T61:0,  1,  2,  3,  4,  5,
  { 16, 11, 1, 22,  2, 17},
  //T62:0,  1,  2,  3,  4,  5,
  { 16, 11, 1, 22,  2, 17},
  //T63:0,  1,  2,  3,  4,  5,
  { 16, 13, 15, 7, 14, 19},
  //T64:0,  1,  2,  3,  4,  5,
  {  8, 12, 1, 19, 16, 23},
  //T65:0,  1,  2,  3,  4,  5,
  {  6, 12, 7,  9, 13, 21},
  //T66:0,  1,  2,  3,  4,  5,
  {  6, 12, 13, 2,  7, 18},
  //T67:0,  1,  2,  3,  4,  5,
  { 16, 12, 1, 21, 11, 17},
  //T68:0,  1,  2,  3,  4,  5,
  { 16, 11, 7, 19, 12, 15},
  //T69:0,  1,  2,  3,  4,  5,
  {  8, 12, 7, 11, 14, 21},
  //T70:0,  1,  2,  3,  4,  5,
  {  6, 12, 7, 11, 8, 9},
  //T71:0,  1,  2,  3,  4,  5,
  { 6, 12,  7, 11, 2, 21},
  //T72:0,  1,  2,  3,  4,  5,
  {  6, 12, 1, 11, 21, 22},
  //T73:0,  1,  2,  3,  4,  5,
  {  6,  7, 11, 16, 9, 12},
  //T74:0,  1,  2,  3,  4,  5,
  { 6,  12,  7, 11, 9, 21},
  //T75:0,  1,  2,  3,  4,  5,
  { 6, 12,  7, 11, 13, 17},
  //T76:0,  1,  2,  3,  4,  5,
  {  6, 12, 11, 21, 2, 7},
  //T77:0,  1,  2,  3,  4,  5,
  {  6, 12, 1,  11, 2, 7},
  //T78:0,  1,  2,  3,  4,  5,
  { 6,  12, 7, 11, 16, 21},
  //T79:0,  1,  2,  3,  4,  5,
  { 6,  12, 7, 11, 9, 16},
};
#else
const uint8_t g_aucTrSet[80][4] =
{
//T0:0,  1,  2,  3,
{ 17, 18, 23, 24},
//T1:0,  1,  2,  3,
{  3,  7, 18, 22},
//T2:0,  1,  2,  3, 
{  2, 17, 18, 22},
//T3:0,  1,  2,  3,  
{  3, 15, 17, 18},
//T4:0,  1,  2,  3, 
{  3, 12, 18, 19},
//T5:0,  1,  2,  3, 
{ 12, 18, 19, 23},
//T6:0,  1,  2,  3,
{  2, 12, 17, 18},
//T7:0,  1,  2,  3, 
{  2, 17, 18, 22},
//T8:0,  1,  2,  3,  
{  2, 11, 17, 18},
//T9:0,  1,  2,  3, 
{ 12, 18, 19, 23},
//T10:0,  1, 2,   3, 
{ 12, 13, 16, 24},
//T11:0,  1,  2,  3, 
{  2, 11, 16, 23},
//T12:0,  1,  2,  3, 
{  2, 13, 17, 22},
//T13:0,  1,  2,  3,
{  2, 11, 17, 21},
//T14:0,  1,  2,  3, 
{ 13, 16, 19, 22},
//T15:0,  1,  2,  3,
{  7, 12, 13, 18},
//T16:0,  1,  2,  3, 
{  1, 11, 12, 16},
//T17:0,  1,  2,  3,
{  3, 13, 17, 22},
//T18:0,  1,  2,  3,  
{  1,  6, 12, 22},
//T19:0,  1,  2,  3,  
{ 12, 13, 15, 16},
//T20:0,  1,  2,  3,  
{ 18, 19, 23, 24},
//T21:0,  1,  2,  3, 
{  2, 17, 18, 24},
//T22:0,  1,  2,  3,
{  3,  4, 17, 22},
//T23:0,  1,  2,  3, 
{ 12, 18, 19, 23},
//T24:0,  1,  2,  3,
{ 12, 18, 19, 23},
//T25:0,  1,  2,  3,  
{  6, 12, 18, 24},
//T26:0,  1,  2,  3, 
{  2,  6, 12, 21},
//T27:0,  1,  2,  3, 
{  1, 11, 17, 22},
//T28:0,  1,  2,  3, 
{  3, 11, 16, 17},
//T29:0,  1,  2,  3, 
{  8, 12, 19, 23},
//T30:0,  1,  2,  3, 
{  7, 13, 16, 23},
//T31:0,  1,  2,  3, 
{  1,  6, 11, 12},
//T32:0,  1,  2,  3,
{  1, 11, 17, 21},
//T33:0,  1,  2,  3,
{  6, 11, 17, 21},
//T34:0,  1,  2,  3,
{  8, 11, 14, 17},
//T35:0,  1,  2,  3, 
{  6, 11, 12, 21},
//T36:0,  1,  2,  3, 
{  1,  6, 11, 12},
//T37:0,  1,  2,  3, 
{  2,  6, 11, 12},
//T38:0,  1,  2,  3,
{  1,  6, 11, 21},
//T39:0,  1,  2,  3,
{  7, 11, 12, 16},
//T40:0,  1,  2,  3,
{  8, 12, 19, 24},
//T41:0,  1,  2,  3,
{  1, 13, 18, 22},
//T42:0,  1,  2,  3,
{  2,  6, 17, 21},
//T43:0,  1,  2,  3,
{ 11, 12, 16, 19},
//T44:0,  1,  2,  3,
{  8, 12, 17, 24},
//T45:0,  1,  2,  3,
{  6, 12, 19, 21},
//T46:0,  1,  2,  3,
{  6, 12, 13, 21},
//T47:0,  1,  2,  3,
{  2, 16, 17, 21},
//T48:0,  1,  2,  3,
{  6, 17, 19, 23},
//T49:0,  1,  2,  3,
{  6, 12, 14, 17},
//T50:0,  1,  2,  3,
{  6,  7, 11, 21},
//T51:0,  1,  2,  3,  
{  1, 11, 12, 16},
//T52:0,  1,  2,  3,
{  1,  6, 11, 12},
//T53:0,  1,  2,  3,
{  6, 11, 12, 21},
//T54:0,  1,  2,  3,
{  7,  8,  9, 11},
//T55:0,  1,  2,  3,
{  6,  7, 11, 12},
//T56:0,  1,  2,  3,
{  6,  7, 11, 12},
//T57:0,  1,  2,  3,
{  1, 11, 12, 16},
//T58:0,  1,  2,  3,
{  6, 11, 17, 21},
//T59:0,  1,  2,  3,
{  6,  7, 11, 12},
//T60:0,  1,  2,  3,
{ 12, 14, 18, 21},
//T61:0,  1,  2,  3, 
{  1, 11, 16, 22},
//T62:0,  1,  2,  3,
{  1, 11, 16, 22},
//T63:0,  1,  2,  3, 
{  7, 13, 15, 16},
//T64:0,  1,  2,  3, 
{  1,  8, 12, 19},
//T65:0,  1,  2,  3,
{  6,  7,  9, 12},
//T66:0,  1,  2,  3,  
{  2,  6, 12, 13},
//T67:0,  1,  2,  3,  
{  1, 12, 16, 21},
//T68:0,  1,  2,  3, 
{  7, 11, 16, 19},
//T69:0,  1,  2,  3, 
{  7,  8, 11, 12},
//T70:0,  1,  2,  3,  
{  6,  7, 11, 12},
//T71:0,  1,  2,  3,  
{  6,  7, 11, 12},
//T72:0,  1,  2,  3,  
{  1,  6, 11, 12},
//T73:0,  1,  2,  3, 
{  6,  7, 11, 16},
//T74:0,  1,  2,  3,  
{  6,  7, 11, 12},
//T75:0,  1,  2,  3,  
{  6,  7, 11, 12},
//T76:0,  1,  2,  3, 
{  6, 11, 12, 21},
//T77:0,  1,  2,  3,
{  1,  6, 11, 12},
//T78:0,  1,  2,  3, 
{  6,  7, 11, 12},
//T79:0,  1,  2,  3, 
{  6,  7, 11, 12},
};
#endif
#endif
// initialize ROM variables
void initROM()
{
#if LMS_LINEAR_MODEL || TRANSFORM_SIMD_OPT
  // g_aucConvertToBit[ x ]: log2(x/4), if x=4 -> 0, x=8 -> 1, x=16 -> 2, ...
  // g_aucLog2[ x ]: log2(x), if x=1 -> 0, x=2 -> 1, x=4 -> 2, x=8 -> 3, x=16 -> 4, ...
  ::memset(g_aucLog2, 0, sizeof(g_aucLog2));
  int c = 0;
  for (int i = 0, n = 0; i <= MAX_CU_SIZE; i++)
  {
#if LMS_LINEAR_MODEL
    g_aucNextLog2[i] = i <= 1 ? 0 : c + 1;
#endif

    if (i == (1 << n))
    {
      c = n;
      n++;
    }

#if LMS_LINEAR_MODEL
    g_aucPrevLog2[i] = c;
#endif
    g_aucLog2[i] = c;
  }
#elif MMLM
  int c = 0;
  for (int i = 0, n = 0; i <= MAX_CU_SIZE; i++)
  {
    if (i == (1 << n))
    {
      c = n;
      n++;
    }

    g_aucPrevLog2[i] = c;
  }
#elif TU_256
  int c = 0;
#endif

  gp_sizeIdxInfo = new SizeIndexInfoLog2();
  gp_sizeIdxInfo->init(MAX_CU_SIZE);


  SizeIndexInfoLog2 sizeInfo;
  sizeInfo.init(MAX_CU_SIZE);

  // initialize scan orders
  for (uint32_t blockHeightIdx = 0; blockHeightIdx < sizeInfo.numAllHeights(); blockHeightIdx++)
  {
    for (uint32_t blockWidthIdx = 0; blockWidthIdx < sizeInfo.numAllWidths(); blockWidthIdx++)
    {
      const uint32_t blockWidth  = sizeInfo.sizeFrom(blockWidthIdx);
      const uint32_t blockHeight = sizeInfo.sizeFrom(blockHeightIdx);
      const uint32_t totalValues = blockWidth * blockHeight;

      //--------------------------------------------------------------------------------------------------

      //non-grouped scan orders

      for (uint32_t scanTypeIndex = 0; scanTypeIndex < SCAN_NUMBER_OF_TYPES; scanTypeIndex++)
      {
        const CoeffScanType scanType = CoeffScanType(scanTypeIndex);
        ScanElement *       scan     = nullptr;

        if (blockWidthIdx < sizeInfo.numWidths() && blockHeightIdx < sizeInfo.numHeights())
        {
          scan = new ScanElement[totalValues];
        }

        g_scanOrder[SCAN_UNGROUPED][scanType][blockWidthIdx][blockHeightIdx] = scan;

        if (scan == nullptr)
        {
          continue;
        }

        ScanGenerator fullBlockScan(blockWidth, blockHeight, blockWidth, scanType);

        for (uint32_t scanPosition = 0; scanPosition < totalValues; scanPosition++)
        {
          const int rasterPos = fullBlockScan.GetNextIndex( 0, 0 );
          const int posY      = rasterPos / blockWidth;
          const int posX      = rasterPos - ( posY * blockWidth );

          scan[scanPosition].idx = rasterPos;
          scan[scanPosition].x   = posX;
          scan[scanPosition].y   = posY;
        }
      }

      //--------------------------------------------------------------------------------------------------

      //grouped scan orders
      const uint32_t* log2Sbb        = g_log2SbbSize[floorLog2(blockWidth)][floorLog2(blockHeight)];
      const uint32_t  log2CGWidth    = log2Sbb[0];
      const uint32_t  log2CGHeight   = log2Sbb[1];

      const uint32_t  groupWidth     = 1 << log2CGWidth;
      const uint32_t  groupHeight    = 1 << log2CGHeight;
      const uint32_t  widthInGroups = std::min<unsigned>(JVET_C0024_ZERO_OUT_TH, blockWidth) >> log2CGWidth;
      const uint32_t  heightInGroups = std::min<unsigned>(JVET_C0024_ZERO_OUT_TH, blockHeight) >> log2CGHeight;

      const uint32_t  groupSize      = groupWidth    * groupHeight;
      const uint32_t  totalGroups    = widthInGroups * heightInGroups;

      for (uint32_t scanTypeIndex = 0; scanTypeIndex < SCAN_NUMBER_OF_TYPES; scanTypeIndex++)
      {
        const CoeffScanType scanType = CoeffScanType(scanTypeIndex);

        ScanElement *scan = new ScanElement[totalValues];

        g_scanOrder[SCAN_GROUPED_4x4][scanType][blockWidthIdx][blockHeightIdx] = scan;

        if ( blockWidth > JVET_C0024_ZERO_OUT_TH || blockHeight > JVET_C0024_ZERO_OUT_TH )
        {
          for (uint32_t i = 0; i < totalValues; i++)
          {
            scan[i].idx = totalValues - 1;
            scan[i].x   = blockWidth - 1;
            scan[i].y   = blockHeight - 1;
          }
        }

        ScanGenerator fullBlockScan(widthInGroups, heightInGroups, groupWidth, scanType);

        for (uint32_t groupIndex = 0; groupIndex < totalGroups; groupIndex++)
        {
          const uint32_t groupPositionY  = fullBlockScan.GetCurrentY();
          const uint32_t groupPositionX  = fullBlockScan.GetCurrentX();
          const uint32_t groupOffsetX    = groupPositionX * groupWidth;
          const uint32_t groupOffsetY    = groupPositionY * groupHeight;
          const uint32_t groupOffsetScan = groupIndex     * groupSize;

          ScanGenerator groupScan(groupWidth, groupHeight, blockWidth, scanType);

          for (uint32_t scanPosition = 0; scanPosition < groupSize; scanPosition++)
          {
            const int rasterPos = groupScan.GetNextIndex( groupOffsetX, groupOffsetY );
            const int posY      = rasterPos / blockWidth;
            const int posX      = rasterPos - ( posY * blockWidth );

            scan[groupOffsetScan + scanPosition].idx = rasterPos;
            scan[groupOffsetScan + scanPosition].x   = posX;
            scan[groupOffsetScan + scanPosition].y   = posY;
          }

          fullBlockScan.GetNextIndex(0, 0);
        }
      }

      //--------------------------------------------------------------------------------------------------
    }
  }

  // initialize CoefTopLeftDiagScan8x8 for LFNST
  for( uint32_t blockWidthIdx = 0; blockWidthIdx < sizeInfo.numAllWidths(); blockWidthIdx++ )
  {
    const uint32_t blockWidth = sizeInfo.sizeFrom( blockWidthIdx );

    const static uint8_t g_auiXYDiagScan8x8[ 64 ][ 2 ] =
    {
      { 0, 0 }, { 0, 1 }, { 1, 0 }, { 0, 2 }, { 1, 1 }, { 2, 0 }, { 0, 3 }, { 1, 2 },
      { 2, 1 }, { 3, 0 }, { 1, 3 }, { 2, 2 }, { 3, 1 }, { 2, 3 }, { 3, 2 }, { 3, 3 },
      { 0, 4 }, { 0, 5 }, { 1, 4 }, { 0, 6 }, { 1, 5 }, { 2, 4 }, { 0, 7 }, { 1, 6 },
      { 2, 5 }, { 3, 4 }, { 1, 7 }, { 2, 6 }, { 3, 5 }, { 2, 7 }, { 3, 6 }, { 3, 7 },
      { 4, 0 }, { 4, 1 }, { 5, 0 }, { 4, 2 }, { 5, 1 }, { 6, 0 }, { 4, 3 }, { 5, 2 },
      { 6, 1 }, { 7, 0 }, { 5, 3 }, { 6, 2 }, { 7, 1 }, { 6, 3 }, { 7, 2 }, { 7, 3 },
      { 4, 4 }, { 4, 5 }, { 5, 4 }, { 4, 6 }, { 5, 5 }, { 6, 4 }, { 4, 7 }, { 5, 6 },
      { 6, 5 }, { 7, 4 }, { 5, 7 }, { 6, 6 }, { 7, 5 }, { 6, 7 }, { 7, 6 }, { 7, 7 }
    };
    for( int i = 0; i < 64; i++ )
    {
      g_coefTopLeftDiagScan8x8[ blockWidthIdx ][ i ].idx = g_auiXYDiagScan8x8[ i ][ 0 ] + g_auiXYDiagScan8x8[ i ][ 1 ] * blockWidth;
      g_coefTopLeftDiagScan8x8[ blockWidthIdx ][ i ].x   = g_auiXYDiagScan8x8[ i ][ 0 ];
      g_coefTopLeftDiagScan8x8[ blockWidthIdx ][ i ].y   = g_auiXYDiagScan8x8[ i ][ 1 ];
    }
#if JVET_W0119_LFNST_EXTENSION
    for( int blky = 0; blky <= 1; blky++ )
    {
      for( int blkx = 0; blkx <= 1; blkx++ )
      {
        for( int i = 0; i < 64; i++ )
        {
          int i4 = blky * 128 + blkx * 64 + i;
          g_coefTopLeftDiagScan16x16[ blockWidthIdx ][ i4 ].idx = g_auiXYDiagScan8x8[ i ][ 0 ] + g_auiXYDiagScan8x8[ i ][ 1 ] * blockWidth + blkx * 8 + blky * 8 * blockWidth;
          g_coefTopLeftDiagScan16x16[ blockWidthIdx ][ i4 ].x   = g_auiXYDiagScan8x8[ i ][ 0 ];
          g_coefTopLeftDiagScan16x16[ blockWidthIdx ][ i4 ].y   = g_auiXYDiagScan8x8[ i ][ 1 ];
        }
      }
    }
#endif
  }

  initGeoTemplate();

  ::memset(g_isReusedUniMVsFilled, 0, sizeof(g_isReusedUniMVsFilled));
#if INTER_LIC
  ::memset(g_isReusedUniMVsFilledLIC, 0, sizeof(g_isReusedUniMVsFilledLIC));
#endif

  for (int qp = 0; qp < 57; qp++)
  {
    int qpRem = (qp + 12) % 6;
    int qpPer = (qp + 12) / 6;
    int quantiserScale = g_quantScales[0][qpRem];
    int quantiserRightShift = QUANT_SHIFT + qpPer;
    double threshQP = ((double)(1 << quantiserRightShift)) / quantiserScale;
    g_paletteQuant[qp] = (int)(threshQP*0.16 + 0.5);
  }

#if SIGN_PREDICTION && !JVET_W0103_INTRA_MTS
  memset(&g_resiBorderTemplate[0][0][0], 0, sizeof(g_resiBorderTemplate));
  const int8_t *p_data = g_initRomSignPred;
  for( int log2Width = 0; log2Width < 6; ++log2Width)
  {
    int width = 4<<log2Width;
    for( int log2Height = 0; log2Height < 6; ++log2Height)
    {
      int height = 4<<log2Height;
      int length = width + height - 1;
      for( int idx = 0; idx < 9; ++idx)
      {
        if( *p_data != 0 && *p_data != 1)
        {
          int j = 0;
          j+=1;
        }
        if(*p_data++ != 0)
        {
          g_resiBorderTemplate[log2Width][log2Height][idx] = p_data;
          p_data += length * 16;
        }
      }
    }
  }
#endif

#if JVET_Y0141_SIGN_PRED_IMPROVE
  memset(&g_resiBorderTemplate[0][0][0], 0, sizeof(g_resiBorderTemplate));
  memset(&g_resiBorderTemplateLFNST[0][0][0], 0, sizeof(g_resiBorderTemplateLFNST));
#endif
#if TU_256
  c = 256;
  const double s = sqrt( ( double ) c ) * (64 << COM16_C806_TRANS_PREC);
  const double PI = 3.14159265358979323846;

  for( int k = 0; k < c; k++ )
  {
    for( int n = 0; n < c; n++ )
    {
      double w0, v;

      // DCT-II
      w0 = ( k == 0 ) ? sqrt( 0.5 ) : 1.0;
      v = cos( PI * ( n + 0.5 ) * k / c ) * w0 * sqrt( 2.0 / c );
      g_trCoreDCT2P256[k][n] = (short)( s * v + ( v > 0 ? 0.5 : -0.5 ) );

      // DCT-VIII
      v = cos( PI * ( k + 0.5 ) * ( n + 0.5 ) / ( c + 0.5 ) ) * sqrt( 2.0 / ( c + 0.5 ) );
      g_trCoreDCT8P256[k][n] = (short)( s * v + ( v > 0 ? 0.5 : -0.5 ) );

      // DST-VII
      v = sin( PI * ( k + 0.5 ) * ( n + 1 ) / ( c + 0.5 ) ) * sqrt( 2.0 / ( c + 0.5 ) );
      g_trCoreDST7P256[k][n] = (short)( s * v + ( v > 0 ? 0.5 : -0.5 ) );
    }
  }
#endif
#if JVET_W0103_INTRA_MTS
#if LMS_LINEAR_MODEL || TRANSFORM_SIMD_OPT || TU_256 
  c = 2;
#else
  int c = 2;
  const double PI = 3.14159265358979323846;
#endif

  for (int i = 0; i < 8; i++)
  {
    const double s = sqrt((double)c) * (64 << COM16_C806_TRANS_PREC);
    TMatrixCoeff *iT = NULL;

    switch (i)
    {
    case 0: iT = g_aiTr2[0][0]; break;
    case 1: iT = g_aiTr4[0][0]; break;
    case 2: iT = g_aiTr8[0][0]; break;
    case 3: iT = g_aiTr16[0][0]; break;
    case 4: iT = g_aiTr32[0][0]; break;
    case 5: iT = g_aiTr64[0][0]; break;
    case 6: iT = g_aiTr128[0][0]; break;
    case 7: iT = g_aiTr256[0][0]; break;
    case 8: exit(0); break;
    }

    for (int k = 0; k < c; k++)
    {
      for (int n = 0; n < c; n++)
      {
        double w0, w1, v;

        // DCT-II
        w0 = k == 0 ? sqrt(0.5) : 1;
        v = cos(PI*(n + 0.5)*k / c) * w0 * sqrt(2.0 / c);
        iT[DCT2*c*c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // DCT-V
        w0 = (k == 0) ? sqrt(0.5) : 1.0;
        w1 = (n == 0) ? sqrt(0.5) : 1.0;
        v = cos(PI*n*k / (c - 0.5)) * w0 * w1 * sqrt(2.0 / (c - 0.5));
        iT[DCT5*c*c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // DCT-VIII
        v = cos(PI*(k + 0.5)*(n + 0.5) / (c + 0.5)) * sqrt(2.0 / (c + 0.5));
        iT[DCT8*c*c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // DST-I
        v = sin(PI*(n + 1)*(k + 1) / (c + 1)) * sqrt(2.0 / (c + 1));
        iT[DST1*c*c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // DST-VII
        v = sin(PI*(k + 0.5)*(n + 1) / (c + 0.5)) * sqrt(2.0 / (c + 0.5));
        iT[DST7*c*c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // DST4
        v = sin(PI * (k + 0.5) * (n + 0.5) / c) * sqrt(2.0 / c);
        iT[DST4 * c * c + k * c + n] = (short)(s * v + (v > 0 ? 0.5 : -0.5));

        // ID
        iT[IDTR * c * c + k * c + n] = (k == n) ? (short)(s + (s > 0 ? 0.5 : -0.5)) : 0;
      }
    }
    c <<= 1;
  }
#endif
#if JVET_AA0107_RMVF_AFFINE_MERGE_DERIVATION
  g_rmvfMultApproxTbl[0] = 0;
  for (int k = 1; k < (3 << sizeof(int64_t)); k++)
  {
    g_rmvfMultApproxTbl[k] = ((1 << 15) + (k >> 1)) / k;
  }
#endif
}

void destroyROM()
{
  unsigned numWidths = gp_sizeIdxInfo->numAllWidths();
  unsigned numHeights = gp_sizeIdxInfo->numAllHeights();

  for (uint32_t groupTypeIndex = 0; groupTypeIndex < SCAN_NUMBER_OF_GROUP_TYPES; groupTypeIndex++)
  {
    for (uint32_t scanOrderIndex = 0; scanOrderIndex < SCAN_NUMBER_OF_TYPES; scanOrderIndex++)
    {
      for (uint32_t blockWidthIdx = 0; blockWidthIdx <= numWidths; blockWidthIdx++)
      {
        for (uint32_t blockHeightIdx = 0; blockHeightIdx <= numHeights; blockHeightIdx++)
        {
          delete[] g_scanOrder[groupTypeIndex][scanOrderIndex][blockWidthIdx][blockHeightIdx];
          g_scanOrder[groupTypeIndex][scanOrderIndex][blockWidthIdx][blockHeightIdx] = nullptr;
        }
      }
    }
  }

  delete gp_sizeIdxInfo;
  gp_sizeIdxInfo = nullptr;

  for( int modeIdx = 0; modeIdx < GEO_NUM_PARTITION_MODE; modeIdx++ )
  {
    delete[] g_GeoParams[modeIdx];
    g_GeoParams[modeIdx] = nullptr;
  }
  delete[] g_GeoParams;
#if JVET_AA0058_GPM_ADP_BLD
  for (int bldIdx = 0; bldIdx < GEO_NUM_BLD; bldIdx++)
  {
    for (int i = 0; i < GEO_NUM_PRESTORED_MASK; i++)
    {
      delete[] g_globalGeoWeights[bldIdx][i];
      g_globalGeoWeights[bldIdx][i] = nullptr;
    }
  }

  for (int i = 0; i < GEO_NUM_PRESTORED_MASK; i++)
  {
    delete[] g_globalGeoEncSADmask[i];
    g_globalGeoEncSADmask[i] = nullptr;
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    delete[] g_globalGeoWeightsTpl[i];
    g_globalGeoWeightsTpl[i] = nullptr;
#endif
  }
#else
  for( int i = 0; i < GEO_NUM_PRESTORED_MASK; i++ )
  {
    delete[] g_globalGeoWeights   [i];
    delete[] g_globalGeoEncSADmask[i];
    g_globalGeoWeights   [i] = nullptr;
    g_globalGeoEncSADmask[i] = nullptr;
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    delete[] g_globalGeoWeightsTpl[i];
    g_globalGeoWeightsTpl[i] = nullptr;
#endif
  }
#endif
#if JVET_Y0141_SIGN_PRED_IMPROVE
  for (int log2Width = 0; log2Width < 6; log2Width++)
  {
    for (int log2Height = 0; log2Height < 6; log2Height++)
    {
      for (int idx = 0; idx < NUM_TRANS_TYPE*NUM_TRANS_TYPE; idx++)
      {
        if (g_resiBorderTemplate[log2Width][log2Height][idx])
        {
          xFree(g_resiBorderTemplate[log2Width][log2Height][idx]);
          g_resiBorderTemplate[log2Width][log2Height][idx] = nullptr;
        }
      }
#if JVET_W0119_LFNST_EXTENSION || EXTENDED_LFNST
      for (int idx = 0; idx < 210; idx++)
#else
      for (int idx = 0; idx < 16; idx++)
#endif
      {
        if (g_resiBorderTemplateLFNST[log2Width][log2Height][idx])
        {
          xFree(g_resiBorderTemplateLFNST[log2Width][log2Height][idx]);
          g_resiBorderTemplateLFNST[log2Width][log2Height][idx] = nullptr;
        }
      }
    }
  }
#endif
}

#if MMLM
int g_aiLMDivTableLow[] = {
  0, 0, 21845, 0, 13107, 43690, 18724, 0, 50972, 39321, 53620, 21845, 15123, 9362, 4369, 0, 3855, 58254, 17246, 52428, 49932, 59578, 25644, 43690, 28835, 40329, 16990, 37449, 56496, 34952, 4228, 0, 61564, 34695, 29959, 29127, 15941, 41391, 26886, 26214, 28771, 24966, 6096, 29789, 23301, 45590, 25098, 21845, 30761, 47185, 1285, 20164, 34622, 41263, 36938, 18724, 49439, 61016, 51095, 17476, 23635, 2114, 16644, 0, 16131, 63550, 9781, 50115, 52238, 14979, 2769, 14563, 49376, 40738, 53302, 20695, 7660, 13443, 37330, 13107, 5663, 14385, 38689, 12483, 771, 3048, 18832, 47662, 23563, 11650, 11522, 22795, 45100, 12549, 55878, 43690, 41213, 48148, 64212, 23592, 57100, 33410, 17815, 10082, 9986, 17311, 31849, 53399, 16233, 51237, 27159, 9362, 63216, 57487, 57557, 63276, 8962, 25547, 47362, 8738, 40621, 11817, 53281, 33825, 18874, 8322, 2064, 0, 2032, 8065, 18009, 31775, 49275, 4890, 29612, 57825, 23918, 58887, 31589, 7489, 52056, 34152, 19248, 7281, 63728, 57456, 53944, 53137, 54979, 59419, 868, 10347, 22273, 36598, 53274, 6721, 27967, 51433, 11540, 39321, 3663, 35599, 4020, 39960, 12312, 52112, 28255, 6241, 51575, 33153, 16479, 1524, 53792, 42184, 32206, 23831, 17031, 11781, 8054, 5825, 5069, 5761, 7878, 11397, 16295, 22550, 30139, 39042, 49238, 60707, 7891, 21845, 37012, 53374, 5377, 24074, 43912, 64874, 21406, 44564, 3260, 28550, 54882, 16705, 45075, 8907, 39258, 5041, 37314, 4993, 39135, 8655, 44613, 15924, 53648, 26699, 604, 40884, 16458, 58386, 35585, 13579, 57895, 37449, 17767, 64376, 46192, 28743, 12019, 61546, 46244, 31638, 17720, 4481, 57448, 45541, 34288, 23681, 13710, 4369, 61185, 53078, 45578, 38676, 32366, 26640, 21491, 16912, 12896, 9437, 6527, 4161, 2331, 1032, 257, 0, 255, 1016, 2277, 4032, 6277, 9004, 12210, 15887, 20031, 24637, 29699, 35213, 41173, 47574, 54411, 61680, 3840, 11959, 20494, 29443, 38801, 48562, 58724, 3744, 14693, 26028, 37746, 49844, 62316, 9624, 22834, 36408, 50342, 64632, 13737, 28728, 44063, 59740, 10219, 26568, 43249, 60257, 12055, 29709, 47682, 434, 19033, 37941, 57155, 11136, 30953, 51067, 5938, 26637, 47624, 3360, 24916, 46751, 3328, 25716, 48376, 5770, 28967, 52428, 10616, 34599, 58840, 17799, 42547, 2010, 27256, 52748, 12947, 38924, 65140, 26056, 52743, 14127, 41277, 3120, 30726, 58555, 21072, 49344, 12300, 41007, 4394, 33530, 62876, 26896, 56659, 21092, 51264, 16103, 46678, 11915, 42886, 8515, 39875, 5890, 37632, 4027, 36145, 2912, 35400, 2534, 35385, 2880, 36089, 3939, 37500, 5698, 39605, 8147, 42395, 11275, 45857, 15069, 49982, 19521, 54758, 24619, 60175, 30353, 688, 36713, 7357, 43690, 14639, 51274, 22522, 59455, 30999, 2688, 40059, 12037, 49693, 21956, 59894, 32437, 5117, 43471, 16425, 55050, 28273, 1630, 40655, 14275, 53561, 27441, 1449, 41120, 15382, 55305, 29818, 4453, 44748, 19629, 60166, 35288, 10529, 51425, 26902, 2496, 43742, 19567, 61042, 37095, 13261, 55074, 31463, 7962, 50106, 26824, 3649, 46117, 23157, 302, 43088, 20442, 63436, 40997, 18660, 61961, 39826, 17792, 61393, 39557, 17819, 61715, 40171, 18724, 62908, 41651, 20489, 64956, 43980, 23096, 2304, 47139, 26529, 6009, 51115, 30773, 10519, 55890, 35811, 15819, 61448, 41628, 21892, 2240, 48208, 28724, 9322, 55538, 36301, 17144, 63604, 44608, 25692, 6855, 53632, 34952, 16349, 63360, 44911, 26539, 8242, 55557, 37410, 19338, 1340, 48951, 31099, 13320, 61149, 43513, 25949, 8456, 56569, 39216, 21932, 4718, 53109, 36031, 19022, 2080, 50741, 33933, 17191, 516, 49441, 32896, 16416, 0,
};
int g_aiLMDivTableHigh[] = {
  65536, 32768, 21845, 16384, 13107, 10922, 9362, 8192, 7281, 6553, 5957, 5461, 5041, 4681, 4369, 4096, 3855, 3640, 3449, 3276, 3120, 2978, 2849, 2730, 2621, 2520, 2427, 2340, 2259, 2184, 2114, 2048, 1985, 1927, 1872, 1820, 1771, 1724, 1680, 1638, 1598, 1560, 1524, 1489, 1456, 1424, 1394, 1365, 1337, 1310, 1285, 1260, 1236, 1213, 1191, 1170, 1149, 1129, 1110, 1092, 1074, 1057, 1040, 1024, 1008, 992, 978, 963, 949, 936, 923, 910, 897, 885, 873, 862, 851, 840, 829, 819, 809, 799, 789, 780, 771, 762, 753, 744, 736, 728, 720, 712, 704, 697, 689, 682, 675, 668, 661, 655, 648, 642, 636, 630, 624, 618, 612, 606, 601, 595, 590, 585, 579, 574, 569, 564, 560, 555, 550, 546, 541, 537, 532, 528, 524, 520, 516, 512, 508, 504, 500, 496, 492, 489, 485, 481, 478, 474, 471, 468, 464, 461, 458, 455, 451, 448, 445, 442, 439, 436, 434, 431, 428, 425, 422, 420, 417, 414, 412, 409, 407, 404, 402, 399, 397, 394, 392, 390, 387, 385, 383, 381, 378, 376, 374, 372, 370, 368, 366, 364, 362, 360, 358, 356, 354, 352, 350, 348, 346, 344, 343, 341, 339, 337, 336, 334, 332, 330, 329, 327, 326, 324, 322, 321, 319, 318, 316, 315, 313, 312, 310, 309, 307, 306, 304, 303, 302, 300, 299, 297, 296, 295, 293, 292, 291, 289, 288, 287, 286, 284, 283, 282, 281, 280, 278, 277, 276, 275, 274, 273, 271, 270, 269, 268, 267, 266, 265, 264, 263, 262, 261, 260, 259, 258, 257, 256, 255, 254, 253, 252, 251, 250, 249, 248, 247, 246, 245, 244, 243, 242, 241, 240, 240, 239, 238, 237, 236, 235, 234, 234, 233, 232, 231, 230, 229, 229, 228, 227, 226, 225, 225, 224, 223, 222, 222, 221, 220, 219, 219, 218, 217, 217, 216, 215, 214, 214, 213, 212, 212, 211, 210, 210, 209, 208, 208, 207, 206, 206, 205, 204, 204, 203, 202, 202, 201, 201, 200, 199, 199, 198, 197, 197, 196, 196, 195, 195, 194, 193, 193, 192, 192, 191, 191, 190, 189, 189, 188, 188, 187, 187, 186, 186, 185, 185, 184, 184, 183, 183, 182, 182, 181, 181, 180, 180, 179, 179, 178, 178, 177, 177, 176, 176, 175, 175, 174, 174, 173, 173, 172, 172, 172, 171, 171, 170, 170, 169, 169, 168, 168, 168, 167, 167, 166, 166, 165, 165, 165, 164, 164, 163, 163, 163, 162, 162, 161, 161, 161, 160, 160, 159, 159, 159, 158, 158, 157, 157, 157, 156, 156, 156, 155, 155, 154, 154, 154, 153, 153, 153, 152, 152, 152, 151, 151, 151, 150, 150, 149, 149, 149, 148, 148, 148, 147, 147, 147, 146, 146, 146, 145, 145, 145, 144, 144, 144, 144, 143, 143, 143, 142, 142, 142, 141, 141, 141, 140, 140, 140, 140, 139, 139, 139, 138, 138, 138, 137, 137, 137, 137, 136, 136, 136, 135, 135, 135, 135, 134, 134, 134, 134, 133, 133, 133, 132, 132, 132, 132, 131, 131, 131, 131, 130, 130, 130, 130, 129, 129, 129, 129, 128, 128, 128, 128,
};
#endif

// ====================================================================================================================
// Data structure related table & variable
// ====================================================================================================================

const int g_quantScales[2][SCALING_LIST_REM_NUM] = // can be represented as a 9 element table
{
    { 26214,23302,20560,18396,16384,14564 },
    { 18396,16384,14564,13107,11651,10280 } // Note: last 3 values of second row == half of the first 3 values of the first row
};

const int g_invQuantScales[2][SCALING_LIST_REM_NUM] = // can be represented as a 9 element table
{
  { 40,45,51,57,64,72 },
  { 57,64,72,80,90,102 } // Note: last 3 values of second row == double of the first 3 values of the first row
};

//--------------------------------------------------------------------------------------------------
//structures
//--------------------------------------------------------------------------------------------------
//coefficients
//--------------------------------------------------------------------------------------------------
// ====================================================================================================================
// Intra prediction
// ====================================================================================================================

const uint8_t g_aucIntraModeNumFast_UseMPM_2D[7 - MIN_CU_LOG2 + 1][7 - MIN_CU_LOG2 + 1] =
{
  {3, 3, 3, 3, 2, 2},  //   4x4,   4x8,   4x16,   4x32,   4x64,   4x128,
  {3, 3, 3, 3, 3, 2},  //   8x4,   8x8,   8x16,   8x32,   8x64,   8x128,
  {3, 3, 3, 3, 3, 2},  //  16x4,  16x8,  16x16,  16x32,  16x64,  16x128,
  {3, 3, 3, 3, 3, 2},  //  32x4,  32x8,  32x16,  32x32,  32x64,  32x128,
  {2, 3, 3, 3, 3, 2},  //  64x4,  64x8,  64x16,  64x32,  64x64,  64x128,
  {2, 2, 2, 2, 2, 3},  // 128x4, 128x8, 128x16, 128x32, 128x64, 128x128,
};

const uint8_t g_aucIntraModeNumFast_UseMPM[MAX_CU_DEPTH] =
{
  3,  //   2x2
  8,  //   4x4
  8,  //   8x8
  3,  //  16x16
  3,  //  32x32
  3,  //  64x64
  3   // 128x128
};
const uint8_t g_aucIntraModeNumFast_NotUseMPM[MAX_CU_DEPTH] =
{
  3,  //   2x2
  9,  //   4x4
  9,  //   8x8
  4,  //  16x16   33
  4,  //  32x32   33
  5,  //  64x64   33
  5   // 128x128
};

const uint8_t g_chroma422IntraAngleMappingTable[NUM_INTRA_MODE] =
//                                    *                                H                              *                                D      *   *   *   *       *   *   *                   *        V       *                   *   *   *      *   *   *   *
//0, 1,  2,  3,  4,  5,  6,  7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 65, 66, DM
{ 0, 1, 61, 62, 63, 64, 65, 66, 2, 3,  5,  6,  8, 10, 12, 13, 14, 16, 18, 20, 22, 23, 24, 26, 28, 30, 31, 33, 34, 35, 36, 37, 38, 39, 40, 41, 41, 42, 43, 43, 44, 44, 45, 45, 46, 47, 48, 48, 49, 49, 50, 51, 51, 52, 52, 53, 54, 55, 55, 56, 56, 57, 57, 58, 59, 59, 60, DM_CHROMA_IDX };



// ====================================================================================================================
// Misc.
// ====================================================================================================================
SizeIndexInfo*           gp_sizeIdxInfo = NULL;

const int                 g_ictModes[2][4] = { { 0, 3, 1, 2 }, { 0, -3, -1, -2 } };

#if LMS_LINEAR_MODEL || TRANSFORM_SIMD_OPT
int8_t                    g_aucLog2[MAX_CU_SIZE + 1];
#endif
#if LMS_LINEAR_MODEL
int8_t                    g_aucNextLog2[MAX_CU_SIZE + 1];
int8_t                    g_aucPrevLog2[MAX_CU_SIZE + 1];
#endif

#if MMLM && !LMS_LINEAR_MODEL
int8_t                    g_aucPrevLog2[MAX_CU_SIZE + 1];
#endif

UnitScale g_miScaling( MIN_CU_LOG2, MIN_CU_LOG2 );


// ====================================================================================================================
// Scanning order & context model mapping
// ====================================================================================================================

// scanning order table
ScanElement *g_scanOrder[SCAN_NUMBER_OF_GROUP_TYPES][SCAN_NUMBER_OF_TYPES][MAX_CU_SIZE / 2 + 1][MAX_CU_SIZE / 2 + 1];
#if JVET_W0119_LFNST_EXTENSION
ScanElement  g_coefTopLeftDiagScan8x8  [ MAX_CU_DEPTH + 1 ][  64 ];
ScanElement  g_coefTopLeftDiagScan16x16[ MAX_CU_DEPTH + 1 ][ 256 ];
#else
ScanElement  g_coefTopLeftDiagScan8x8[ MAX_CU_SIZE / 2 + 1 ][ 64 ];
#endif

#if TU_256
// starting position of the coefficient in a group
// 2^suffix_length * ( 2 + (prefix & 1) )
const uint32_t g_uiMinInGroup[LAST_SIGNIFICANT_GROUPS] = { 0,1,2,3,4,6,8,12,16,24,32,48,64,96,128,192 };
#else
const uint32_t g_uiMinInGroup[LAST_SIGNIFICANT_GROUPS] = { 0,1,2,3,4,6,8,12,16,24,32,48,64,96 };
#endif

// suffix_length = prefix/2 - 1
// lastPos = 2^suffix_length * ( 2 + (prefix & 1) ) + suffix
// length of the prefix code
#if TU_256
const uint32_t g_uiGroupIdx[] = { 0,1,2,3,4,4,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9, 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11
,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13
,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14
,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15
};
#else
const uint32_t g_uiGroupIdx[MAX_TB_SIZEY] = { 0,1,2,3,4,4,5,5,6,6,6,6,7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,9,9,9,9, 10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11,11 };
#endif
const uint32_t g_auiGoRiceParsCoeff[32] =
{
  0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3
};

#if TCQ_8STATES
const uint64_t g_stateTransTab[3] = { UINT64_C(0), UINT64_C(587280912), UINT64_C(6274468666787591456) };
// the values represent the following state transition tables (curr : next[parity=0]  next[parity=1])
//
//    (1) no dep quant:  0 : 0 0
//
//    (2) 4 states:      0 : 0 1
//                       1 : 2 3
//                       2 : 1 0
//                       3 : 3 2
//
//    (3) 8 states:      0 : 0 2
//                       1 : 5 7
//                       2 : 1 3
//                       3 : 6 4
//                       4 : 2 0
//                       5 : 4 6
//                       6 : 3 1
//                       7 : 7 5
#endif

const char *MatrixType[SCALING_LIST_SIZE_NUM][SCALING_LIST_NUM] =
{
  {
    "INTRA1X1_LUMA",
    "INTRA1X1_CHROMAU",
    "INTRA1X1_CHROMAV",
    "INTER1X1_LUMA",
    "INTER1X1_CHROMAU",
    "INTER1X1_CHROMAV"
  },
  {
    "INTRA2X2_LUMA",
    "INTRA2X2_CHROMAU",
    "INTRA2X2_CHROMAV",
    "INTER2X2_LUMA",
    "INTER2X2_CHROMAU",
    "INTER2X2_CHROMAV"
  },
  {
    "INTRA4X4_LUMA",
    "INTRA4X4_CHROMAU",
    "INTRA4X4_CHROMAV",
    "INTER4X4_LUMA",
    "INTER4X4_CHROMAU",
    "INTER4X4_CHROMAV"
  },
  {
    "INTRA8X8_LUMA",
    "INTRA8X8_CHROMAU",
    "INTRA8X8_CHROMAV",
    "INTER8X8_LUMA",
    "INTER8X8_CHROMAU",
    "INTER8X8_CHROMAV"
  },
  {
    "INTRA16X16_LUMA",
    "INTRA16X16_CHROMAU",
    "INTRA16X16_CHROMAV",
    "INTER16X16_LUMA",
    "INTER16X16_CHROMAU",
    "INTER16X16_CHROMAV"
  },
  {
    "INTRA32X32_LUMA",
    "INTRA32X32_CHROMAU",
    "INTRA32X32_CHROMAV",
    "INTER32X32_LUMA",
    "INTER32X32_CHROMAU",
    "INTER32X32_CHROMAV"
  },
  {
    "INTRA64X64_LUMA",
    "INTRA64X64_CHROMAU",
    "INTRA64X64_CHROMAV",
    "INTER64X64_LUMA",
    "INTER64X64_CHROMAU",
    "INTER64X64_CHROMAV"
  },
  {
  },
};

const char *MatrixType_DC[SCALING_LIST_SIZE_NUM][SCALING_LIST_NUM] =
{
  {  //1x1
  },
  {
  },
  {
  },
  {
  },
  {
    "INTRA16X16_LUMA_DC",
    "INTRA16X16_CHROMAU_DC",
    "INTRA16X16_CHROMAV_DC",
    "INTER16X16_LUMA_DC",
    "INTER16X16_CHROMAU_DC",
    "INTER16X16_CHROMAV_DC"
  },
  {
    "INTRA32X32_LUMA_DC",
    "INTRA32X32_CHROMAU_DC",
    "INTRA32X32_CHROMAV_DC",
    "INTER32X32_LUMA_DC",
    "INTER32X32_CHROMAU_DC",
    "INTER32X32_CHROMAV_DC"
  },
  {
    "INTRA64X64_LUMA_DC",
    "INTRA64X64_CHROMAU_DC",
    "INTRA64X64_CHROMAV_DC",
    "INTER64X64_LUMA_DC",
    "INTER64X64_CHROMAU_DC",
    "INTER64X64_CHROMAV_DC"
  },
  {
  },
};

const int g_quantTSDefault4x4[4 * 4] =
{
  16,16,16,16,
  16,16,16,16,
  16,16,16,16,
  16,16,16,16
};

const int g_quantIntraDefault8x8[8 * 8] =
{
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16
};

const int g_quantInterDefault8x8[8 * 8] =
{
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16,
  16,16,16,16,16,16,16,16
};

const uint32_t g_scalingListSize [SCALING_LIST_SIZE_NUM] = { 1, 4, 16, 64, 256, 1024, 4096, 16384 };
const uint32_t g_scalingListSizeX[SCALING_LIST_SIZE_NUM] = { 1, 2,  4,  8,  16,   32,   64,   128 };

const uint32_t g_scalingListId[SCALING_LIST_SIZE_NUM][SCALING_LIST_NUM] =
{
  {  0,  0,  0,  0,  0,  0},  // SCALING_LIST_1x1
  {  0,  0,  0,  0,  0,  1},  // SCALING_LIST_2x2
  {  2,  3,  4,  5,  6,  7},  // SCALING_LIST_4x4
  {  8,  9, 10, 11, 12, 13},  // SCALING_LIST_8x8
  { 14, 15, 16, 17, 18, 19},  // SCALING_LIST_16x16
  { 20, 21, 22, 23, 24, 25},  // SCALING_LIST_32x32
  { 26, 21, 22, 27, 24, 25},  // SCALING_LIST_64x64
  {  0,  0,  0,  0,  0,  0},  // SCALING_LIST_128x128
};

#if CTU_256
Mv   g_reusedUniMVs        [MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_DEPTH + 1][MAX_CU_DEPTH + 1][2][33];
bool g_isReusedUniMVsFilled[MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_DEPTH + 1][MAX_CU_DEPTH + 1];
#else
Mv   g_reusedUniMVs[32][32][8][8][2][33];
bool g_isReusedUniMVsFilled[32][32][8][8];
#endif
#if INTER_LIC
Mv   g_reusedUniMVsLIC        [MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_DEPTH + 1][MAX_CU_DEPTH + 1][2][33];
bool g_isReusedUniMVsFilledLIC[MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_SIZE >> MIN_CU_LOG2][MAX_CU_DEPTH + 1][MAX_CU_DEPTH + 1];
#endif

uint16_t g_paletteQuant[57];
uint8_t g_paletteRunTopLut [5] = { 0, 1, 1, 2, 2 };
uint8_t g_paletteRunLeftLut[5] = { 0, 1, 2, 3, 4 };

void initGeoTemplate()
{
  g_GeoParams = new int16_t*[GEO_NUM_PARTITION_MODE];
  int modeIdx = 0;
  for( int angleIdx = 0; angleIdx < GEO_NUM_ANGLES; angleIdx++ )
  {
    for( int distanceIdx = 0; distanceIdx < GEO_NUM_DISTANCES; distanceIdx++ )
    {
      if( (distanceIdx == 0 && angleIdx >= 16)
        || ((distanceIdx == 2 || distanceIdx == 0) && (g_angle2mask[angleIdx] == 0 || g_angle2mask[angleIdx] == 5))
        || g_angle2mask[angleIdx] == -1 )
      {
        continue;
      }
      g_GeoParams[modeIdx]    = new int16_t[2];
      g_GeoParams[modeIdx][0] = (int16_t)angleIdx;
      g_GeoParams[modeIdx][1] = (int16_t)distanceIdx;
      modeIdx++;
    }
  }
#if JVET_AA0058_GPM_ADP_BLD
  // initialization of blending weights
  for (int angleIdx = 0; angleIdx < (GEO_NUM_ANGLES >> 2) + 1; angleIdx++)
  {
    if (g_angle2mask[angleIdx] == -1)
    {
      continue;
    }
    for (int bldIdx = 0; bldIdx < GEO_NUM_BLD; bldIdx++)
    {
      g_globalGeoWeights[bldIdx][g_angle2mask[angleIdx]] = new int16_t[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];

      int distanceX = angleIdx;
      int distanceY = (distanceX + (GEO_NUM_ANGLES >> 2)) % GEO_NUM_ANGLES;
      int16_t rho = (g_Dis[distanceX] << (GEO_MAX_CU_LOG2 + 1)) + (g_Dis[distanceY] << (GEO_MAX_CU_LOG2 + 1));
      static const int16_t maskOffset = (2 * GEO_MAX_CU_SIZE - GEO_WEIGHT_MASK_SIZE) >> 1;
      int index = 0;
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
      for (int y = 0; y < GEO_WEIGHT_MASK_SIZE; y++)
      {
        int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
        for (int x = 0; x < GEO_WEIGHT_MASK_SIZE; x++, index++)
        {
          int16_t sx_i = ((x + maskOffset) << 1) + 1;
          int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;

          if (g_bld2Width[bldIdx] > 1)
          {
            int weightLinearIdx = 8 * g_bld2Width[bldIdx] + weightIdx;
            g_globalGeoWeights[bldIdx][g_angle2mask[angleIdx]][index] = Clip3(0, 32, (weightLinearIdx + (g_bld2Width[bldIdx] >> 2)) >> floorLog2(g_bld2Width[bldIdx] >> 1));
          }
          else
          {
            int weightLinearIdx = 8 + weightIdx;
            g_globalGeoWeights[bldIdx][g_angle2mask[angleIdx]][index] = Clip3(0, 32, weightLinearIdx << 1);
          }
        }
      }
#else
      for (int y = 0; y < GEO_WEIGHT_MASK_SIZE; y++)
      {
        int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
        for (int x = 0; x < GEO_WEIGHT_MASK_SIZE; x++, index++)
        {
          int16_t sx_i = ((x + maskOffset) << 1) + 1;
          int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;
          if (g_bld2Width[bldIdx] > 1)
          {
            int weightLinearIdx = 8 * g_bld2Width[bldIdx] + weightIdx;
            g_globalGeoWeights[bldIdx][g_angle2mask[angleIdx]][index] = Clip3(0, 32, (weightLinearIdx + (g_bld2Width[bldIdx] >> 2)) >> floorLog2(g_bld2Width[bldIdx] >> 1));
          }
          else
          {
            int weightLinearIdx = 8 + weightIdx;
            g_globalGeoWeights[bldIdx][g_angle2mask[angleIdx]][index] = Clip3(0, 32, weightLinearIdx << 1);
          }
        }
      }
#endif
    }
  }

  // initialization of mask weights
  for (int angleIdx = 0; angleIdx < (GEO_NUM_ANGLES >> 2) + 1; angleIdx++)
  {
    if (g_angle2mask[angleIdx] == -1)
    {
      continue;
    }
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    g_globalGeoWeightsTpl[g_angle2mask[angleIdx]] = new Pel[GEO_WEIGHT_MASK_SIZE_EXT * GEO_WEIGHT_MASK_SIZE_EXT];
#endif
#if JVET_R0351_HIGH_BIT_DEPTH_SUPPORT
    g_globalGeoEncSADmask[g_angle2mask[angleIdx]] = new Pel[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];
#else
    g_globalGeoEncSADmask[g_angle2mask[angleIdx]] = new int16_t[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];
#endif

    int distanceX = angleIdx;
    int distanceY = (distanceX + (GEO_NUM_ANGLES >> 2)) % GEO_NUM_ANGLES;
    int16_t rho = (g_Dis[distanceX] << (GEO_MAX_CU_LOG2 + 1)) + (g_Dis[distanceY] << (GEO_MAX_CU_LOG2 + 1));
    static const int16_t maskOffset = (2 * GEO_MAX_CU_SIZE - GEO_WEIGHT_MASK_SIZE) >> 1;
    int index = 0;
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    int indexGeoWeight = 0;
    for (int y = -GEO_TM_ADDED_WEIGHT_MASK_SIZE; y < GEO_WEIGHT_MASK_SIZE + GEO_TM_ADDED_WEIGHT_MASK_SIZE; y++)
    {
      int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
      for (int x = -GEO_TM_ADDED_WEIGHT_MASK_SIZE; x < GEO_WEIGHT_MASK_SIZE + GEO_TM_ADDED_WEIGHT_MASK_SIZE; x++)
      {
        int16_t sx_i = ((x + maskOffset) << 1) + 1;
        int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;
        g_globalGeoWeightsTpl[g_angle2mask[angleIdx]][index++] = weightIdx > 0 ? 1 : 0;
        if (x >= 0 && x < GEO_WEIGHT_MASK_SIZE && y >= 0 && y < GEO_WEIGHT_MASK_SIZE)
        {
          g_globalGeoEncSADmask[g_angle2mask[angleIdx]][indexGeoWeight] = weightIdx > 0 ? 1 : 0;
          ++indexGeoWeight;
        }
      }
    }
#else
    for (int y = 0; y < GEO_WEIGHT_MASK_SIZE; y++)
    {
      int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
      for (int x = 0; x < GEO_WEIGHT_MASK_SIZE; x++, index++)
      {
        int16_t sx_i = ((x + maskOffset) << 1) + 1;
        int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;
        g_globalGeoEncSADmask[g_angle2mask[angleIdx]][index] = weightIdx > 0 ? 1 : 0;
      }
    }
#endif
  }
#else
  for (int angleIdx = 0; angleIdx < (GEO_NUM_ANGLES >> 2) + 1; angleIdx++)
  {
    if (g_angle2mask[angleIdx] == -1)
    {
      continue;
    }
    g_globalGeoWeights[g_angle2mask[angleIdx]] = new int16_t[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    g_globalGeoWeightsTpl[g_angle2mask[angleIdx]] = new Pel[GEO_WEIGHT_MASK_SIZE_EXT * GEO_WEIGHT_MASK_SIZE_EXT];
#endif
#if JVET_R0351_HIGH_BIT_DEPTH_SUPPORT
    g_globalGeoEncSADmask[g_angle2mask[angleIdx]] = new Pel[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];
#else
    g_globalGeoEncSADmask[g_angle2mask[angleIdx]] = new int16_t[GEO_WEIGHT_MASK_SIZE * GEO_WEIGHT_MASK_SIZE];
#endif

    int distanceX = angleIdx;
    int distanceY = (distanceX + (GEO_NUM_ANGLES >> 2)) % GEO_NUM_ANGLES;
    int16_t rho = (g_Dis[distanceX] << (GEO_MAX_CU_LOG2+1)) + (g_Dis[distanceY] << (GEO_MAX_CU_LOG2 + 1));
    static const int16_t maskOffset = (2*GEO_MAX_CU_SIZE - GEO_WEIGHT_MASK_SIZE) >> 1;
    int index = 0;
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
    int indexGeoWeight = 0;
    for( int y = -GEO_TM_ADDED_WEIGHT_MASK_SIZE; y < GEO_WEIGHT_MASK_SIZE + GEO_TM_ADDED_WEIGHT_MASK_SIZE; y++ )
    {
      int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
      for( int x = -GEO_TM_ADDED_WEIGHT_MASK_SIZE; x < GEO_WEIGHT_MASK_SIZE + GEO_TM_ADDED_WEIGHT_MASK_SIZE; x++ )
      {
        int16_t sx_i = ((x + maskOffset) << 1) + 1;
        int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;
        int weightLinearIdx = 32 + weightIdx;
        g_globalGeoWeightsTpl[g_angle2mask[angleIdx]][index++] = weightIdx > 0 ? 1 : 0;
        if (x >= 0 && x < GEO_WEIGHT_MASK_SIZE && y >=0 && y < GEO_WEIGHT_MASK_SIZE)
        {
          g_globalGeoWeights   [g_angle2mask[angleIdx]][indexGeoWeight] = Clip3(0, 8, (weightLinearIdx + 4) >> 3);
          g_globalGeoEncSADmask[g_angle2mask[angleIdx]][indexGeoWeight] = weightIdx > 0 ? 1 : 0;
          ++indexGeoWeight;
        }
      }
    }
#else
    for( int y = 0; y < GEO_WEIGHT_MASK_SIZE; y++ )
    {
      int16_t lookUpY = (((y + maskOffset) << 1) + 1) * g_Dis[distanceY];
      for( int x = 0; x < GEO_WEIGHT_MASK_SIZE; x++, index++ )
      {
        int16_t sx_i = ((x + maskOffset) << 1) + 1;
        int16_t weightIdx = sx_i * g_Dis[distanceX] + lookUpY - rho;
        int weightLinearIdx = 32 + weightIdx;
        g_globalGeoWeights[g_angle2mask[angleIdx]][index] = Clip3(0, 8, (weightLinearIdx + 4) >> 3);
        g_globalGeoEncSADmask[g_angle2mask[angleIdx]][index] = weightIdx > 0 ? 1 : 0;
      }
    }
#endif
  }
#endif

  for( int hIdx = 0; hIdx < GEO_NUM_CU_SIZE; hIdx++ )
  {
    int16_t height = 1 << ( hIdx + GEO_MIN_CU_LOG2);
    for( int wIdx = 0; wIdx < GEO_NUM_CU_SIZE; wIdx++ )
    {
      int16_t width = 1 << (wIdx + GEO_MIN_CU_LOG2);
      for( int splitDir = 0; splitDir < GEO_NUM_PARTITION_MODE; splitDir++ )
      {
        int16_t angle         = g_GeoParams[splitDir][0];
        int16_t distance      = g_GeoParams[splitDir][1];
        int16_t offsetX       = (GEO_WEIGHT_MASK_SIZE - width) >> 1;
        int16_t offsetY       = (GEO_WEIGHT_MASK_SIZE - height) >> 1;
        if( distance > 0 )
        {
          if( angle % 16 == 8 || (angle % 16 != 0 && height >= width) )
          {
            offsetY += angle < 16 ? ((distance * (int32_t)height) >> 3) : -((distance * (int32_t)height) >> 3);
          }
          else
          {
            offsetX += angle < 16 ? ((distance * (int32_t)width) >> 3) : -((distance * (int32_t)width) >> 3);
          }
        }
        g_weightOffset[splitDir][hIdx][wIdx][0] = offsetX;
        g_weightOffset[splitDir][hIdx][wIdx][1] = offsetY;
      }
    }
  }


}

int16_t** g_GeoParams;
#if JVET_AA0058_GPM_ADP_BLD
int16_t*  g_globalGeoWeights[GEO_NUM_BLD][GEO_NUM_PRESTORED_MASK];
int       g_bld2Width[GEO_NUM_BLD] = { 1, 2, 4, 8, 16 };
#else
int16_t*  g_globalGeoWeights   [GEO_NUM_PRESTORED_MASK];
#endif
#if JVET_Z0056_GPM_SPLIT_MODE_REORDERING
Pel*      g_globalGeoWeightsTpl[GEO_NUM_PRESTORED_MASK];
#endif
#if JVET_R0351_HIGH_BIT_DEPTH_SUPPORT
Pel*      g_globalGeoEncSADmask[GEO_NUM_PRESTORED_MASK];
#else
int16_t*  g_globalGeoEncSADmask[GEO_NUM_PRESTORED_MASK];
#endif
int16_t   g_weightOffset       [GEO_NUM_PARTITION_MODE][GEO_NUM_CU_SIZE][GEO_NUM_CU_SIZE][2];
int8_t    g_angle2mask[GEO_NUM_ANGLES] = { 0, -1, 1, 2, 3, 4, -1, -1, 5, -1, -1, 4, 3, 2, 1, -1, 0, -1, 1, 2, 3, 4, -1, -1, 5, -1, -1, 4, 3, 2, 1, -1 };
int8_t    g_Dis[GEO_NUM_ANGLES] = { 8, 8, 8, 8, 4, 4, 2, 1, 0, -1, -2, -4, -4, -8, -8, -8, -8, -8, -8, -8, -4, -4, -2, -1, 0, 1, 2, 4, 4, 8, 8, 8 };
int8_t    g_angle2mirror[GEO_NUM_ANGLES] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 2 };
#if JVET_Y0065_GPM_INTRA
int8_t    g_geoAngle2IntraAng[GEO_NUM_ANGLES] = {50, 0, 44, 41, 34, 27, 0, 0, 18, 0, 0, 9, 66, 59, 56, 0, 50, 0, 44, 41, 34, 27, 0, 0, 18, 0, 0, 9, 66, 59, 56, 0};
#endif
#if MULTI_HYP_PRED
const int g_addHypWeight[MULTI_HYP_PRED_NUM_WEIGHTS] = { 2, -1 };
static_assert(g_BcwLog2WeightBase == MULTI_HYP_PRED_WEIGHT_BITS, "number of bits for gbi and multi-hyp weights do not match");
#endif
#if (JVET_W0097_GPM_MMVD_TM && TM_MRG) || JVET_Y0065_GPM_INTRA
uint8_t g_geoTmShape[2][GEO_NUM_ANGLES] = {
                                          { GEO_TM_SHAPE_A,  0,               GEO_TM_SHAPE_A,  GEO_TM_SHAPE_A,
                                            GEO_TM_SHAPE_A,  GEO_TM_SHAPE_AL, 0,               0,
                                            GEO_TM_SHAPE_AL, 0,               0,               GEO_TM_SHAPE_AL,
                                            GEO_TM_SHAPE_AL, GEO_TM_SHAPE_A,  GEO_TM_SHAPE_A,  0,
                                            GEO_TM_SHAPE_A,  0,               GEO_TM_SHAPE_A,  GEO_TM_SHAPE_A,
                                            GEO_TM_SHAPE_A,  GEO_TM_SHAPE_AL, 0,               0,
                                            GEO_TM_SHAPE_AL, 0,               0,               GEO_TM_SHAPE_AL,
                                            GEO_TM_SHAPE_AL, GEO_TM_SHAPE_A,  GEO_TM_SHAPE_A,  0, },
                                          { GEO_TM_SHAPE_AL, 0,               GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL,
                                            GEO_TM_SHAPE_L,  GEO_TM_SHAPE_L,  0,               0,
                                            GEO_TM_SHAPE_L,  0,               0,               GEO_TM_SHAPE_L,
                                            GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL, 0,
                                            GEO_TM_SHAPE_AL, 0,               GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL,
                                            GEO_TM_SHAPE_L,  GEO_TM_SHAPE_L,  0,               0,
                                            GEO_TM_SHAPE_L,  0,               0,               GEO_TM_SHAPE_L,
                                            GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL, GEO_TM_SHAPE_AL, 0, } };
#endif
#if JVET_W0066_CCSAO
const int8_t g_ccSaoCandPosX[MAX_NUM_LUMA_COMP][MAX_CCSAO_CAND_POS_Y] = { {-1,  0,  1, -1,  0,  1, -1,  0,  1} };
const int8_t g_ccSaoCandPosY[MAX_NUM_LUMA_COMP][MAX_CCSAO_CAND_POS_Y] = { {-1, -1, -1,  0,  0,  0,  1,  1,  1} };
#endif
#if JVET_Y0106_CCSAO_EDGE_CLASSIFIER
const int8_t g_ccSaoEdgeTypeX[CCSAO_EDGE_TYPE][2] = { { -1, 1 }, { 0, 0 }, { -1, 1 }, { 1, -1 } };
const int8_t g_ccSaoEdgeTypeY[CCSAO_EDGE_TYPE][2] = { { 0, 0 }, { -1, 1 }, { -1, 1 }, { -1, 1 } };
const short  g_ccSaoQuanValue[CCSAO_QUAN_NUM]      = { 2, 4, 6, 8, 10, 14, 18, 22, 30, 38, 54, 70, 86, 118, 150, 182 };
#endif
#if JVET_V0130_INTRA_TMP
unsigned int g_uiDepth2Width[5] = { 4, 8, 16, 32, 64 };
#endif
#if JVET_X0149_TIMD_DIMD_LUT
int g_gradDivTable[16] = { 0, 7, 6, 5, 5, 4, 4, 3, 3, 2, 2, 1, 1, 1, 1, 0 };
#endif
#if JVET_AA0107_RMVF_AFFINE_MERGE_DERIVATION
int g_rmvfMultApproxTbl[3 << sizeof(int64_t)];
#endif
//! \}
